# Copyright (C) 2014 Association of Universities for Research in Astronomy, Inc.
# Contact: mschirme@gemini.edu
#  
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software 
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# $Id: vmAstroCat.tcl,v 1.10 2013/08/14 15:46:39 gmmps Exp $
#
#***********************************************************************
#****     D A O   I N S T R U M E N T A T I O N   G R O U P        *****
#
# (c) <year>				(c) <year>
# National Research Council		Conseil national de recherches
# Ottawa, Canada, K1A 0R6 		Ottawa, Canada, K1A 0R6
# All rights reserved			Tous droits reserves
# 					
# NRC disclaims any warranties,		Le CNRC denie toute garantie
# expressed, implied, or statu-		enoncee, implicite ou legale,
# tory, of any kind with respect	de quelque nature que se soit,
# to the software, including		concernant le logiciel, y com-
# without limitation any war-		pris sans restriction toute
# ranty of merchantability or		garantie de valeur marchande
# fitness for a particular pur-		ou de pertinence pour un usage
# pose.  NRC shall not be liable	particulier.  Le CNRC ne
# in any event for any damages,		pourra en aucun cas etre tenu
# whether direct or indirect,		responsable de tout dommage,
# special or general, consequen-	direct ou indirect, particul-
# tial or incidental, arising		ier ou general, accessoire ou
# from the use of the software.		fortuit, resultant de l'utili-
# 					sation du logiciel.
#
#***********************************************************************
#
# FILENAME
# vmAstroCat.tcl
#
# PURPOSE:
# User interface class for viewing catalog info
#
# WARNING:
# This needs cleanup.  There are a LOT of functions that are never called.
# Also, the catalog file is both kept and used in memory AND for some weird
# reason also written to - WITHOUT being asked.  A lot of re-processing of the
# catalog when the darn thing is already in memory.  Its almost like 2 different
# people worked on it with completely different ideas.  Anyway, since we
# only have time for modifications - I have followed this haphazard way
# of dealing with the catalog file - so sorry.
#
# CLASS NAME(S)
# itk::usual vmAstroCat - Main menu class
#
# METHOD/PROC NAME(S)
#  constructor		   - constructor of the class
#  destructor		   - destructor of the class
#  init			   - called after options have been eval'ed.
#  set_logfile		   - setup logfile in .skycat dir.
#  readInFits		   - read in a fits file and create a catalog.
#  set_feedback		   - setup feeback for HTTP transfers.
#  set_instance		   - keep array of instances.
#  add_menubar		   - add menubar in object table window
#  add_menu_state	   - enable/disable some menues depending on selection.
#  add_catalog_menu	   - adds the catalog menu item under GMMS
#  reload_config_file	   - re-load default skycat catalog config file.
#  check_proxies	   - check for proxy server and init if nec.
#  catalog_directory	   - pops up window to browse the cat. dir.
#  update_catalog_menus    - update all cat. menus to show current catalog.
#  my_set_plot_symbols     - Dialog to set plot symbols for a catalog.
#  set_search_cols         - Dialog to set search columns.
#  update_search_cols	   - Update search option after edit.
#  close		   - close this window
#  sort_dialog		   - sort list in dialog
#  set_sort_cols	   - called when user has set col's to sort by
#  select_columns	   - pop up a dialog to select table columns
#  set_show_cols	   - called when user has selected col's to show
#  get_name_servers	   - adds name server catalogs to menu
#  reset_table		   - reset table dialogs if needed.
#  clear		   - clear the table listing
#  print		   - popup to print table listing
#  local_catalog	   - Called when Load from file is requested
#  fill_catalog_menu	   - fills menu list with catalogs to load
#  open_catalog_window     - opens a window for the given catalog
#  save_as		   - save the current data to a local catalog
#  add_to		   - add the rows in the list to a catalog file.
#  add_select		   - add the currently selected rows to a catalog
#  remove_selected	   - remove selected froms fro a catalog
#  enter_new_object	   - dialog to enter data for a new object to cat.
#  edit_selected_object    - dialog to allow user to edit object
#  slits		   - draws the slits for ODF files.
#  clearslits		   - clears the slits drawn from ODF files.
#  _sel			   - updates the priority col. in cat for an object
#  open_catalog		   - opens the catalog for a window
#  add_result_table	   - fills in a results_ table from a catalog
#  arch			   - called when a curved slit button is pressed
#  mps_add_dialog_buttons  - fills in the buttons in an object table window.
#  slit_add_dialog_buttons - fills in the buttons in an ODF.
#  add_dialog_buttons      - Adds dialog buttons to some window?
#  conv			   - Adds missing columns to a catalog, Image co-ordinates
#  symbol		   - Write image & symbol information from image to cat.
#  col_PRIO		   - Add the PRIOIRTY column.
#  col_ST		   - Add the SLITTYPE column.
#  col_STSIZE		   - Add the slitsize_y column.
#  export		   - Runs the slit algorithm stuff.
#  disable_search	   - Disable search
#  help			   - Help menu
#  add_progress_bar	   - Add progress bar.
#  layout_dialog	   - Does the dialog window for a new catalog.
#  make_short_help	   - Create short help info.
#  select_catalog	   - Called when a catalog is selected from a menu
#  set_namesvr		   - Set the name server to use to resolve object names
#  interrupt		   - Interrupt the current search
#  set_state		   - Set/reset widget states while busy
#  feedback		   - Provide feedback.
#  set_info		   - Set contents of a listing.
#  query_done		   - Called when background query is done.
#  search		   - Called to do search in catalog.
#  getimage_from_args      - Get image from arg's.
#  check_local_catalog     - Check catalog file.
#  set_pos_radius	   - Set position and radius
#  set_pos_width_height	   - Set position, width and height entries
#  select_result_row       - Called when a result row is selected.
#  preview		   - Get preview of URL
#  get_preview		   - Get the preview
#  preview_done		   - Called when preview data is recieved.
#  send_to_netscape        - Send URL to be displayed by netscape.
#  more			   - Use the more URL to display more info.
#  get_col		   - Return given column val in a row.
#  new_catalog		   - Open a window for the named catalog.
#  instances		   - Return list of instance for this class.
#  get_instance		   - Return widget instance for a catalog.
#  iscat		   - Return 1 if window is for searchable catalog.

#### Then various classes to deal with images and plotting
#  plot, display_image_file, get_equinox, get_table, ...
#
# Original taken from:
# E.S.O. - VLT project/ESO Archive
# @(#) $Id: vmAstroCat.tcl,v 1.10 2013/08/14 15:46:39 gmmps Exp $
# Created originally by:
# D.Bottini, 01 Apr 00
#
# $Log: vmAstroCat.tcl,v $
# Revision 1.10  2013/08/14 15:46:39  gmmps
# Set instType and targetInst to upper case after reading INSTRUME from header
#
# Revision 1.9  2013/06/10 15:56:00  gmmps
# added workaround to make reading of PIXSCALE from image header more reliable
#
# Revision 1.8  2013/03/18 21:17:56  gmmps
# Added procedure "get_ccdsum_image" in vmAstroCat.tcl to get CCDSUM if nothing returned by the original method
#
# Revision 1.7  2013/02/12 23:59:22  gmmps
# Fix type in comment related to slittype column test
#
# Revision 1.6  2013/02/04 18:38:06  gmmps
# Small features and bug fixes from Dec2012 and Jan2013.
#
# Revision 1.5  2011/08/18 22:37:18  gmmps
# Bugfixes and small improvements, focus on mask design algorithm
#
# Revision 1.4  2011/04/29 16:31:03  gmmps
# Quieted down GMMPS in terminal and cleaned up src folder.
#
# Revision 1.3  2011/04/29 02:26:50  gmmps
# Slit wiggle is active!
#
# Revision 1.2  2011/04/25 18:27:33  gmmps
# Forked from 0.401.12 .
#
# Revision 1.29  2011/04/25 17:15:58  gmmps
# More small bugfixes/improvements (see 0.401.12 in changelog).
#
# Revision 1.28  2011/04/25 16:35:19  gmmps
# Bugfixes.
#
# Revision 1.27  2011/04/19 20:58:03  gmmps
# Bugfixes.
#
# Revision 1.26  2011/04/19 17:57:43  gmmps
# Changed the way the barcode is displayed.
#
# Revision 1.25  2011/04/18 21:36:36  gmmps
# Fixed N&S bug.
#
# Revision 1.24  2011/04/15 20:09:23  gmmps
# Now displays a barcode rectangle when plotting GMOS gaps.
#
# Revision 1.23  2011/04/14 22:30:12  gmmps
# Max Mode bugfix.
# N&S Bandfile bugfix.
# Added cancel button to N&S dialog.
#
# Revision 1.22  2011/04/08 21:38:51  gmmps
# Flamingos-2 ODF load bugfix.
#
# Revision 1.21  2011/04/08 20:13:24  gmmps
# N&S bugfixes.
#
# Revision 1.20  2011/04/07 21:27:42  gmmps
# Fixed catalog window title bug.
#
# Revision 1.19  2011/04/07 17:05:50  gmmps
# Band definition files are now associated with OT and ODF files.
#
# Revision 1.18  2011/04/06 22:28:34  gmmps
# Continued N&S tweaks.
#
# Revision 1.17  2011/03/31 19:37:32  gmmps
# N&S properties are now remembered by the config file. Bugfixes.
#
# Revision 1.16  2011/03/29 18:04:06  gmmps
# "plot bands" behavour tweak.
#
# Revision 1.15  2011/03/29 16:31:04  gmmps
# Bugfix on plot bands method.
# Moved gmmps.cfg to hidden file in user's home directory.
#
# Revision 1.14  2011/03/28 16:55:49  gmmps
# Added "DET_IMG" and "DET_SPEC" to generated ODF files.
# Slit conflict detection butfix in gmMakeMasks_2.cc.
#
# Revision 1.13  2011/03/25 23:30:21  gmmps
# *** empty log message ***
#
# Revision 1.12  2011/03/25 00:59:25  gmmps
# Changed colorscheme for OT and ODF catalog windows.
#
# Revision 1.11  2011/03/23 16:39:49  gmmps
# 'Plot GMOS -detname- Gaps' now consistant over multiple open windows.
# Grating/Filter/etc now hold thier values in the make masks window.
#
# Revision 1.10  2011/03/21 23:13:01  gmmps
# configuration file development continues.
#
# Revision 1.9  2011/03/21 14:37:44  gmmps
# More adjustments for new GMOS-N detector. Also added gmmps.cfg file
# and some associated code in vmAstroCat.tcl.
#
# Revision 1.8  2011/03/17 17:23:50  gmmps
# Generalized detector settings and fixed an install script bug.
#
# Revision 1.7  2011/03/04 18:05:46  gmmps
# Tweaked slittilt check.
#
# Revision 1.6  2011/03/04 17:34:56  gmmps
# New install procedure implemented, slittilt check implemented.
#
# Revision 1.5  2011/02/22 19:19:04  gmmps
# Minor changes.
#
# Revision 1.4  2011/01/27 16:53:23  gmmps
# F2 Waveccd adjustments.
#
# Revision 1.3  2011/01/26 22:16:00  gmmps
# Added more robustness with regard to waveccd. Shouldn't crash gmmps now.
#
# Revision 1.2  2011/01/24 20:17:57  gmmps
# Removed white line indicating "Center of Spectrum".
#
# Revision 1.1  2011/01/24 20:02:15  gmmps
# Compiled for RedHat 5.5 32 and 64 bit.
#
# Revision 1.38  2007/08/22 01:51:28  callen
# some bug fixes from James Turner again fixed while I was at
# scipy, I could not reproduce these errors but they appear related to
# some modifications for FLAMINGOS-2
#
# Revision 1.37  2007/06/26 06:51:51  callen
# patch given by Bryan Miller... should be and tests as innocuous... so I am
# patching it... though in reality I did not see the same bug in
# FC6 with the same data when plotting slits.
#
# Revision 1.36  2007/05/12 00:15:49  callen
# updates for a more robust set of build tools and specifically for
# FC6... also brought forward some FC6 fixes which were supposed to have
# been checked in.
#
# Revision 1.35  2007/05/09 02:29:35  callen
# FC6 mods
#
# Revision 1.34  2006/05/06 09:29:39  callen
# more changes for FLAMINGOS-2
#
# Revision 1.33  2006/05/04 19:38:11  callen
# changes
#
# Revision 1.32  2004/11/11 02:02:01  callen
# gelys supplied fix that shows slit tilts
#
# Revision 1.31  2003/02/27 02:22:24  callen
# fixed gmmps so that you can plot the object symbols for objects in
# the ODF when it is loaded from a FITS file.
#
# Revision 1.30  2003/02/20 03:01:26  callen
# files: band_def_UI.tcl vmAstroCat.tcl
#
# improved a faulty default and button names
#
# Revision 1.29  2003/02/06 06:31:00  callen
# added feature to grab SPEC_LEN from the #fits "preserved" headers in the .cat
# file when plotting slits... if the "Spectrum:" cat variable isn't present.
#
# Revision 1.28  2003/02/06 01:49:45  callen
# finished with the "plot bands"
#
# Revision 1.27  2003/02/05 23:20:15  callen
# added new "Convert and Load ODF from FITS" menuitem
#
# Revision 1.26  2003/02/05 01:21:03  callen
# changes to the GUI
#
# Revision 1.25  2003/01/18 08:04:05  callen
# added a line to pause, improving the linux bug, but not really solving it yet, although I've tracked it to the paste which attempts to join the new column
#
#
# S: ----------------------------------------------------------------------
#
# Revision 1.24  2003/01/18 07:24:28  callen
# removed extra info_dialog
#
# Revision 1.22  2003/01/18 06:34:27  callen
# more debug output for comparison between linux and solaris
#
# Revision 1.21  2003/01/18 06:28:41  callen
# debug output comparison linux vs solaris
#
# Revision 1.20  2003/01/18 06:09:55  callen
# debug output
#
# Revision 1.19  2003/01/18 06:06:59  callen
# more debug information for linux bug
#
# Revision 1.18  2003/01/17 13:38:02  callen
# these files changed tonight for these additions:
# NODSIZE/SHUFSIZE or BANDSIZE/SHUFSIZE into FITS ODF header
# slitwidth given via LabelEntry (not slider)
# save and load *.bands files for values in bandGUI
# clear bands:
# debugging output for linux bug
#
# Modified Files:
#  	band_def_UI.tcl gmConvert2Cat.c gmMakeMasks.cc gmmps_spoc.tcl
#  	vmAstroCat.tcl vmTableList.tcl
#
# Revision 1.17  2003/01/17 05:02:17  callen
# more of the validation and interface features from inger's testing
#
# Revision 1.16  2003/01/16 10:27:42  callen
# debugging puts for linux bug (!!! meant to work on regular list! but the next users are linux based and I run into the bug when I use my laptop/linux)  I did track it down quite a bit
#
# Revision 1.15  2003/01/06 11:31:55  callen
#
# i done from home... more bandUI
#
# Revision 1.14  2002/12/21 03:34:53  callen
# got past the roadblocks using vmTableList but, made some changes to text messages.
#
# Not yet done with band shuffling interface.  Checking things in for safety
# before vacation.
#
# Revision 1.13  2002/12/05 23:32:02  callen
# added slitlen to N&S screen, file, and makemasks: allows different slit
# and shuffle distances, and also ensures that the slitsize is accurate (since shufflePx is rounded to the next higher int, and converting that back to arcsec
# then gives a different number... now we don't convert, we use slitlen
# directly when writing the ODF.
#
# I also changed the "Band" to "N & S" to avoid confusion with frequency bands.
#
# Revision 1.12  2002/12/05 03:07:52  callen
# added another shuffle amount so it's represented in
# arcsec and pixels and converts.  Slow going as I learned a bit
# more tcl (and about the ESO LabelEntry widget)
#
# Revision 1.11  2002/11/21 19:43:28  callen
# stable point, band definition file name is passed through from
# band def --> gmmps_spoc --> gmMakeMask
#
# NOTE: MOST OF THESE FILES WERE NOT INVOLVED... I must have tweaked the
# dates with one experiment I did where I backed up the files and restored them
# outside CVS!
#
# Revision 1.10  2002/11/20 03:45:20  callen
# these changes bring up a nod and shuffle interface which can start the mask design interface.
#
# note: the feature is not yet active... it's just a GUI!
#
# Revision 1.9  2002/09/13 02:46:42  callen
# this allowed editing of the following columns
# slitpos_x
# slitpos_y
# slitsize_x
# slitsize_y
# tilt
#
# from within the vmAstroCat interface... next... saving the changes to FITS.
#
# Revision 1.8  2002/09/12 08:38:55  callen
# finished up most of the GUI side of the row editing for the input FITS table
# of objects.  need to add help and catch the flags in the _sel function.
# checking in as checkpoint, done for the night from home.
#
# Revision 1.7  2002/09/12 03:40:24  callen
# made changes to add more editing controls... these will be things like
# slitsize_x, _y, position, etc. from the task list.  This fix includes just
# slitsize_y as my test implementation.  It's working, I'm checking it in so I
# can do the other fields from home tonight (and to have a micromilestone for
# this feature)
#
# Revision 1.6  2002/08/22 01:41:01  callen
# had to remove some info_dialogs that were in for testing... for some reason
# they were not firing off in solaris but did when I tested in linux... definately
# not supposed to be there though.
#
# Revision 1.5  2002/08/20 00:49:17  callen
# added "-i" in grep spawning when looking for column names to work toward case
# insensitivity.  NOTE: other case sensitive checks seem spread out throughout the code
# and components.
#
# Revision 1.4  2002/08/13 23:14:12  callen
# thought I should add a "Clear GMOS Gaps" button, I also made the plot erase
# the previous marks (the gap graphics are taged as "gaps")
#
# Revision 1.3  2002/08/13 04:31:11  callen
# after further testing, found I was not converting properly on the GMOS
# gaps visualization... so I fixed this.  There is still a problem when loading another image
# in which case the marks are not erased (unless the user manually does it)
#
# Revision 1.2  2002/08/13 04:06:12  callen
# checking in changes to:
# vmAstroCat.tcl (to draw GMOS detector gaps)
# gmmps.sh (to allow passing an image argument to the gmmps script to load and image immediately)
#
# Revision 1.1.1.1  2002/07/19 00:02:09  callen
# importing gmmps as recieved from Jennifer Dunn
# gmmps is a skycat plugin and processes for creating masks
#
# Revision 1.23  2001/12/11 00:09:05  dunn
# Fixed checking for PIXSCALE keyword, yet again.
#
# Revision 1.22  2001/11/27 23:07:13  dunn
# Change of name
#
# Revision 1.21  2001/11/01 23:17:59  dunn
# Changes to handle different pixel scale.
#
# Revision 1.20  2001/10/22 17:38:41  dunn
# *** empty log message ***
#
# Revision 1.19  2001/10/22 17:35:46  dunn
# Fixed symbol's grep for fits keywords.
#
# Revision 1.18  2001/10/19 23:53:23  dunn
# Make pixel scale be passed in and around .
#
# Revision 1.17  2001/10/11 21:49:56  dunn
# Cleared debugLevel
#
# Revision 1.16  2001/10/11 21:04:43  dunn
# Added mag to require col list, Changed to get pix scale from file.
#
# Revision 1.15  2001/08/22 22:55:24  dunn
# *** empty log message ***
#
# Revision 1.14  2001/08/20 23:17:51  dunn
# Better messages.
#
# Revision 1.13  2001/08/20 18:52:41  dunn
# Changed it so that a user-friendly msg comes up when they try to
# open the Mask Design popup twice.
#
# Revision 1.12  2001/08/13 21:44:32  dunn
# *** empty log message ***
#
# Revision 1.11  2001/08/13 21:24:51  dunn
# Got rid of globals, and added them to the window instead.
#
# Revision 1.10  2001/08/03 21:07:32  dunn
# Changes to handle multiple windows and object selection.
#
# Revision 1.9  2001/08/02 20:07:43  dunn
# Fixes
#
# Revision 1.8  2001/07/17 19:40:11  dunn
# *** empty log message ***
#
# Revision 1.7  2001/06/26 21:45:17  dunn
# *** empty log message ***
#
# Revision 1.6  2001/05/11 18:06:59  dunn
# Added ability to change priority of user-define object.
#
# Revision 1.5  2001/05/04 22:48:58  dunn
# *** empty log message ***
#
# Revision 1.4  2001/04/25 22:25:24  dunn
# Got rid of some puts statements.
#
# Revision 1.3  2001/04/25 18:26:07  dunn
# Changed prior 2 to be green square.
#
# Revision 1.2  2001/04/25 17:01:22  dunn
# Initial revisions.
#
#
#****     D A O   I N S T R U M E N T A T I O N   G R O U P        *****
#***********************************************************************
#

itk::usual vmAstroCat {}

######################################################################
# The vmAstroCat widget class defines a top level widget for searching and
# displaying astronomical catalog data. It contains a menubar with items
# for loading, editing, and saving catalog data and a table displaying
# rows and columns of catalog data. This class does not know anything
# about images or plotting objects in images, however these features may
# be added in a derived class (see the SkySearch(n) in the skycat
# package for an example).
#
# You can also run this class as a standalone application with the
# command "astrocat". The application options are the same as the class
# options.
######################################################################

itcl::class cat::vmAstroCat {
    inherit util::TopLevelWidget

    #############################################################
    #  Name: constructor
    #
    #  Description:
    #   Constructor, what can I say.
    #############################################################
    #############################################################

    constructor {args} {
	#debugStatement "\n#vmAstroCat::constructor Entry.#"
	
    	set home $::env(GMMPS)
	
	#
	# This is where the debug level is set. 
	# (Unless there are more places where someone has decided to turn it off...)
	global acDebugLevel
	set acDebugLevel 0

	eval itk_initialize $args
	
	wm protocol $w_ WM_DELETE_WINDOW [code $this close ]
	
	#debugStatement "#vmAstroCat::constructor Exit.#\n"
	
    }

    
    #############################################################
    #  Name: destructor
    #
    #  Description:
    #   destructor - delete C++ based objects so that the temp 
    #   files are deleted
    #############################################################
    #############################################################

    destructor {
	#debugStatement "\n#vmAstroCat::destructor Entry.#"  	
    	
        catch {
            if {"[$w_.cat servtype]" == "local"} {
		if {"[string range $itk_option(-catalog) 0 4]" == "/tmp/"} {
		    file delete $itk_option(-catalog)
		}
            }
        }
        catch {$w_.cat delete}
        catch {close $rfd_}
        catch {close $wfd_}
        catch {unset instances_($instance_idx_)}
        if {"$current_instance_" == "$w_"} {
            set current_instance_ {}
        }
        
        # Clear all marks if this is the last catalog closed.
        if {[array size instances ] == 0} {
	    set target_canvas_ .skycat1.image.imagef.canvas
	    
	    $target_canvas_ delete gaps
	    $target_canvas_ delete f2
            $target_canvas_ delete boundary
            $target_canvas_ delete slitMark
	    $target_canvas_ delete
	    set plotGaps_ 0
	    
        }
    }

    
    #############################################################
    #  Name: init
    #
    #  Description:
    #   This function is called ??? after options have been evaluated
    #   It will : set the logfile
    #             do the window layout: which opens the catalog.
    #             enter widget name in instance array
    #             add a short help window
    #             create an object for running interruptable batch queries
    #             position from previous window, if pos.
    #		  for local catalogs, start search automatically
    #             Then ... something I don't understand.
    #############################################################
    #############################################################

    protected method init {} {
	#debugStatement "\n#vmAstroCat::init Entry.#"

	#debugStatement "init: catType = $itk_option(-catType)"
	#debugStatement "init: catalog = $itk_option(-catalog)"

	#CBA: not working; I'm following what seems a bad use of globals to initialize an object
        # because it's already followed for catType and I think it may make a 
        # bigger mess to screw with that in this code.  So, instrument type...
        
	set instType $itk_option(-instType)

	#debugStatement "vmAC::init: instType = $instType"

	#
	# Set a default central wavelength for each instrument.
	if {$instType == "GMOS-N" || $instType == "GMOS-S"} {
	    set itk_option(-bw) 600
	} elseif {$instType == "F2"} {
	    set itk_option(-bw) 1300
	}

	#
	# if running stand-alone, setup X defaults
	if {$itk_option(-standalone)} {
	    util::setXdefaults
	    cat::setXdefaults  
	}

	#
	# set the file used to keep track of URLs (for debugging)
	set_logfile

	#
	# do window layout.  This opens the catalog, and saves to table.
	layout_dialog

	#
	# enter widget name in instance array for later reference
	set_instance

	#
	# add a short help window
	make_short_help

	#
	# create an object for running interruptable batch queries
	Batch $w_.batch \
	    -command [code $this preview_done] \
	    -debug $itk_option(-debug)
        #
	# position from previous window, if pos.
	if {"$current_instance_" != ""} {
	    wm geometry $w_ +[winfo x $current_instance_]+[winfo y $current_instance_]
	}
	set current_instance_ $w_

	#
	# for local catalogs, start search automatically
	# Search will add the columns missing: x_ccd, y_ccd, MPS, ST
	set name $itk_option(-catalog)
	if {"[$w_.cat servtype]" == "local"} {
	    #
	    #  Set the title on the window to be the name listed
	    #  after the longname in the file.  And icon is the shortname
	    #debugStatement "vmAstroCat: $name ([$w_.cat longname $name])"
	    wm title $w_ "[file tail $name] ($itk_option(-number))"
	    wm iconname $w_ "[file tail [$w_.cat shortname $name]]"
	    search
	}

	set initialized_ 1

	#
	#  Appears to check if there is an priority column in something?, 
	#  and if so then it must be an object table. 
	#  If it has an slitpos_x column, then must be a 
	#  mask ODF file, otherwise must be an objects table.
	#  It brings up 2 different dialog boxes.
	set catType $itk_option(-catType)		
	#debugStatement "init: about to check for columns... "
	#debugStatement "init: catType = $catType ... "
	#puts "init: later, catType = $catType"
	if { $catType == 1 || $catType == 3 } {
	    # Change window color for OT.
	    $w_ configure -background $ot_bg_
	    
	    mps_add_dialog_buttons $instType
	} elseif { $catType == 2 || $catType == 4} {
	    # ODF loading! 2 is .cat file, and 4 is convert from FITS first
	    
	    # Change window color for ODF.
	    $w_ configure -background $odf_bg_
	    
	    slit_add_dialog_buttons $instType
	    # and look for 
	} else {
	    error_dialog "Error, Invalid pulldown type" 
	}
	
	# Change shorthelp colors.
	$w_.shelp configure -background $ot_bg2_ -foreground $ot_text_	
	
       	#debugStatement "#vmAstroCat::init Exit.\n#"
    } 


    #############################################################
    #  Name: debugStatement
    #
    #  Description:
    #   Pop up a window so that the user can edit the selected object(s)
    #############################################################
    #############################################################

    protected proc debugStatement {line} {
	global acDebugLevel
	
	if { $acDebugLevel == 1 } {
	    puts $line
	}
    }


    #############################################################
    #  Name: set_logfile
    #
    #  Description:
    #   Create the ~/.skycat dir if it does not already exists and
    #   keep a log file there.  Funny, not used in the file.
    #############################################################
    #############################################################

    protected method set_logfile {} {
	#debugStatement "\n#vmAstroCat::set_logfile Entry.#"

	global ::env

	# open log file used to keep track of URLs (for debugging)
	set dir $env(HOME)/.skycat

	if {! [file isdirectory $dir]} {
	    catch {mkdir $dir}
	}
	set logfile_name_ $dir/log

	#debugStatement "#vmAstroCat::set_logfile Exit.#\n"
    }


    #############################################################
    #  Name: readInFitsFile
    #
    #  Description:
    #   Convert a fits file to a catalog file format.
    #############################################################
    #############################################################

    protected proc readInFitsFile {fitsFile args } {
	#debugStatement "\n#vmAstroCat::readInFitsFile Entry.#"

	set result " " 
	set buff " "

	
	# args is tType... 4 is a ODF.fits file
	set convertProgram gmConvert2Cat

	#debugStatement "\nreadInFitsFile: Entry. fitsFile=$fitsFile"
	#puts "readInFitsFile: about to exec: $convertProgram $fitsFile"
	#debugStatement "readInFitsFile: using $convertProgram to convert type $args"

	if {[catch {set buff [exec $convertProgram $fitsFile]} msg]} {
	    error_dialog "$msg." 
	    return -1
	}

	return 0
	#debugStatement "#vmAstroCat::readInFitsFile Exit.\n#"
    }

    
    #############################################################
    #  Name: set_feedback
    #
    #  Description:
    #   open or close a pipe to get feedback during HTTP transfers.
    #   (To save limited fds, we close the feedback pipe after each 
    #   HTTP op).  The arg should be "on" to turn feedback on, or
    #   "off" to turn it off.
    #############################################################
    #############################################################

    protected method set_feedback {onoff} {
	#debugStatement "\n#vmAstroCat::set_feedback Entry.#"

	# Process any pending file events.
	# Note: this is important: if we don't process the events before
	# closing the feedback file, a crash may result.
	update 

	if {$itk_option(-debug)} {
	    # if -debug was given, the feedback bit wont work, since the
	    # query is done in the foreground and trying to read the feedback
	    # from the pipe may cause the application to hang...
	    return
	}
	if {"$onoff" == "on"} {
	    lassign [pipe] rfd_ wfd_
	    fileevent $rfd_ readable [code $this feedback]
	    $w_.cat feedback $wfd_
	} elseif {[info exists rfd_]} {
	    ::close $rfd_
	    ::close $wfd_
	    unset rfd_ wfd_
	    $w_.cat feedback {}
	}
        
	#debugStatement "#vmAstroCat::set_feedback Exit.\n#"
    }

    
    #############################################################
    #  Name: set_instance
    #
    #  Description:
    #   Keep an array of instances(name,id) to help locate the 
    #   window for a catalog
    #############################################################
    #############################################################
    
    protected method set_instance {} {
	#debugStatement "\n#vmAstroCat::set_instance Entry.#"

	set name [$w_.cat longname $itk_option(-catalog)]
	set id $itk_option(-id)
	set instance_idx_ "$name,$id"
	set instances_($instance_idx_) $w_ 

	#debugStatement "#vmAstroCat::set_instance Exit.\n#"
    }


    #############################################################
    #  Name: add_menubar
    #
    #  Description:
    #    Add the menu bar in the object table window.
    #############################################################
    #############################################################

    protected method add_menubar {} {
	#debugStatement "\n#vmAstroCat::add_menubar Entry.#"
	TopLevelWidget::add_menubar
	
	set home $::env(GMMPS)
	set m [add_menubutton File "Display File menu"]
	set file_menu_ $m

	if {$itk_option(-catType) == 1 || $itk_option(-catType) == 3} {
	    $m configure \
		-foreground $menu_text_ot_ \
		-background $menu_bg_ot_ 			
	} else {
	    $m configure \
		-foreground $menu_text_odf_ \
		-background $menu_bg_odf_ 
	}
	
	if {$iscat_} {
	    add_menuitem $m command "Open" \
		{Open OT local catalog in tab table format} \
		-command [code cat::vmAstroCat::local_catalog \
			      $itk_option(-id) \
			      [info class] \
			      $itk_option(-debug) 1 $w_] \
		-accelerator "Control-o"
	    
	    $m add separator
	    
	    add_menuitem $m command "Save as..." \
		{Save listed objects to a local catalog file} \
		-command [code $this save_as] \
		-accelerator "Control-s"
	    
	    add_menuitem $m command "Add to..." \
		{Add listed objects to a local catalog file} \
		-command [code $this add_to] \
		-accelerator "Control-a"
	    
	    add_menuitem $m command "Add selected..." \
		{Add selected rows to a local catalog file} \
		-command [code $this add_selected] \
		-accelerator "Control-A"
	    
	    $m add separator
	    
	    add_menuitem $m command "Print..." \
		{Print the listing to a printer or file} \
		-command [code $this print] \
		-accelerator "Control-p"
	    
	    add_menuitem $m command "Clear" \
		{Clear the catalog listing} \
		-command [code $this clear]
	    
	    $m add separator
	}
	
	add_menuitem $m command "Close" \
	    {Close this window} \
	    -command [code $this close]
	
	if {$itk_option(-standalone)} {
	    add_menuitem $m command "Exit" \
		{Exit the application} \
		-command [code $this quit] \
		-accelerator "Control-q"
	}
	
	if {$iscat_} {
	    # Edit menu
	    set m [add_menubutton Edit "Display Edit menu"]
	    set edit_menu_ $m
	    
	    if {$itk_option(-catType) == 1 || $itk_option(-catType) == 3} {
		$m configure \
		    -foreground $menu_text_ot_ \
		    -background $menu_bg_ot_ 			
	    } else {
		$m configure \
		    -foreground $menu_text_odf_ \
		    -background $menu_bg_odf_ 
	    }
	    
	    add_menuitem $m command "Clear Slits" \
		{Clear slits from the GMOS display} \
		-command [code $this clearslits]
	    
	    add_menuitem $m command "Remove selected" \
		{Remove selected rows from the local catalog} \
		-command [code $this remove_selected] \
		-state disabled
	    
	    add_menuitem $m command "Enter new object..." \
		{Enter the data for a new object for the local catalog} \
		-command [code $this enter_new_object] \
		-state disabled
	    
	    add_menuitem $m command "Edit selected object..." \
		{Edit the data for the selected object in the local catalog} \
		-command [code $this edit_selected_object] \
		-state disabled
	}
	
	# Options menu
	set m [add_menubutton Options "Display Options menu"]
	set options_menu_ $m
	add_menuitem $m cascade "Set Name Server" \
	    {Select the name server used to resolve astronomical object names} \
	    -menu [set ns_menu_ [menu $itk_component(options).m.ns]]
	
	if {$itk_option(-catType) == 1 || $itk_option(-catType) == 3} {
	    $m configure \
		-foreground $menu_text_ot_ \
		-background $menu_bg_ot_ 			
	} else {
	    $m configure \
		-foreground $menu_text_odf_ \
		-background $menu_bg_odf_ 
	}
	
	if {$iscat_} {
	    add_menuitem $m command "Set Sort Columns..." \
		{Set options for sorting the query results} \
		-command [code $this sort_dialog] \
		-state disabled
	    
	    add_menuitem $m command "Hide/Show Columns..." \
		{Set options for displaying columns of the query results} \
		-command [code $this select_columns] \
		-state disabled
	    
	    add_menuitem $m command "Set Search Columns..." \
		{Set the list of columns to search by for this catalog} \
		-command [code $this set_search_cols] \
		-state disabled
	    
	    add_menuitem $m command "Set Plot Symbols..." \
		{Set the symbol (color, size, etc.) to use to plot objects} \
		-command [code my_set_plot_symbols] \
		-state disabled
	    
	    $m add separator
	    
	    #puts "child m: [pack slaves $m ]"
	    
	    # Open detector config file. 
	    set detFile [open "$home/config/Detectors.dat" r ]
	    
	    add_menuitem $m command "Select $itk_option(-instType) Detector" {Select Detector} \
		-state disabled 
	    
	    set imageDetId [string trim [get_detid ]]
	    set curDetId ""	
	    
	    if {[array size instances_ ] != 1} {
		# If there is another open window find out which detector is selected.
		
		set curDetId [lindex $detector_ 2]
	    }
	    
	    set currentDetMode 0
	    set curIndex ""
	    set cnt 0
	    while {[gets $detFile line] >= 0} {
		
		if {$line == ""} {
		    continue
		} elseif {[string index $line 0 ] == "#" } {
		    continue
		}
		
		set line [split $line "\t" ]
		
		# Extract instrument type from detector property list. 
		# List order: (more info in config/Detectors.dat)
		# Currently-Installed Instrument-Name Detector-Name Detector-Id Gap1 Gap2 GapL Pixel-Scale
		set inst [string trim [lindex $line 1 ]]

		if {$inst == $itk_option(-instType)} {
		    # If this is a detector for our current instrument, add it to 
		    # the detector list.
		    
		    incr cnt	
		    
		    set detName [string trim [lindex $line 3 ]]
		    set detId [string trim [lindex $line 2 ]]
		    set installed [string trim [lindex $line 0 ]]
		    
		    if {$detId == $curDetId} {
			set curIndex [expr $cnt + 6 ]
		    }				
		    
		    if {$installed} {
			# If this is the currently installed detector for our instrument,
			# then select it. 
			
			# Preimage loaded was taken using the current detector. Go
			# into current detector mode.
			if {$imageDetId == $detId} {
			    set currentDetMode 1
			}
			
			set installedIndex [expr $cnt + 6 ]
			
			$m add radiobutton \
			    -label "\"$detName\" : $detId" \
			    -state active \
			    -selectcolor $detRadioColor_ \
			    -command [code $this set_detector $line ]
			
		    } else {
			$m add radiobutton \
			    -label "\"$detName\" : $detId" \
			    -selectcolor $detRadioColor_ \
			    -command [code $this set_detector $line ]
		    }
		}
		# End Detector input loop.
	    }
	    
	    # If another instance has already set a detector, use that one,
	    # other wise default to installed detector.
	    if {$curIndex != ""} {
		set activeIndex $curIndex
	    } else {
		set activeIndex $installedIndex
	    }
	    
	    # Select menuitem in widget.
	    $m invoke $activeIndex
	    
	    # Current detector mode, do not allow other detectors to be 
	    # selected.
	    if {$currentDetMode} {
		set num $cnt 
		set cnt 0
		while {$cnt <= $num} {
		    set curIndex [expr $cnt + 6 ]
		    if {$curIndex != $activeIndex} {
			$m entryconfigure $curIndex -state disabled
		    }
		    incr cnt
		}
	    }
	    
	    ::close $detFile
	}
	
	#	add_menuitem $m command "Proxies..."  \
	    #	    "Define an HTTP proxy server for use with a firewall." \
	    #	    -command [code cat::vmAstroCat::proxies]
	#	check_proxies

	# add a menu of catalogs ("Data-Servers")
	#add_catalog_menu $w_ $itk_option(-id) [info class] $itk_option(-debug)

	#debugStatement "#vmAstroCat::add_menubar Exit.\n#"
    }


    #############################################################
    #  Name: set_detector
    #
    #  Description:
    #   Sets the current instrument detector being used by gmmps. 
    #############################################################
    #############################################################


    public method set_detector {detInfo} {
	
	set detector_ $detInfo
	set detName [lindex $detInfo 3 ]
	set instType $itk_option(-instType)

	if {$plotGaps_} {
	    # If we are plotting gaps, then draw the new gaps.
	    
	    set target_canvas_ .skycat1.image.imagef.canvas
	    set plotGaps_ 0
	    
	    if {$instType != "F2"} {
		$target_canvas_ delete gaps		
		catch {plotGmosGaps}
	    } else { 
		$target_canvas_ delete F2
		catch {plotF2Detector}
	    }
	    
	} else {
	    # If we aren't plotting we still need to change the name. 
	    
	    # Change gap plotting button name, when the button exists.
	    if {$instType != "F2"} {
		change_gap_button_names "Plot GMOS $detName Detector Gaps "
	    } else {
		change_gap_button_names "Plot FLAMINGOS-2 Field Layout "
	    }
	}
    }        


    #############################################################
    #  Name: change_gap_button_names
    #
    #  Description:
    #   Change the name of the detector gap button name for all open catalog
    #	widgets. 
    #############################################################
    #############################################################

    protected method change_gap_button_names {newname} {
	set names [array names instances_ ]

	foreach name $names {
	    set instance $instances_($name)
	    
	    #debugStatement "change gap button names: $instance.plotGmosGaps"
	    
	    catch {$instance.plotGmosGaps configure -text $newname}
	}
    }


    #############################################################
    #  Name: set_menu_states
    #
    #  Description:
    #   Enable or disable some menus
    #############################################################
    #############################################################
    
    protected method set_menu_states {} {
	#debugStatement "\n#vmAstroCat::set_menu_states Entry.#"
	


	$options_menu_ entryconfig "Set Sort Columns..." -state normal
	$options_menu_ entryconfig "Hide/Show Columns..." -state normal

	if {[$w_.cat iswcs] || [$w_.cat ispix]} {
	    set state normal
	} else {
	    set state disabled
	}
	$options_menu_ entryconfig "Set Plot Symbols..." -state $state
	$options_menu_ entryconfig "Set Name Server" -state $state

	# determine states for some menu items
	if {"[$w_.cat servtype]" == "local"} {
	    set state [set sstate normal]
	} else {
	    set state [set sstate disabled]
	    # allow user to edit search cols if URL contains "%cond"
	    if {[string first "%cond" [$w_.cat url]] >= 0} {
		set sstate normal
	    } 
	}
	$edit_menu_ entryconfig "Remove selected" -state $state
	$edit_menu_ entryconfig "Enter new object..." -state $state
	$edit_menu_ entryconfig "Edit selected object..." -state $state

	$options_menu_ entryconfig "Set Search Columns..." -state $sstate

	#debugStatement "#vmAstroCat::set_menu_states Exit.\n#"
    }
    
    
    #############################################################
    #  Name: add_catalog_menu
    #
    #  Description:
    #   This function adds a menubutton with catalog items to 
    #   the given TopLevelWidget.  This is specifically for the
    #   popup windows, NOT for the top level Skycat window (see
    #   gmmps_menu for that.)
    #
    #  Args:
    #   w is the caller's toplevel window (an instance of TopLevelWidget).
    #
    #   id is an optional unique id to be associated with a new catalog 
    #   widget.
    #
    #   classname is the name of the vmAstroCat subclass to use to 
    #   create new catalog widgets (defaults to "vmAstroCat").
    #
    #   debug is a flag is passed from the command line. If true, 
    #   querries are run in the foreground, to make debugging easier.
    #
    #   note: this is defined as a proc rather than a method so that it 
    #   can be accessed from outside the class to add the Data-Servers 
    #   menu
    #############################################################
    #############################################################


    public proc add_catalog_menu {w {id ""} {classname vmAstroCat} {debug 0}} {
	#debugStatement "\n#vmAstroCat::add_catalog_menu Entry.#"

	# save this info so we can update the menu automatically later
	set catalog_menu_info_($w) \
	    [code cat::vmAstroCat::add_catalog_menu $w $id $classname $debug]

	# add the menu to the caller's menubar
	set m [$w add_menubutton "GMOS-MMS" {GMOS Mask Making Software Plugin}]
	
	$w add_menuitem $m cascade "Load a catalog" \
	    {Select a catalog from the menu} \
	    -menu [menu $m.local]

	#	fill_catalog_menu $w $id $classname $m.local $debug local shortname

	$m.local add separator

	$w add_menuitem $m.local command "Load an OT Catalog (*.cat)..." \
	    {Open an OT local catalog} \
	    -command [code cat::vmAstroCat::local_catalog $id $classname $debug  1 $w] \
	    -accelerator "Control-O"
	################3
	#### JENNIFER, ADD other MMS Menu options here....
	####      this appears on the small catalog screen.
	################

	############## REMOVE this because we can't handle browsing directories.
	#	$m add separator
	#
	#	$w add_menuitem $m command "Browse Catalog Directories..."  \
	    #	    "Browse the catalog directory hierarchy to view \
	    #             catalogs or add them to the default list" \
	    #	    -command [code cat::vmAstroCat::catalog_directory $id $classname $debug $w] \
	    #	    -accelerator "Control-b"
	#
	#	$w add_menuitem $m command "Reload config file..."  \
	    #	    "Reload the default catalog config file after it was edited by hand" \
	    #	    -command [code cat::vmAstroCat::reload_config_file $w] \
	    #		-accelerator "Control-R"
	#	
	#	# if there is a local catalog called "history", add it to the menu also
	#	if {"[$astrocat_ servtype history]" == "local"} {
	#	    $m add separator
	#	    $w add_menuitem $m command "History..."  \
	    #		"Open an automatically generated catalog of previously viewed images" \
	    #		-command [code cat::vmAstroCat::select_catalog history local $id $classname $debug $w] \
	    #		-accelerator "Control-h"
	#	}#######################################

	#debugStatement "#vmAstroCat::add_catalog_menu Exit.\n#"
    }

    
    #############################################################
    #  Name: reload_config_file
    #
    #  Description:
    #   Reload the default skycat catalog config file using the given 
    #   astrocat object
    #############################################################
    #############################################################

    public proc reload_config_file {w} {
	#debugStatement "\n#vmAstroCat::reload_config_file Entry.#"

	global ::env
	set config_file $env(HOME)/.skycat/skycat.cfg
	if {[file exists $config_file]} {
	    set env(CATLIB_CONFIG) "file:$config_file"
	}

	if {[catch {$astrocat_ reload} msg]} {
	    error_dialog $msg $w
	}
	# make sure these windows are updated
	catch {destroy $w_.symconf}
	catch {destroy $w_.searchconf}
	if {[winfo exists .catinf]} {
	    .catinf reinit_tree
	}
	
	if {[info exists instances_]} {
	    foreach i [array names instances_] {
		$instances_($i) update_search_options
	    }
	}
	update_catalog_menus

	#debugStatement "#vmAstroCat::reload_config_file Exit.\n#"
    }


    #############################################################
    #  Name: check_proxies
    #
    #  Description:
    #    Check for a file ~/.skycat/proxies, once each session, and
    #    use it to initialize environment variables for a proxy server
    #    (see also tclutil/util/src/HTTP.C).
    #############################################################
    #############################################################

    public proc check_proxies {} {
	#debugStatement "\n#vmAstroCat::check_proxies Entry.#"

	global ::env
	# only do it once
	if {$checked_proxies_} {
	    return 
	}
	cat::ProxyDialog::check_proxies $env(HOME)/.skycat/proxies
	set checked_proxies_ 1
    }

    # Pop up a dialog to set or change the HTTP proxy server.

    public proc proxies {} {
	global ::env
	utilReUseWidget ProxyDialog .proxy \
	    -configfile $env(HOME)/.skycat/proxies

	#debugStatement "#vmAstroCat::check_proxies Exit.\n#"
    }

    
    #############################################################
    #  Name: catalog_directory
    #
    #  Description:
    #   This function pops up a window to browse the catalog 
    #   directories. 
    #
    #  Args:
    #   id is an optional unique id to be associated with a new 
    #   catalog widget.
    #
    #   classname is the name of the vmAstroCat subclass to use to 
    #   create new catalog widgets (defaults to "vmAstroCat").
    #
    #   debug is a flag is passed from the command line. If true, 
    #   querries are run in the foreground, to make debugging easier.
    #
    #   w should be the caller's top level window, if specified.
    #
    #############################################################
    #############################################################

    proc catalog_directory {{id ""} {classname vmAstroCat} {debug 0} {w ""}} {
	#debugStatement "\n#vmAstroCat::catalog_directory Entry.#"

	utilReUseWidget cat::CatalogInfo .catinf \
	    -id $id \
	    -classname $classname \
	    -debug $debug \
	    -callerw $w \
	    -command [code cat::vmAstroCat::update_catalog_menus]
	
	# put the window clone number in the title
	if {[winfo exists $w]} {
	    catch {wm title .catinf "Catalog Directory ([$w cget -number])"}
	}

	#debugStatement "#vmAstroCat::catalog_directory Exit.\n#"
    }

    
    #############################################################
    #  Name: update_catalog_menus
    #
    #  Description:
    #   Update all of the catalog menus in all instances to show 
    #   the current catalog info. 
    #############################################################
    #############################################################

    public proc update_catalog_menus {} {
	#debugStatement "\n#vmAstroCat::update_catalog_menus Entry.#"

	foreach i [array names catalog_menu_info_] {
	    if {[winfo exists [utilNamespaceTail $i]]} {
		eval $catalog_menu_info_($i)
	    }
	}

	#debugStatement "#vmAstroCat::update_catalog_menus Exit.\n#"
    }
    

    #############################################################
    #  Name: my_set_plot_symbols
    #
    #  Description:
    #   Pop up a dialog to set the plot symbols to use for this
    #   catalog.
    #############################################################
    #############################################################

    public method my_set_plot_symbols {} {
	#debugStatement "\n#vmAstroCat::set_plot_symbols Entry.#"

	set columns $itk_option(-headingx)
	#debugStatement "vmAC: set_plot_symbols: $columns"
	if {[llength $columns] == 0} {
	    info_dialog "Please make a query first so that the column names are known" $w_
	    return
	}
	utilReUseWidget cat::SymbolConfig $w_.symconf \
	    -catalog $itk_option(-catalog) \
	    -astrocat [code $w_.cat] \
	    -columns $columns \
	    -command [code $this plot]

	#debugStatement "#vmAstroCat::set_plot_symbols Exit.\n#"
    }


    #############################################################
    #  Name: set_search_cols
    #
    #  Description:
    #   Pop up a dialog to set the search columns to use for 
    #   this catalog
    #############################################################
    #############################################################

    public  method set_search_cols {} {
	#debugStatement "\n#vmAstroCat::set_search_cols Entry.#"

	set columns $itk_option(-headingx)
	if {[llength $columns] == 0} {
	    info_dialog "Please make a query first so that the column names are known" $w_
	    return
	}
	utilReUseWidget cat::SearchConfig $w_.searchconf \
	    -catalog $itk_option(-catalog) \
	    -command [code $this update_search_options] \
	    -astrocat [code $w_.cat] \
	    -columns $columns

	#debugStatement "#vmAstroCat::set_search_cols Exit.\n#"
    }


    #############################################################
    #  Name: update_search_options
    #
    #  Description:
    #    Update the search option entries after they have been edited
    #############################################################
    #############################################################

    public method update_search_options {} {
	$searchopts_ update_search_options
    }


    #############################################################
    #  Name: close
    #
    #  Description:
    #    Close this window
    #############################################################
    #############################################################

    public method close {} {
	#debugStatement "close: entry...."

	catch {$this delete_objects }
	
	if {$itk_option(-catType) == 2 || $itk_option(-catType) == 4} {

	    if {$plotSlits_} {
		catch {clearslits }
	    }
	    
	    if {$plotBands_} {
		catch {.skycat1.image.imagef.canvas delete bandrects}
	    }
	}
	
        destroy $w_ 
    }

    
    #############################################################
    #  Name: sort_dialog
    #
    #  Description:
    #   Pop up a dialog to sort the list
    #############################################################
    #############################################################

    public method sort_dialog {} {
	$results_ sort_dialog
    }

    
    
    #############################################################
    #  Name: set_sort_cols
    #
    #  Description:
    #   This function is called when the user has selected columns 
    #   to sort the results by.
    #   The first arg is the sort columns, the second arg is the 
    #   order (increasing, decreasing)
    #############################################################
    #############################################################
    
    public method set_sort_cols {sort_cols sort_order} {
	if {"[$w_.cat sortcols]" != "$sort_cols" \
		|| "[$w_.cat sortorder]" != "$sort_order"} {
	    $w_.cat sortcols $sort_cols
	    $w_.cat sortorder $sort_order
	    cat::CatalogInfo::save {} $w_ 0
	    $results_ config -sort_cols $sort_cols -sort_order $sort_order
	    search
	}
    }


    #############################################################
    #  Name: select_columns
    #
    #  Description:
    #   Pop up a dialog to select table columns to display
    #############################################################
    #############################################################

    public method select_columns {} {
	$results_ select_columns
    }


    #############################################################
    #  Name: set_show_cols
    #
    #  Description:
    #   Called when the user has selected columns to show
    #############################################################
    #############################################################
    
    public method set_show_cols {cols} {
	set show [$w_.cat showcols]
	if {"$show" == ""} {
	    set show [$results_ get_headings]
	}
    }


    #############################################################
    #  Name: get_name_servers
    #
    #  Description:
    #   Add the name server catalogs to the name server menu
    #############################################################
    #############################################################

    protected method get_name_servers {} {
	set m $ns_menu_
	if {[catch {set list [$w_.cat info namesvr]} msg]} {
	    error_dialog $msg $w_
	    return
	}

	foreach namesvr $list {
	    $m add radiobutton \
		-label $namesvr \
		-command [code $this set_namesvr $namesvr] \
		-variable $m
	}

	if {![info exists namesvr]} {
	    error_dialog "No default name server found for astronomical objects"
	    return
	}

	# set default name server
	global ::$m
	set $m $namesvr
	set_namesvr $namesvr
    }


    #############################################################
    #  Name: reset_table
    #
    #  Description:
    #    Reset table dialogs if needed
    #############################################################
    #############################################################

    public method reset_table {} {
	#debugStatement "reset_table: entry...."
	$results_ reset
	$results_ set_options {MORE PREVIEW more preview} Show 0
	$w_.cat showcols {}
	$w_.cat sortcols {}
	cat::CatalogInfo::save {} $w_ 0
    }

    
    
    #############################################################
    #  Name: clear
    #
    #  Description:
    #   Clear the table listing
    #############################################################
    #############################################################
    
    public method clear {} {
	catch {$results_ clear}
    }

    
    #############################################################
    #  Name: print
    #
    #  Description:
    #   Pop up a dialog to print the table listing
    #############################################################
    #############################################################
    
    public method print {} {
	$results_ print_dialog
    }

    
    #############################################################
    #  Name: local_catalog
    #
    #  Description:
    #   Called from menu item "Load from file..." to load in a catalog.
    #   This function asks the user for the name of a local 
    #   catalog file and then opens a window for the catalog.
    #
    #   id is an optional unique id to be associated with a new
    #   catalog widget.
    #
    #   classname is the name of the vmAstroCat subclass to use to 
    #   create new catalog widgets (defaults to "vmAstroCat").
    #
    #   $debug is a flag set from the command line arg.
    #
    #   $tType is the catalog type, Object Table, Master OT.
    #
    #   $w should be the top level window of the caller, 
    #   if specified.
    #############################################################
    #############################################################

    
    public proc local_catalog {{id ""} {classname vmAstroCat} {debug 0} tType {w ""} {instType "NONE"}} {

	# Debuglevel is being set here because this is a procedure.
	global acDebugLevel
	set acDebugLevel 0

	set loadedImage [$target_image_ cget -file ]
	
	if {$loadedImage == ""} {
	    error_dialog "Please load a preimage file into Skycat before loading an object catalog."
	    return
	}
	
	# Set file pattern based on catalog type. 
	if { $tType == 1 } {
	    # .cat OT
	    set filePattern *.cat
	} elseif { $tType == 3 } {
	    # .fits OT	
	    set filePattern *.fits
	} elseif { $tType == 2 } {
	    # .cat ODF
	    set filePattern *ODF*.cat
	} elseif { $tType == 4 } {
	    # .fits ODF
	    set filePattern *ODF*.fits
	}
	
	# Prompt user to choose file to open.
	set file [filename_dialog [pwd ] $filePattern]
	if {"$file" != ""} {
	    if {[file isfile $file]} {
		if {[catch {$astrocat_ check $file} msg]} {
		    error_dialog $msg
		} else {
		    select_catalog $file local $id $classname $debug $tType $w $instType
		}
	    } else {
		error_dialog "There is no file named '$file'"
	    }
	}
	
        if {$tType == 1} {
            # OT Catalog loaded in .cat format.
            
            puts "Loading object table (OT) catalog: $file"       
            
        } elseif {$tType == 2} {
            # ODF Catalog loaded in .cat format.
            
            puts "Loading minimal object definition file (ODF) catalog: $file"
            
        } elseif {$tType == 3} {
            # OT Catalog loaded from .fits format.
            
            puts "Loading object table (OT) catalog: $file"
            puts "(.fits format table will be converted to .cat for display.)"
	    
        } elseif {$tType == 4} {
            # ODF Catalog loaded from .fits format. 
            
            puts "Loading minimal object definition file (ODF) catalog: $file"
            puts "(.fits format table will be converted to .cat for display.)"
        }
    }
    
    #############################################################
    #  Name: fill_catalog_menu
    #
    #  Description:
    #   This function fills up a menu full of catalog files.  If one
    #   is choosen, then select_catalog is called which will look
    #   for manditory columns, open the file, and save an instance of it.
    #   Called when the main menu is set up & later.
    # 
    #  Args:
    #   w is the caller's top level widget (containing menubar)
    #
    #   id is an optional unique id to be associated with a new 
    #   catalog widget.
    #
    #   classname is the name of the vmAstroCat subclass to use to 
    #   create new catalog widgets (defaults to "vmAstroCat").
    #
    #   $debug is a flag set from the command line arg.
    #
    #   The serv_type argument should be one of: "catalog", "archive".
    #
    #   An optional "cmd" argument may be set to longname or shortname
    #   depending on the names you want in the menu (for local catalogs,
    #   specify shortname, to avoid getting the complete path name in
    #   the menu.
    #############################################################
    #############################################################


    public proc fill_catalog_menu {w id classname m debug tType serv_type {cmd longname}} {

	#
	#  This appears to have bugs, so for now, lets just not list past
	#  files, thank you.
	return

	#
	#  Create a catalog list to choose from
	#

	if {[catch {set catalog_list [lsort [$astrocat_ info $serv_type]]} msg]} {
	    error_dialog $msg $w
	    return
	}

	#
	#  If there is anything in the list then... if one is selected
	#  call select_catalog for it.
	if {[llength $catalog_list]} {
	    #
	    #  For each item in the list
	    foreach i $catalog_list {
		set name [$astrocat_ $cmd $i]
		if { $name == "history" } { 
		    continue
		}
		$w add_menuitem $m command $name \
		    "Open $serv_type: \"$i\"" \
		    -command [code cat::vmAstroCat::select_catalog $i $serv_type $id \
				  $classname $tType $debug $w]
	    }
	}
    }

    
    #############################################################
    #  Name: open_catalog_window
    #
    #  Description:
    #   This function opens a window for the given catalog, or 
    #   reports an error if the catalog does not exist. 
    #      $name is the name of the catalog to open,
    #      $classname is the class to use to open it,
    #      $debug is an optional debugging flag,
    #      $w should be set to the top level window of the caller (optional).
    #############################################################
    #############################################################
    ####### NOT CALLED BY ANYONE????,  wait there are lots in this file
    ########## that are not called by anyone.  Note to self: get rid of them.

    public proc open_catalog_window {name {id ""} {classname vmAstroCat} {debug 0} {w ""}} {

	if {[catch {$astrocat_ open $name} msg]} {
	    error_dialog $msg
	    return
	}

	cat::vmAstroCat::select_catalog $name catalog $id $classname $debug $w
    }


    #############################################################
    #  Name: save_as
    #
    #  Description:
    #   Save the current data to a local catalog
    #############################################################
    #############################################################
    
    public method save_as {} {
	$results_ save_as
    }


    #############################################################
    #  Name: add_to
    #
    #  Description:
    #   Add the rows in the current listing to a local catalog file
    #############################################################
    #############################################################
    
    public method add_to {} {
	$results_ add_to
    }

    
    #############################################################
    #  Name: add_selected
    #
    #  Description:
    #   Add the currently selected rows to a local catalog file
    #############################################################
    #############################################################
    
    public method add_selected {} {
	$results_ add_selected
    }

    
    #############################################################
    #  Name: remove_selected
    #
    #  Description:
    #   Remove the currently selected rows from a local catalog file
    #############################################################
    #############################################################

    public method remove_selected {} {
	$results_ remove_selected

	# update the display
	clear
	search
    }

    
    #############################################################
    #  Name: enter_new_object
    #
    #  Description:
    #   Pop up a dialog to enter the data for a new object for a local catalog
    #############################################################
    #############################################################

    public method enter_new_object {} {
	$results_ enter_new_object [code $this search]
    }

    
    #############################################################
    #  Name: edit_selected_object
    #
    #  Description:
    #   Pop up a window so that the user can edit the selected object(s)
    #############################################################
    #############################################################

    public method edit_selected_object {} {
	$results_ edit_selected_object [code $this search]
    }

    #############################################################
    #  Name: tilt_selected_object
    #
    #  Description:
    #   Pop up a window so that the user can edit the selected object(s)
    #   angles.
    #############################################################
    #############################################################

    public method tilt_selected_object {} {
	$results_ tilt_selected_object [code $this search]
    }


    #############################################################
    #  Name: drawBoundary
    #
    #  Description:
    #   This function draws the boundary for the fov, 
    #   but first reads in the fov information.
    #############################################################
    #############################################################

    public method drawBoundary { pixScale instType} {

	set bg black
	set fg red
	set tags boundary
	set target_canvas_ .skycat1.image.imagef.canvas
	set target_image_ image2
	set catname $itk_option(-catalog)
	set home $::env(GMMPS)
	
	if {$instType == "GMOS-N"} {
	    catch { set fovList [open $home/config/GMOS-N_current_fov.dat r] }
	    
	} elseif {$instType == "GMOS-S"} {
	    catch { set fovList [open $home/config/GMOS-S_current_fov.dat r] }
	    
	} elseif {$instType == "F2"} {
	    
	    set aofold ""
	    
	    catch {
		set aofold [$target_image_ fits get AOFOLD]
		# debugStatement "AOFOLD = $aofold"
	    }

	    if {$aofold != "in" || $aofold == ""} {
		catch { set fovList [open $home/config/F2-fov.dat r] }

	    } else {
		# F2-AO mode. 
		
		catch { set fovList [open $home/config/F2-AO-fov.dat r]}
	    }
	}
	
	set buf_fov  ""
	set xstart   ""
	set xend     ""
	set ystart   ""
	set yend     ""
	set xcorner1 ""
	set xcorner2 ""
	set ycorner1 ""
	set ycorner2 ""
	set pixScale [lindex $pixScale 0]

	# Read fov values:
	while {[gets $fovList line] >= 0} {
	    if {[string length $line] > 2} { 
		if {[string match *#* $line] == 0} { 
		    if {[string match *XEND* $line] == 1} {
			set xend [expr [lindex $line 1 ]/$pixScale ]
			#debugStatement "got Xend=$xend"
		    } elseif { [string match *XSTART* $line] == 1 } {
			set xstart [expr [lindex $line 1 ]/$pixScale ]
			#debugStatement "got Xstart=$xstart"
		    } elseif { [string match *YSTART* $line] == 1 } {
			set ystart [expr [lindex $line 1 ]/$pixScale ]
			#debugStatement "got ystart=$ystart"
		    } elseif { [string match *YEND* $line] == 1 } {
			set yend [expr [lindex $line 1 ]/$pixScale ]
			#debugStatement "got Yend=$yend"
		    } elseif { [string match *XCORNER1* $line] == 1 } {
			set xcorner1 [expr [lindex $line 1 ]/$pixScale ]
			#debugStatement "got Xcorner1=$xcorner1"
		    } elseif { [string match *YCORNER1* $line] == 1 } {
			set ycorner1 [expr [lindex $line 1 ]/$pixScale ]
			#debugStatement "got ycorner1=$ycorner1"
		    } elseif { [string match *XCORNER2* $line] == 1 } {
			set xcorner2 [expr [lindex $line 1 ]/$pixScale ]
			#debugStatement "got Xcorner2=$xcorner2"
		    } elseif { [string match *YCORNER2* $line] == 1 } {
			set ycorner2 [expr [lindex $line 1 ]/$pixScale ]
			#debugStatement "got ycorner2=$ycorner2"
		    } 
		} 
	    }
	}

	::close $fovList
	#debugStatement "drawBoundary, read fov data"

	if { $xend == "" || $xstart == "" || $ystart == "" || $yend == "" || $xcorner1 == "" || $xcorner2 == "" || $ycorner1 == "" || $ycorner2 == ""  } {
	    error_dialog "Error, fov.dat file missing keywords: xend=$xend, $xstart, $ystart, $yend, $xcorner1, $xcorner2 , $ycorner1, $ycorner2 , can't draw boundary"
	    return
	}
	
	#  This chops :
	#      A( xstart,yend-ycorner2) to B(xstart+xcorner2,yend)
	#      C(xend-xcorner1,yend) to D(xend, yend-ycorner2)
	#      E(xstart,ystart+ycorner1 ) to F(xstart+xcorner2, ystart)
	#      G(xend-xcorner1, ystart) to H(xend, ystart+ycorner1)
	#  
	#           XCor2         XCor1
	#            |              |
	#               B        C
	#  YCor2-     /           \
	    #           A               D
	#   
	#           E               H
	#  YCor1-     \           /
	#               F        G
	#  
	#
	set ay [expr $yend - $ycorner2]
	set bx [expr $xstart + $xcorner2]
	set cx [expr $xend - $xcorner1]
	set hy [expr $ystart + $ycorner1]

	$target_image_ convert coords $xstart $ystart image xstart ystart canvas
	$target_image_ convert coords $xend $yend image xend yend canvas
	$target_image_ convert coords $xcorner1 $ycorner1 image xcorner1 ycorner1 canvas
	$target_image_ convert coords $xcorner2 $ycorner2 image xcorner2 ycorner2 canvas

	$target_image_ convert coords $xcorner2 $ycorner2 image xcorner2 ycorner2 canvas
	$target_image_ convert coords $bx $ay image bx ay canvas
	$target_image_ convert coords $cx $hy image cx hy canvas

	# A( xstart,yend-ycorner2) to B(xstart+xcorner2,yend)
	$target_canvas_ create line $xstart $ay $bx $yend \
            -fill $fg \
            -width 2 \
            -tags "$tags $fg"            
	# B(xstart+xcorner2,yend) to C(xend-xcorner1,yend)
	$target_canvas_ create line $bx $yend $cx $yend \
            -fill $fg \
            -width 2 \
            -tags "$tags $fg"            
	# C(xend-xcorner1,yend) to D(xend, yend-ycorner2)
	$target_canvas_ create line $cx $yend $xend $ay \
            -fill $fg \
            -width 2 \
            -tags "$tags $fg"            
	# D(xend, yend-ycorner2) to H(xend, ystart+ycorner1)
	$target_canvas_ create line $xend $ay $xend $hy \
            -fill $fg \
            -width 2 \
            -tags "$tags $fg"            
	# H(xend, ystart+ycorner1) to G((xend-xcorner1, ystart)
	$target_canvas_ create line $xend $hy $cx $ystart \
            -fill $fg \
            -width 2 \
            -tags "$tags $fg"            
	# G((xend-xcorner1, ystart) to  F(xstart+xcorner2, ystart)
	$target_canvas_ create line $cx $ystart $bx $ystart \
            -fill $fg \
            -width 2 \
            -tags "$tags $fg"            
	# F(xstart+xcorner2, ystart) to  E(xstart,ystart+ycorner1 )
	$target_canvas_ create line $bx $ystart $xstart $hy \
            -fill $fg \
            -width 2 \
            -tags "$tags $fg"            
	#  E(xstart,ystart+ycorner1 ) to  A( xstart,yend-ycorner2)
	$target_canvas_ create line $xstart $hy $xstart $ay \
            -fill $fg \
            -width 2 \
            -tags "$tags $fg"            
	return
    }



    #############################################################
    #  Name: slits
    #
    #  Description:
    #   This function draws the slits for ODF files.
    #############################################################
    #############################################################


    public method slits {} {
	set bg black
	#set fg green
	set fg2 yellow
	set tags slitMark
	set target_canvas_ .skycat1.image.imagef.canvas
	set catname $itk_option(-catalog)
	set instType $itk_option(-instType)
	set scale $itk_option(-pixs)

	# Pi/180
	set rad 0.017453292519943295

	if {$plotSlits_ == 1} {
	    # Slits are being plotted, so, clearslits and return 	
	    
	    $w_.slits configure -text "Plot Slits "
	    
	    clearslits
	    
	    return
	} else {
	    # Slits are not being plotted, change button name then continue on
	    # to slit plotting routine.
	    
	    $w_.slits configure -text "Clear Slits"
	    
	    set plotSlits_ 1
	}

	#
	#  Draw the boundary.
	drawBoundary $scale $instType
	
	#
	#  Determine spectrum length by greping for #Spectrum in catalog file.
	#  Very cheesy way to do it..
	#  Convert that length to image length
	#
	set spect_len ""
	catch { set spect_len [expr [exec grep #Spectrum $catname | cut -f2 -d:] / 2.0] }
	if {$spect_len == "" } {
	    catch { 
		set spect_len [exec grep "#fits SPEC_LEN=" $catname | cut -f2 -d=] 
		set spect_len [string trim $spect_len "' \t"]
		set spect_len [expr $spect_len / 2.0]
	    }
	    if {$spect_len == "" } {
		set spect_len 3100
	    }
	}		

	$target_image_ convert dist $spect_len $spect_len image spect_len spect_len canvas

	
	#
	#  Cycle through each item in the objects list
	#  This gets num rows from total_rows(vmTableList)
	#  and info_ connections all rows in the display.
	set nobj [$results_ total_rows]
	#debugStatement "slits: Cycling thru $nobj objects"
        
	for {set n 0} {$n < $nobj} {incr n} {
            #
            #  For each item in the object list
            #  get the row, then parse out x, y, dimx, dimy.
            # BIG WARNING: it assumes order of the columns.
            # Which is kind of OK, because we create the Master OT file.
            # Col's start at 0, Col3 = x_ccd, Col4 = y_ccd,
            # Col5 = specpos_x, Col6 = specpos_y
            # Col7 = slitpos_x, Col8 = slitpos_y
            # Col9=slitsize_x(slitWidth), Col10=slitsize_y(slitLength)
            # Col13=priority
            # All col's, (except x_ccd&y_ccd, specpos_*) are in arc sec. 
            # so need to convert to pixels, /pixScale
	    
	    #if {$n > 0} {break}
	    
	    set row [lindex $info_ $n]

	    #debugStatement "slits: row = $row"
	    set x [expr [lindex $row 3] + [lindex $row 5]]
	    set y [expr [lindex $row 4] + [lindex $row 6]]
	    # I still don't understand what's the difference between x/xccd and y/yccd (Mischa 2014-08-18)
	    set xccd [expr [lindex $row 3] + ([lindex $row 7]/$itk_option(-pixs)) ]
	    set yccd [expr [lindex $row 4] + ([lindex $row 8]/$itk_option(-pixs)) ]
	    set xshift [expr [lindex $row 5 ]]
	    set yshift [expr [lindex $row 6 ]]
	    set dimx [expr [lindex $row 9] / 2.0 / $itk_option(-pixs)]
	    set dimy [expr [lindex $row 10] / 2.0 / $itk_option(-pixs)]
	    set prior [lindex $row 13]
	    
	    set waveccd ""

	    catch {set waveccd [lindex $row 15]}
	    
	    set slittilt [lindex $row 11]
            
	    set tmp 0
            
	    #
	    #  Set color of slit and spectrum drawn based on priority,
	    #  color Acq. objects different color.
	    if { $prior == "0" } {
		set fg cyan 
	    } else {
		set fg GhostWhite
	    }
	    
	    #debugStatement "slits: Object position before: $x, $y"
	    $target_image_ convert coords $x $y image x y canvas
	    $target_image_ convert coords $xccd $yccd image xccd yccd canvas
	    $target_image_ convert dist $dimx $dimy image dimx dimy canvas
	    if {$waveccd != ""} {
                if {$instType != "F2"} {
                    $target_image_ convert coords $waveccd $tmp image waveccd tmp canvas
                } else {
                    $target_image_ convert coords $tmp $waveccd image tmp waveccd canvas
                }
	    }

	    if {$instType != "F2"} {
		# GMOS (horizontal dispersion)
		set slit_xmin [expr $x - $dimx]
		set slit_xmax [expr $x + $dimx]
		set slit_ymin [expr $y - $dimy]
		set slit_ymax [expr $y + $dimy]
		set spec_xmin [expr $x - $spect_len]
		set spec_xmax [expr $x + $spect_len]
		set spec_ymin [expr $y - $dimy]
		set spec_ymax [expr $y + $dimy]

	    } else {
		# F2 (vertical dispersion)
		set slit_xmin [expr $x - $dimx]
		set slit_xmax [expr $x + $dimx]
		set slit_ymin [expr $y - $dimy]
		set slit_ymax [expr $y + $dimy]
		set spec_xmin [expr $x - $dimx]
		set spec_xmax [expr $x + $dimx]
		set spec_ymin [expr $y - $spect_len]
		set spec_ymax [expr $y + $spect_len]
	    }

	    if {$instType != "F2"} {
		
		######################################
		# GMOS spectral footprint outlines
		######################################
		
		# For plotting tilted slits we need to shift vertices. Introducing three helper variables:
		# tilt_dx1 (offset for tilting a 1D slit)
		# tilt_dx2 (additional offset to compensate for narrowing effect when tilting 2D slit)
		# tilt_dcwl (offset by which we have to shift the "green line"; a simple fraction of tilt_dx1)

		# Zero the offsets in case of no tilt
		set tilt_dx1  0.
		set tilt_dx2  0.
		set tilt_dcwl 0.
		# If there is a tilt
		if {[string is double -strict $slittilt] } {
		    set tilt_dx1 [expr $dimy * tan ($slittilt * $rad)]
		    set tilt_dx2 [expr $dimx / cos ($slittilt * $rad)]
		    set tilt_dcwl [expr $tilt_dx1 * $yshift / $dimy]
		}
		
		# Draw spectrum box: right, bottom, left, top lines (in this order)
		$target_canvas_ create line \
		    [expr $spec_xmax + $tilt_dx1] $spec_ymax \
		    [expr $spec_xmax - $tilt_dx1] $spec_ymin \
		    -fill $fg -width 1 -tags "$tags $fg"
		$target_canvas_ create line \
		    [expr $spec_xmax - $tilt_dx1] $spec_ymin \
		    [expr $spec_xmin - $tilt_dx1] $spec_ymin \
		    -fill $fg -width 1 -tags "$tags $fg"
		$target_canvas_ create line \
		    [expr $spec_xmin - $tilt_dx1] $spec_ymin \
		    [expr $spec_xmin + $tilt_dx1] $spec_ymax \
		    -fill $fg -width 1 -tags "$tags $fg"
		$target_canvas_ create line \
		    [expr $spec_xmin + $tilt_dx1] $spec_ymax \
		    [expr $spec_xmax + $tilt_dx1] $spec_ymax \
		    -fill $fg -width 1 -tags "$tags $fg"
		
                if {$waveccd != "" } {
		    # Draw red line indicating central wavelength position
		    $target_canvas_ create line \
			[expr { $waveccd + $tilt_dx1 }] $slit_ymax \
			[expr { $waveccd - $tilt_dx1 }] $slit_ymin \
			-fill red -width 2 -tags "$tags $fg"

		    # Draw green line from CWL to slit center, passing through the science object
		    $target_canvas_ create line \
			[expr $waveccd + $tilt_dcwl] [expr $y] \
			[expr $xccd + $tilt_dcwl] [expr $yccd] \
			-fill green -width 1 -tags "$tags $fg2 "
#			[expr $waveccd + $tilt_dcwl] [expr $y + $yshift] \
#			[expr $xccd + $tilt_dcwl] [expr $yccd + $yshift] \
		}

		# Draw the tilted slit; first the right edge, then the left edge
		$target_canvas_ create line \
		    [expr $xccd - $tilt_dx1 + $tilt_dx2] $slit_ymin \
		    [expr $xccd + $tilt_dx1 + $tilt_dx2] $slit_ymax \
		    -fill $fg2 -width 2 -tags "$tags $fg2 " 
		$target_canvas_ create line \
		    [expr $xccd - $tilt_dx1 - $tilt_dx2] $slit_ymin \
		    [expr $xccd + $tilt_dx1 - $tilt_dx2] $slit_ymax \
		    -fill $fg2 -width 2 -tags "$tags $fg2 " 

	    } else {
		
		######################################
		# F2 spectral footprint outlines
		######################################

		# Zero the offsets in case of no tilt
		set tilt_dy1  0.
		set tilt_dy2  0.
		set tilt_dcwl 0.
		# If there is a tilt
		if {[string is double -strict $slittilt] } {
		    set tilt_dy1 [expr $dimx * tan ($slittilt * $rad)]
		    set tilt_dy2 [expr $dimy / cos ($slittilt * $rad)]
		    set tilt_dcwl [expr $tilt_dy1 * $xshift / $dimx]
		}

		# Draw spectrum box: right, bottom, left, top lines (in this order)
		$target_canvas_ create line \
		    $spec_xmax [expr $spec_ymax - $tilt_dy1] \
		    $spec_xmax [expr $spec_ymin - $tilt_dy1] \
		    -fill $fg -width 1 -tags "$tags $fg"   		
		$target_canvas_ create line \
		    $spec_xmax [expr $spec_ymin - $tilt_dy1] \
		    $spec_xmin [expr $spec_ymin + $tilt_dy1] \
		    -fill $fg -width 1 -tags "$tags $fg"   		
		$target_canvas_ create line \
		    $spec_xmin [expr $spec_ymin + $tilt_dy1] \
		    $spec_xmin [expr $spec_ymax + $tilt_dy1] \
		    -fill $fg -width 1 -tags "$tags $fg"   		
		$target_canvas_ create line \
		    $spec_xmin [expr $spec_ymax + $tilt_dy1] \
		    $spec_xmax [expr $spec_ymax - $tilt_dy1] \
		    -fill $fg -width 1 -tags "$tags $fg" 			
		
                if {$waveccd != "" } {
		    # Draw red line indicating central wavelength position
		    $target_canvas_ create line \
			$slit_xmin [expr { $waveccd + $tilt_dy1 }] \
			$slit_xmax [expr { $waveccd - $tilt_dy1 }] \
			-fill red -width 2 -tags "$tags $fg"
		    # Draw green line from CWL to slit center, passing through the science object
		    $target_canvas_ create line \
			[ expr $x + $xshift ] [expr $waveccd  + $tilt_dcwl ] \
			[ expr $xccd + $xshift ] [expr $yccd  + $tilt_dcwl ] \
			-fill green -width 1 -tags "$tags $fg2 "
		}
		
		# Draw the tilted slit; first the upper edge, then the lower edge
		$target_canvas_ create line \
		    $slit_xmax [expr $yccd - $tilt_dy1 + $tilt_dy2] \
		    $slit_xmin [expr $yccd + $tilt_dy1 + $tilt_dy2] \
		    -fill $fg2 -width 2 -tags "$tags $fg2 "
		$target_canvas_ create line \
		    $slit_xmax [expr $yccd - $tilt_dy1 - $tilt_dy2] \
		    $slit_xmin [expr $yccd + $tilt_dy1 - $tilt_dy2] \
		    -fill $fg2 -width 2 -tags "$tags $fg2 "
	    }
	}
	
	return
    }


    #############################################################
    #  Name: clearslits
    #
    #  Description:
    #   Clear the slits
    #############################################################
    #############################################################

    public method clearslits {} {
        set target_canvas_ .skycat1.image.imagef.canvas
        $target_canvas_ delete slitMark
	
	catch { $w_.slits configure -text "Plot Slits " }
	
	set plotSlits_ 0
	
	#debugStatement "clearslits: plotSlits_ = $plotSlits_"
	
        return
    }


    #############################################################
    #  Name: define_bands
    #
    #  Description:
    #   opens the "define bands" interface 
    # 
    #  Author: cba
    #############################################################
    #############################################################
    
    public method define_bands {args} {
	
	
	if {$args == "drawBandsOnly" && $plotBands_ == 1} {
	    
	    # Clear band objects.
	    .skycat1.image.imagef.canvas delete bandrects
	    
	    $w_.plotBands configure -text "Plot Bands "
	    set plotBands_ 0
	    
	    return 
	} elseif {$args == "drawBandsOnly"} {
	    
	    set bandPlotFail [::gmmps::band_def_UI::drawBands $w_ $itk_option(-catalog)]
	    
	    if {$bandPlotFail == 0} {
		$w_.plotBands configure -text "Clear Bands"
		
		set plotBands_ 1
	    }
	    
	    return
	}
	
	promptBandFileSelect
	
	if {$itk_option(-bdfile) == ""} {
	    debugStatement "No band file selected, cannot enter N&S."
	    puts "Nod & Shuffle cancelled."
	    return
	}
	
	set instType $itk_option(-instType)
	
	#  Determine if slitpos_x and slitpos_y exist in the
	#  catalog files.
	set addPX ""
	set addPY ""
	
	catch {set addPX [exec grep slitpos_x $itk_option(-catalog) ]}
	catch {set addPY [exec grep slitpos_y $itk_option(-catalog) ]}
	if { $addPX == "" } {
	    set addPX "yes" 
	} else {
	    set addPX "no" 
	}
	
	if { $addPY == "" } {
	    set addPY "yes" 
	} else {
	    set addPY "no" 
	}
	
	# Make sure there are no invalid slit tilts. 
	if {[check_tilts 0.0 0.0 ]} {
	    return 
	}
	
	# Destroy any open band_def_ui and spoc window.
	catch {destroy $bw_}
	catch {destroy $spoc_}
	
	update idletasks

	# Find detector info.		
	setDetInfo $instType
	
	#
	#  Call the gmmps_spoc.tcl init function
	set bw_ [utilReUseWidget gmmps::band_def_UI .skycat1.bandDefUI $this $itk_option(-catalog) $itk_option(-xsize) $itk_option(-tilt) $addPX $addPY $itk_option(-pixs) $itk_option(-bw) $this.results $config_file_ $DET_IMG_ $DET_SPEC_ $itk_option(-instType) $itk_option(-bdfile)]
	#if {[catch {set bw_ [utilReUseWidget gmmps::band_def_UI .skycat1.bandDefUI $this $itk_option(-catalog) $itk_option(-xsize) $itk_option(-tilt) $addPX $addPY $itk_option(-pixs) $itk_option(-bw) $this.results $config_file_ $DET_IMG_ $DET_SPEC_ $itk_option(-instType)] } msg]} {
	#	debugStatement "Failed to spoc"
	#	if { $args == "drawBandsOnly" } { 
	#		error_dialog "Error Drawing Bands... Band Def UI must be closed as it is configured and used for this operation."
	#	} else {
	#		error_dialog "Error Opening Band Def UI\nIs the UI already up? $msg"
	#	}
	return 
    }
    
    return
}


#########################################################################
#  Name: promptBandFileSelect
#
#  Description: Creates a dialog window prompting the user to select or 
# 	create a band definition file for use with any created ODF.
#########################################################################
#########################################################################

protected method promptBandFileSelect {} {
    # Search for bandfiles associated with our window.
    # (needs to be global for use with listbox widget)
    global foundBandFiles
    set foundBandFiles [findBandFiles [file dirname $itk_option(-catalog)]]
    set foundBandFiles [linsert $foundBandFiles 0 "tmp.bands" ]
    
    # find max length so we can configure the listbox correctly.
    set maxNameLen 0
    foreach foundFile $foundBandFiles {
	if {[string length $foundFile ] > $maxNameLen} {
	    set maxNameLen [string length $foundFile ]
	}
    }
    incr maxNameLen
    
    # Create dialog window.
    set pbf [toplevel $w_.pickbf ]
    wm title $w_.pickbf "Pick .bands File"
    
    # Associate pressing the 'x' with choosing tmp.bands.
    wm protocol $pbf WM_DELETE_WINDOW [code $this chooseBandFile $pbf "cancel"]
    
    pack \
	[label $pbf.bandfileprompt \
	     -text "Please select or create a new band definition file." ] \
	[frame $pbf.oldFiles ] \
	[frame $pbf.bandButtons ] \
	[label $pbf.explainTextBands \
	     -text "\nThe band definition file you choose will be associated with \nany ODF file(s) created during this N&S design session.     \n\nChoose tmp.bands if you do not wish to have a dedicated BDF."]
    
    pack \
	[button $pbf.cancel \
	     -command [code $this chooseBandFile $pbf "cancel"] \
	     -text "Cancel" ] \
	[button $pbf.selectOld \
	     -text "Use Selected" \
	     -command [code $this chooseBandFile $pbf "selected"]] \
	[button $pbf.newFile \
	     -command [code $this chooseBandFile $pbf "new"] \
	     -text "Create New .band File" ] \
	-side right \
	-padx 2 \
	-in $pbf.bandButtons
    
    pack \
	[label $pbf.prompt \
	     -text "Select file:" ]\
	[listbox $pbf.selectBox \
	     -listvariable foundBandFiles \
	     -selectmode single \
	     -width $maxNameLen ] \
	-side left \
	-in $pbf.oldFiles
    
    $pbf.selectBox selection set 0
    
    tkwait window $pbf
}


#########################################################################
#  Name: chooseBandFile
#
#  Description: Creates a dialog window prompting the user to select or 
# 	create a band definition file for use with any created ODF.
#########################################################################
#########################################################################

protected method chooseBandFile {dialog_window mode} {
    
    if {$mode == "cancel"} {
	
	set itk_option(-bdfile) ""
	
	destroy $dialog_window
	
    } elseif {$mode == "new"} {
	set bandType [list [list ".bands Band Definition File" {".bands" ".BANDS"}]]
	
	# Search for appropriate initfile name.
	set cnt 1
	set initFile [format "%sbanddef%d.bands" [file rootname $itk_option(-catalog) ] $cnt]
	while {[file exists $initFile ]} {
	    incr cnt
	    set initFile [format "%sbanddef%d.bands" [file rootname $itk_option(-catalog) ] $cnt]
	}
	set initFile [file tail $initFile ]
	
	# Open "new file" dialog.
	set newBandFile [tk_getSaveFile \
			     -defaultextension ".bands" \
			     -filetypes $bandType \
			     -title "New Band Definition" \
			     -initialdir [file dirname $itk_option(-catalog) ]\
			     -initialfile $initFile \
			     -parent $w_ ]
	
	# If a name was choosen, create that file.
	if {$newBandFile != ""} {
	    
	    set bdfile [open $newBandFile "w" ]
	    puts $bdfile "OT:$itk_option(-catalog)"
	    puts $bdfile "shuffleMode:microShuffle"
	    puts $bdfile "shuffleAmt:5.0"
	    puts $bdfile "slitlen:5.0"				
	    ::close $bdfile
	    
	    # Set bdfile.
	    set itk_option(-bdfile) $newBandFile
	    
	    gmmps_config_ [list "shuffle_mode" "slitlen_ms" "shuffleamt_ms" ] [list "microShuffle" "5.0" "5.0" ]
	    
	    destroy $dialog_window
	}		

    } elseif {$mode == "selected"} {

	# Query listbox for selected bdfile name.
	set cur [$dialog_window.selectBox curselection ]

	if {$cur == ""} {
	    error_dialog "Please select a band definition file before pressing this button."
	    return
	}
	
	set bandFileName [$dialog_window.selectBox get $cur ]
	
	set bandFileLongName [format %s/%s [file dirname $itk_option(-catalog)] $bandFileName ]
	
	# If in "tmp.bands" mode and "tmp.bands" doesn't exist, create it. 
	if {$bandFileName == "tmp.bands"} {
	    makeTmpBands
	    
	}
	
	set bandFileName $bandFileLongName
	
	set shufMode ""
	set slitlen_ms ""
	set shuffleAmt_ms ""
	set bandSize_bs ""
	set shufflePx_bs ""
	set yOffset_bs ""
	
	# Parse file data.
	set bandFile [open $bandFileName "r" ]
	while {[gets $bandFile line] >= 0} {
	    set line [split $line ":" ]

	    if {[lindex $line 0 ] == "shuffleMode"} {
		set shufMode [string trim [lindex $line 1 ]]
		#debugStatement "shuffleMode : $shufMode loaded from file."
	    } elseif {[lindex $line 0 ] == "shuffleAmt"} {
		set shuffleAmt_ms [string trim [lindex $line 1 ]]
		#debugStatement "sA_ms $shuffleAmt_ms"
	    } elseif {[lindex $line 0 ] == "yOffset"} {
		set yOffset_bs [string trim [lindex $line 1 ]]
		#debugStatement "yOffset $yOffset_bs"
	    } elseif {[lindex $line 0 ] == "shufflePx"} {
		set shufflePx_bs [string trim [lindex $line 1 ]]
		#debugStatement "shufflePx_bs $shufflePx_bs"
	    } elseif {[lindex $line 0 ] == "slitlen"} {
		set slitlen_ms [string trim [lindex $line 1 ]]
		set bandSize_bs $slitlen_ms
		#debugStatement "slitlen_ms $slitlen_ms"
		#debugStatement "bandSize_bs $slitlen_ms"
	    }
	}			
	::close $bandFile
	
	set itk_option(-bdfile) $bandFileName
	
	# Load the data into the .gmmps.cfg file, so that it will be loaded
	# into the N&S GUI when it is created.
	if {$shufMode == "bandShuffle"} {
	    debugStatement "Sending $shufMode $bandSize_bs $shufflePx_bs $yOffset_bs to config file"
	    gmmps_config_ [list "shuffle_mode" "bandsize_bs" "shufflepx_bs" "yoffset_bs" ] [list $shufMode $bandSize_bs $shufflePx_bs $yOffset_bs ]
	} else { 
	    debugStatement "Sending $shufMode $slitlen_ms $shuffleAmt_ms to config file."
	    gmmps_config_ [list "shuffle_mode" "slitlen_ms" "shuffleamt_ms" ] [list $shufMode $slitlen_ms $shuffleAmt_ms ]
	}
	
	destroy $dialog_window
	
	# Tell the N&S GUI which mode we are in.
	global shuffleMode 
	set shuffleMode $shufMode
    }
}


#########################################################################
#  Name: makeTmpBands
#
#  Description: makes a fresh bands.tmp file, if one doesn't already exist.
#########################################################################
#########################################################################	

protected method makeTmpBands {args} {

    # If in "tmp.bands" mode and "tmp.bands" doesn't exist, create it. 
    set tmpName [format %s/tmp.bands [file dirname $itk_option(-catalog) ]]
    if {![file exists $tmpName ]} {
	
	set mkTmp [open $tmpName "w" ]
	puts $mkTmp "shuffleMode:microShuffle"
	puts $mkTmp "slitlen:5.000"
	puts $mkTmp "shuffleAmt:5.000"
	puts $mkTmp "shufflePx:69"
	::close $mkTmp
    }
    
    set itk_option(-bdfile) $tmpName
    
    if {$args == "closeafter"} {
	catch {destroy $w_.pickbf}
    }
}


#########################################################################
#  Name: odfBandSearch
#
#  Description: Finds an .bands file associated with the given odf.
#########################################################################
#########################################################################	

protected proc odfBandSearch {odfFile} {
    set dir [file dirname $odfFile ]
    set allBandFiles ""
    catch {set allBandFiles [lsort [glob -directory $dir *.bands ]]}	
    
    set foundBFs [list ]
    
    foreach bandFileName $allBandFiles {
	# tmp.bands will be inserted into the list later.
	if {[file tail $bandFileName ] == "tmp.bands"} {
	    continue
	}
	
	set bandFile [open $bandFileName "r" ]
	
	while {[gets $bandFile line ] >= 0} {
	    if {$line == "ODF:$odfFile"} {
		set  [split $bandOdf ":" ]
		set foundBFs [lappend foundBFs [file tail $bandFileName ]]
	    }
	}
	::close $bandFile
    }
    
    
    if {[llength $foundBFs ] > 1} {
	# If more than one associated band file was found prompt user to choose
	set maxNameLen 0
	foreach bandName $foundBFs {
	    if {[string length $bandName ] > $maxNameLen} {
		set maxNameLen $[string length $bandName ]
	    }
	}
	
	# Create dialog window.
	set pbf [toplevel $w_.pickbf2 ]
	wm title $w_.pickbf "Pick .bands File"
	
	# Associate pressing the 'x' with choosing tmp.bands.
	wm protocol $pbf WM_DELETE_WINDOW [code $this makeTmpBands "closeafter"]
	
	pack \
	    [label $pbf.bandfileprompt \
		 -text "More than one bandfile found. Please select a bandfile to display." ] \
	    [frame $pbf.oldFiles ] \
	    [button $pbf.selectOld \
		 -text "Select" \
		 -command {set chosenBF [$pbf.selectBox curselection ]} ]
	
	pack \
	    [label $pbf.prompt \
		 -text "Select file:" ]\
	    [listbox $pbf.selectBox \
		 -listvariable foundBFs \
		 -selectmode single \
		 -width $maxNameLen ] \
	    -side left \
	    -in $pbf.oldFiles
	
	$pbf.selectBox selection set 0
	
	tkwait variable chosenBF $pbf2			
	
	global chosenBF
	
	#debugStatement "chosenBF: $chosenBF"
	
	#debugStatement [format "returning: %s/%s" [file dirname $odfFile ] [pbf.selectBox get $chosenBF ]]
	
	return [format "%s/%s" [file dirname $odfFile ] [pbf.selectBox get $chosenBF ]]
	
    } elseif {[llength $foundBFs ] == 1} {
	return [lindex $foundBFs 0 ]
	
    } else {
	
	return ""
    }
}

#########################################################################
#  Name: findBandFiles
#
#  Description: Searches a directory for .bands files associated with the 
#	current OT file that created the N&S window.
#########################################################################
#########################################################################

protected method findBandFiles {dir} {
    set foundFiles [list ]
    set allBandFiles ""
    catch {set allBandFiles [glob -directory $dir *.bands ]}
    
    foreach bandFileName $allBandFiles {
	
	# tmp.bands will be inserted into the list later.
	if {[file tail $bandFileName ] == "tmp.bands"} {
	    continue
	}
	
	set associatedCat ""
	catch {set associatedCat [string trim [exec grep "OT:" $bandFileName | cut -f2 -d: ]]}
	
	#debugStatement "findBandFiles checked: $bandFileName\n.$associatedCat"
	
	# Match! 
	if {$associatedCat == $itk_option(-catalog)} {
	    set foundFiles [lappend foundFiles [file tail $bandFileName ]]
	}
    }
    
    return $foundFiles
}


#############################################################
#  Name: plotGmosGaps 
#
#  Description:
#   plot the gaps that exist between the three GMOS CCDs
#   40 pixels centered at 2067 and 4153
#   opens the "define bands" interface 
# 
#  Author: cba
#############################################################
#############################################################


public method plotGmosGaps {} {

    set det $detector_
    set detName [lindex $det 3 ]

    if {$plotGaps_} {
	set plotGaps_ 0
	set target_canvas_ .skycat1.image.imagef.canvas

	$target_canvas_ delete gaps
	
	change_gap_button_names "Plot GMOS $detName Detector Gaps "
	
	
        return
    } else {
	set plotGaps_ 1
	
	change_gap_button_names "Clear GMOS $detName Detector Gaps"
	
    }

    set tags gaps
    
    set home $::env(GMMPS)
    set target_image_ image2
    
    set pixelScale $itk_option(-pixs)
    set instType $itk_option(-instType)
    
    set detId 0  
    
    set gapAx 0
    set gapBx 0
    set gapW  0

    set gapAxL 0
    set gapBxL 0
    set gapWL  0
    
    set pxFound 0
    set pxFoundL 0
    
    set gapAx [lindex $det 4 ]
    set gapBx [lindex $det 5 ]
    set gapW  [lindex $det 6 ]

    set bottomy 0

    # calculated values
    # alter set values above based on pixelScale
    set gapAx [ expr $gapAx / $pixelScale ]
    set gapBx [ expr $gapBx / $pixelScale ]
    set gapW  [ expr $gapW  / $pixelScale ]

    # debugStatement "GMOS Gap values: gapAx = $gapAx, gapBx = $gapBx, gapW = $gapW."

    # other variable calculations
    set halfGapW [expr $gapW / 2]
    set gapAx0 [expr $gapAx - $halfGapW ]
    set gapAx1 [expr $gapAx0 + $gapW]
    set topy [$target_image_ height]
    set gapBx0 [expr $gapBx - $halfGapW ]
    set gapBx1 [expr $gapBx0 + $gapW]

    # convert to canvas coords
    set tmpcc 0
    $target_image_ convert coords $gapAx0 $bottomy image gapAx0 bottomy canvas
    $target_image_ convert coords $gapAx1 $topy image gapAx1 topy canvas
    $target_image_ convert coords $gapBx0 $tmpcc image gapBx0 tmpcc canvas
    $target_image_ convert coords $gapBx1 $tmpcc image gapBx1 tmpcc canvas

    set target_canvas_ .skycat1.image.imagef.canvas
    # delete current marks @@note: marks not cleared when image is reloaded
    $target_canvas_ delete $tags
    # plot first gap
    $target_canvas_ create line \
	$gapAx0 $bottomy \
	$gapAx0 $topy \
	$gapAx1 $topy \
	$gapAx1 $bottomy \
	$gapAx0 $bottomy \
	-fill blue \
	-width 1 \
	-tags $tags

    # plot second gap
    $target_canvas_ create line \
	$gapBx0 $bottomy \
	$gapBx0 $topy \
	$gapBx1 $topy \
	$gapBx1 $bottomy \
	$gapBx0 $bottomy \
	-fill blue \
	-width 1 \
	-tags $tags
    
    plotGmosBarcode $tags $target_canvas_

    #  Draw the boundary.
    drawBoundary $pixelScale $instType

}


#############################################################
#  Name: plotGmosBarcode
#
#  Description: Plots the GMOS barcode on the image.
#  
#############################################################

protected method plotGmosBarcode {tags target_canvas_} {
    
    set pixelScale $itk_option(-pixs)
    
    set instType $itk_option(-instType)

    # Load FoV data.
    if {$instType == "GMOS-N"} {
	set fovList [open "$home_/config/GMOS-N_current_fov.dat" "r" ]
    } else {
	set fovList [open "$home_/config/GMOS-S_current_fov.dat" "r" ]
    }

    set xstart ""
    set xend ""
    set yend ""
    set ystart ""
    
    while {[gets $fovList line] >= 0} {
	if {[string length $line] > 2} { 
	    if {[string match *#* $line] == 0} { 
		 if {[string match *XEND* $line] == 1} {
		     set xend [expr [lindex $line 1 ] ]
		     #debugStatement "got Xend=$xend"
		 } elseif { [string match *YSTART* $line] == 1 } {
		     set ystart [expr [lindex $line 1 ] ]
		     #debugStatement "got ystart=$ystart"
		 } elseif { [string match *XSTART* $line] == 1 } {
		     set xstart [expr [lindex $line 1 ] ]
		     #debugStatement "got ystart=$ystart"
		 } elseif { [string match *YEND* $line] == 1 } {
		     set yend [expr [lindex $line 1 ] ]
		     #debugStatement "got Yend=$yend"
		 } 
	     } 
	    }
	}	
	
	::close $fovList
	
	if {$xend == "" || $yend == "" || $ystart == ""} {
	    debugStatement "Could not draw barcode because fov values could not be loaded."
	    return 1
	}
	
	# Find the center of the FoV in x and y.
	set ymid [expr ($yend - $ystart) / 2 + $ystart]
	set xmid [expr ($xend - $xstart) / 2 + $xstart]
	
	#debugStatement "ymid is $ymid"
	
	# Load barcode dimensions.
	set barFile [open "$home_/config/barcode.dat" "r" ]
	
	# xdelta_bar is the distance from the center of the FOV to the near
	#  edge of the barcode.
	set xdelta_bar ""
	set xend_bar ""
	set xdim_bar ""
	set ydim_bar ""
	
	while {[gets $barFile line] >= 0} {
	    if {[string length $line] > 2} { 
		if {[string match *#* $line] == 0} { 
		     if { [string match *XDELTA* $line] == 1 } {
			 set xdelta_bar [expr [lindex $line 1 ] ]
			 #debugStatement "got xdel_bar=$xdelta_bar"
		     } elseif { [string match *YDIM* $line] == 1 } {
			 set ydim_bar [expr [lindex $line 1 ] ]
			 #debugStatement "got ydim_bar=$ydim_bar"
		     } elseif { [string match *XDIM* $line] == 1 } {
			 set xdim_bar [expr [lindex $line 1 ] ]
			 #debugStatement "got xdim_bar=$xdim_bar"
		     } 
		 } 
		}
	    }
	    ::close $barFile
	    
	    if {$xdelta_bar == "" || $xdim_bar == "" || $ydim_bar == ""} {
		debugStatement "Could not draw barfile: dimensions not found."
		return 1
	    }
	    
	    # Calculate barcode rectangle dimensions.
	    
	    debugStatement "$xmid + $xdelta_bar [expr $xmid + $xdelta_bar] [expr ($xmid + $xdelta_bar) / $pixelScale]" 
	    
	    set barTop [expr $ymid + $ydim_bar / 2.0 ]
	    set barBottom [expr $ymid - $ydim_bar / 2.0 ]
	    set barLeft [expr $xmid + $xdelta_bar]
	    set barRight [expr  $barLeft + $xdim_bar ]
	    
	    # Convert to pixel coordinates.
	    set barTop [expr $barTop / $pixelScale ]
	    set barBottom [expr $barBottom / $pixelScale ]
	    set barLeft [expr $barLeft / $pixelScale ]
	    set barRight [expr $barRight / $pixelScale ]
	    
	    # Convert to skycat canvase coordinates.
	    $target_image_ convert coords $barLeft $barTop image barLeft barTop canvas
	    $target_image_ convert coords $barRight $barBottom image barRight barBottom canvas
	    
	    debugStatement "drawing barcode with coords $barLeft $barTop $barRight $barBottom"
	    
	    # Draw barcode rectangle.
	    $target_canvas_ create rect $barLeft $barTop $barRight $barBottom \
		-outline blue \
		-width 1 \
		-tags $tags
	    
	    return 0
	}


	#############################################################
	#  Name: plotF2detector 
	#
	#  Plots the Flamingos-2 detector in the skycat display.
	#  
	#############################################################


	public method plotF2Detector {} {
	    
	    # Toggle drawing and clearing depending on the current state.
	    if {$plotGaps_ == 1} {
		set target_canvas_ .skycat1.image.imagef.canvas
		$target_canvas_ delete f2
		$w_.plotGmosGaps configure -text "Plot FLAMINGOS-2 Field Layout "
		
		set plotGaps_ 0
		
		return
	    } else {
		$w_.plotGmosGaps configure -text "Clear FLAMINGOS-2 Field Layout"
		set plotGaps_ 1
	    }
	    
	    debugStatement "plotting F2 Detector"
	    
	    set tags f2
	    set x0 1 
	    set y0 1
	    set x1 2048
	    set y1 2048
	    
	    set moswid 671
	    set mosy0 688
	    set mosy1 [expr $mosy0 + $moswid]
	    
	    set mosx0 $x0
	    set mosx1 $x1
	    
	    set target_image_ image2

	    # convert to canvas coords
	    $target_image_ convert coords $x0 $y0 image x0 y0 canvas
	    $target_image_ convert coords $x1 $y1 image x1 y1 canvas
	    $target_image_ convert coords $mosx0 $mosy0 image mosx0 mosy0 canvas
	    $target_image_ convert coords $mosx1 $mosy1 image mosx1 mosy1 canvas
	    
	    
	    set target_canvas_ .skycat1.image.imagef.canvas
	    # delete current marks @@note: marks not cleared when image is reloaded
	    $target_canvas_ delete $tags

	    # draw circular illuminated area
	    $target_canvas_ create oval \
		$x0 $y0 $x1 $y1 \
		-outline red \
		-tags $tags
	    
	    # draw MOS region
	    $target_canvas_ create line \
		$mosx0 $mosy0 \
		$mosx0 $mosy1 \
		$mosx1 $mosy1 \
		$mosx1 $mosy0 \
		$mosx0 $mosy0 \
		-fill blue \
		-width 1 \
		-tags $tags
	    
	    return
	}



	public method _sel {flag {del 0}} {
	    #  Get position of PRIOIRTY, slittype and id col.

	    #  Get the results from the display (might be multiple rows)
	    set rows [$results_ get_selected ]
	    
	    #puts "sel entry: rows is:\n$rows"
	    
	    # Save selection.
	    $results_ save_yview
	    $results_ save_selection		
	    
	    set head $itk_option(-headingx)
	    set PRIORITY [lsearch -regex $head {(?i)priority}]
	    set TYPE [lsearch -regex $head {(?i)slittype}]
	    set ID [lsearch -regex $head {(?i)ID}]
	    set SLITSIZEY [lsearch -regex $head {(?i)slitsize_y}]
	    set SLITSIZEX [lsearch -regex $head {(?i)slitsize_x}]
	    set SLITTILT  [lsearch -regex $head {(?i)slittilt}]
	    set SLITPOSX  [lsearch -regex $head {(?i)slitpos_x}]
	    set SLITPOSY  [lsearch -regex $head {(?i)slitpos_y}]
	    
	    set originalRowIds [list ]
	    foreach row $rows {
		set originalRowIds [lappend originalRowIds [lindex $row $ID ]]
	    }
	    
	    set catname $itk_option(-catalog)
	    
	    # Make sure the list of rows is in increasing order by ID.
	    #if { ![is_sorted $rows ]} {	
	    #	set rows [sort_rowlist $rows ]
	    #}
	    
	    #  Check that the flag value pased in is valid.
	    set flagCheck [regexp {Ssx|Ssy|Spx|Spy|St|Sall|0|1|2|3|X|Y} $flag ]

	    if { $flagCheck == 0 } {
		error_dialog "Invalid priority or slittype <$flag>."
		return 1
	    }
	    
	    # The current row index in rows that we are replacing.
	    set allRows [$results_ get_contents ]
	    
	    # construct list of columns that will be changed. 
	    if {$flag == "Ssx"} {
		set cols [list $SLITSIZEX ]
	    } elseif {$flag == "Ssy"} {
		set cols [list $SLITSIZEY ]
	    } elseif {$flag == "Spx"} {
		set cols [list $SLITPOSX ]
	    } elseif {$flag == "Spy"} {
		set cols [list $SLITPOSY ]
	    } elseif {$flag == "St"} {
		set cols [list $SLITTILT ]
	    } elseif {$flag == "Sall"}  {
		set cols [list $SLITSIZEX $SLITSIZEY $SLITPOSX $SLITPOSY $SLITTILT ]
	    } else {
		set cols [list $PRIORITY ]
	    }
	    
	    # Send start signal to undo object.

	    # Here we are hoping that the rows are sorted in order by ID, 
	    # if they are not then we will need several iterations of the while
	    # loop... but, this is still going to be more efficient than before.
	    while { [llength $rows ] > 0 } {
		
		#puts "while start rows is:\n$rows"
		
		# Clear the catalog widget. 
		$results_ clear
		
		# Set found flag to true for init.
		set found 1
		
		# Scan through all rows			
		foreach catRow $allRows {
		    
		    if { $found == 1 && [llength $rows ] > 0 } {
			set row [lindex $rows 0 ]
			
			# Remove first item from the list.
			set rows [lrange $rows 1 end ]
			
			# found flag off.
			set found 0
			#puts "searching for item: $row"
			#puts "rows is:\n$rows"
		    }
		    
		    if { [llength $rows ] >= 0 && [lindex $row $ID ] == [lindex $catRow $ID ] } { 
			# ID match, found a row to edit...
			
			#puts "row [lindex $row $ID ] found."
			
			set found 1
			set oldrow $row
			set newrow $row
			set problemsStr ""

			# Confirm when editing acq objects.
			if {[lindex $row $PRIORITY ] == "0"} {
			    set answer [tk_messageBox -message "Object [lindex $row $ID ] is an acquisition object.\n\nAre you sure you want to edit this object?" -title "WARNING" -parent $w_ -type yesno ]
			    
			    if {$answer == "no"} {
				tk_messageBox -message "Edit canceled for object [lindex $row $ID ]." -parent $w_
				$results_ append_row $catRow
				continue
			    }
			}
			
			if { $flag == "A" || $flag == "R" } {
			    set newrow [lreplace $row $TYPE $TYPE $flag]
			} elseif { [string index $flag 0] == "S" } {
			    switch -regexp $flag {
				Ssx { 
				    set tsx  [$w_.slitsize_x get]
				    if {$tsx == ""} {
					append problemsStr "Slit Size X is null\n"
				    } elseif { $SLITSIZEX == -1} {
					append problemsStr "Slit Size X Column Not Present\n"
				    } else {
					set newrow [lreplace $newrow $SLITSIZEX $SLITSIZEX [format "%.6f" $tsx]]
				    }
				}
				Ssy { 
				    set tsy  [$w_.slitsize_y get]	
				    if {$tsy == ""} {
					append problemsStr "Slit Size Y is null\n"
				    } elseif { $SLITSIZEY == -1} {
					append problemsStr "Slit Size Y Column Not Present\n"
				    } else {
					set newrow [lreplace $newrow $SLITSIZEY $SLITSIZEY [format "%.6f" $tsy]]
				    }
				}
				St  {
				    set tt   [$w_.slittilt get]
				    if {$tt == ""} {
					append problemsStr "Slit Tilt is null\n"
				    } elseif {$SLITTILT == -1} {
					append problemsStr "Slit Tilt Column Not Present\n"
				    } elseif {($tt > 45.0 && !($tt <= 360.0 && $tt >= 315.0) || $tt < -45.0)} {
					append problemsStr "Slit Tilt $tt is outside valid range (-45.0 to 45 degrees).\n"
				    } else {
					set newrow [lreplace $newrow $SLITTILT $SLITTILT [format "%.6f" $tt]]
				    }
				}
				Spx { 
				    set tspx [$w_.slitpos_x get]
				    if {$tspx == ""} {
					append problemsStr "Slit Pos X is null\n"
				    } elseif {$SLITPOSX == -1} {
					append problemsStr "Slit Pos X Column Not Present\n"
				    } else {
					set newrow [lreplace $newrow $SLITPOSX $SLITPOSX [format "%.6f" $tspx]]
					#set newrow [recalculate_specpos $newrow ]
				    }
				}
				Spy {
				    set tspy [$w_.slitpos_y get]
				    if {$tspy == ""} {
					append problemsStr "Slit Pos Y is null\n"
				    } elseif {$SLITPOSY == -1} {
					append problemsStr "Slit Pos Y Column Not Present\n"
				    } else {
					set newrow [lreplace $newrow $SLITPOSY $SLITPOSY [format "%.6f" $tspy]]
				    }
				}
				Sall {
				    set newrow $row
				    set tsx  [$w_.slitsize_x get]
				    set tsy  [$w_.slitsize_y get]
				    set tt   [$w_.slittilt get]
				    set tspx [$w_.slitpos_x get]
				    set tspy [$w_.slitpos_y get]
				    if {$tsx != "" && $SLITSIZEX != -1} {
					set newrow [lreplace $newrow $SLITSIZEX $SLITSIZEX [format "%.6f" $tsx]]
				    }
				    if {$tsy != "" && $SLITSIZEY != -1} {
					set newrow [lreplace $newrow $SLITSIZEY $SLITSIZEY [format "%.6f" $tsy]]
				    }
				    if {$tt != "" && $SLITTILT != -1} {
					# Make sure slit tilts are valid.
					if {($tt > 45.0 && !($tt <= 360.0 && $tt >= 315.0) || $tt < -45.0)} {
					    append problemsStr "Slit Tilt $tt is outside valid range (-45.0 to 45 degrees).\n"
					} else {
					    set newrow [lreplace $newrow $SLITTILT $SLITTILT [format "%.6f" $tt]]
					}
				    }
				    if {$tspx != "" && $SLITPOSX != -1} {
					set newrow [lreplace $newrow $SLITPOSX $SLITPOSX [format "%.6f" $tspx]]
				    }
				    if {$tspy != "" && $SLITPOSY != -1} {
					set newrow [lreplace $newrow $SLITPOSY $SLITPOSY [format "%.6f" $tspy]]
				    }
				}
			    }
			} elseif {$flag == "X"} { 
			    # If we are ignoring an object be careful not to lose priority info 
			    # for that object.
			    
			    set newrow [lreplace $row $PRIORITY $PRIORITY "X"]
			    
			} else {
			    # Change priority. Here the prio will be changed to one of 0, 1, 2, 3 from any
			    # other priority. Flag can have any of the aformentioned values. 
			    if {$SLITTILT != -1} { 
				set tt [lindex $row $SLITTILT ]
				if {($tt > 45.0 && !($tt <= 360.0 && $tt >= 315.0) || $tt < -45.0)} {
				    append problemsStr "Cannot restore slit. Slittilt $tt is outside valid range (-45.0 to 45 degrees).\n"					
				}
			    }
			    
			    set prio [lindex $row $PRIORITY]
			    set newrow [lreplace $row $PRIORITY $PRIORITY $flag]
			}
			
			# If there were problems add the old unmodified row. Else add new row. 
			if {$problemsStr != ""} {
			    error_dialog $problemsStr
			    $results_ append_row $catRow
			} else {
			    
			    $results_ append_row $newrow
			    
			}
			
		    } else {
			# This row is not one we wish to edit at this time. 
			# So add it back to the catalog unmodified.
			$results_ append_row $catRow
		    }
		}
	    }
	    
	    # Finished input signal to catalog display widget. 
	    $results_ new_info
	    
	    $results_ save_to_file $itk_option(-catalog) [$this.results get_contents ] "" 1
	    
	    # Restore the selection.
	    reload_catalog
	    $results_ restore_yview
	    $results_ restore_selection

	    # Check to see if row order or contents has changed. If so warn user.
	    set post_rows [$results_ get_selected ]
	    set cur 0
	    foreach row $post_rows {
		if {[lindex $row $ID ] != [lindex $originalRowIds $cur ]} {
		    #tk_messageBox -message "Warning: Table order changed.\nSelected rows may be different."
		    tk_messageBox -message "Warning: Table row order changed.\nHighlighted rows may not reflect previous selection." -parent $w_ -title "Row Order Changed"
		    break
		}
		
		incr cur
	    }
	    
	    return 0
	}


	#############################################################
	#  Name: open_catalog
	#
	#  Description:
	#   This function opens the catalog for this window.
	#   Called by layoutdialog(), which is called by init()
	#   Specifically:
	#        This create the w_ object used throughout this code.
	#        Initializes it
	#        
	#############################################################
	#############################################################

	public method open_catalog {} {

	    #
	    # create tcs object
	    #
	    global ::$w_.tcs

	    astrocat $w_.cat

	    # normally -catalog should be specified when creating this widget
	    # if not, choose a default...
	    if {"$itk_option(-catalog)" == ""} {
		return
		if {[catch {set catalog_list [$w_.cat info $itk_option(-catalogtype)]} msg]} {
		    error_dialog $msg $w_
		    return
		}
		set itk_option(-catalog) [lindex $catalog_list [expr [llength $catalog_list]-1]]
	    }

	    # open the catalog file
	    #
	    set name $itk_option(-catalog)
	    if {[catch {$w_.cat open $name} msg]} {
		error_dialog $msg $w_
		return
	    }
	    
	    # set iscat_ to true if the catalog is not an image server
	    #
	    set iscat_ 1
	    if {"[$w_.cat servtype]" == "imagesvr"} {
		set iscat_ 0
	    } 

	    # if this is a local catalog, add it to the catalog menus and tree
	    # XXX should check if local catalog was already known
	    if {"[$w_.cat servtype]" == "local"} {
		update_catalog_menus

		# add to catalog tree, if there is one
		catch {.catinf insert_node $name}
		
		# add an entry to the config file for this catalog
		cat::CatalogInfo::save "" $w_ 0
	    }
	    
	    # display catalog name in header and icon
	    #debugStatement [format "/n/n$itk_option(-catalog) ... %s ... $name" [$w_.cat longname $name ]]
	    #debugStatement [format "Setting cat window title to: %s" "[$w_.cat longname $name] ($itk_option(-number))" ]
	    
	    wm title $w_ "[file tail $name] ($itk_option(-number))"
	    wm iconname $w_ "[$w_.cat shortname $name]"
	    
	}


	#############################################################
	#  Name: reload_catalog
	#
	#  Description:
	#	Reload catalog.
	#############################################################
	#############################################################

	method reload_catalog {} {
	    set catalogName $itk_option(-catalog)
	    #set catType $itk_option(-catType)
	    set instType $itk_option(-instType)

	    # Call select catalog again with specs for this catalog.
	    select_catalog $catalogName local "namespace inscope ::skycat::SkyCat .skycat1.image" "::skycat::vmSkySearch" 0 1 "" $instType
	    
	    update idletasks
	    after idle [code $this redraw_canvas_items ]
	}


	method redraw_canvas_items {} {
	    set catType $itk_option(-catType)
	    
	    # Redraw slits if reloading ODF.
	    if {($catType == 2 || $catType == 4) && $plotSlits_ == 1} {
		
		# Clear slits on canvas.
		set target_canvas_ .skycat1.image.imagef.canvas
		$target_canvas_ delete slitMark
		
		$w_.slits configure -text "Plot Slits "
		
		set plotSlits_ 0
		
		# Change values here so method slits will redraw our slits.
		#set plotSlits_ 0
		
		#slits
	    }
	    
	    # Get the plot symbols flag from the search widget.
	    set plotSymbols [$this get_plot_symbols ]
	    
	    $w_.plot configure -text "Clear Symbols"
	    
	    # BIZARRE!!! It appears that vmAstroCat::set_plot_symbols() and vmSkySearch::set_plot_symbols()
	    # shadow each other. If invoked below as '$this set_plot_symbols 1', as was originally the case,
	    # GMMPS/skycat complains about the extra argument (because in vmAstroCat.tcl this function is
	    # defined withour arguments). But when I drop the argument, it complains about the same thing, namely
	    # a missing argument, so vmSkySearch::set_plot_symbols(arg) shadows the local function declaration.
	    # the odd thing is that vmAstroCat does not inherit from vmSkySearch, but only the other way round,
	    # so that problem should not appear here, but there.
	    # Words fail me...
	    # So I ended up just renaming the function, or alternatively put an explicit cat::vmAstroCat:: in front.

	    # Should I possibly precede all local function calls with an explicit vmAstroCat::... geeez

	    # $this set_plot_symbols 1      # wrong, does not work
	    $this skycat::vmSkySearch::set_plot_symbols 1
	    # $this cat::vmAstroCat::set_plot_symbols       # works, too
	    # $this skycat::vmSkySearch::set_plot_symbols 1    # which one is it gonna be?
	    
	    #if {$plotSymbols == 0} {
	    
	    # Tell search widget to delete its objects.
	    #	$this delete_objects
	    #}
	}


	#############################################################
	#  Name: add_search_options
	#
	#  Description:
	# add the search options panel
	#############################################################
	#############################################################

	protected method add_search_options {} {
	    # vmAstroQuery(n) widget for displaying catalog search options.
	    itk_component add searchopts {
		set searchopts_ [::cat::vmAstroQuery $w_.searchopts \
				     -relief groove \
				     -borderwidth 2 \
				     -debug $itk_option(-debug) \
				     -astrocat [code $w_.cat] \
				     -searchcommand [code $this search] \
				     -feedbackcommand [code $this set_feedback] \
				     -command [code $this query_done]]
	    }
	    pack $itk_component(searchopts) \
		-side top -fill x
	}


	#############################################################
	#  Name: add_result_table
	#
	#  Description:
	#   Create the results_ table, by calling vmQueryResult   
	#   Add the table for displaying the query results
	#############################################################
	#############################################################

	protected method add_result_table {} {
	    # QueryResult(n) widget to display catalog query results
	    #debugStatement "add_result_table: entry... "
	    
	    itk_component add results {
		set results_ [::cat::vmQueryResult $w_.results \
				  -astrocat [code $w_.cat] \
				  -title "Search Results" \
				  -hscroll 1 \
				  -height 12 \		      
			      -sortcommand [code $this set_sort_cols] \
				  -layoutcommand [code $this set_show_cols] \
				  -selectmode extended \
				  -exportselection 0]
	    } {
	    }
	    
	    pack $itk_component(results) -side top -fill both -expand 1
	    bind $results_.listbox <ButtonRelease-1> [code $this select_result_row]
	    $results_ set_options {MORE PREVIEW more preview} Show 0
	    
	}


	#############################################################
	#  Name: arch
	#
	#  Description:
	#   This function is called when the "Curved Slits" button
	#   is pressed. It creates the info necessary to a curved
	#   slit.
	#   Specifically it:
	#   	deletes archpoint.tmp file
	#   	calls vmRtdImagePick, which will 
	#		calls update
	#		binds something to someone
	#  		echo's imags2X and image2Y to arcpoint.tmp file
	#
	#############################################################
	#############################################################

	public method arch {} {
	    global image2 mybind
	    global info_ results_ myinfo myresults mythis
	    set myinfo $info_
	    set myresults $results_
	    set mythis $this
	    set catname $itk_option(-catalog)

	    set mybind .skycat1.image.pick.mainf.topf.rectf.rf.zoomView.image.imagef.canvas
	    catch {file delete [file dirname $catname]/arcpoint.tmp}
	    utilReUseWidget rtd::vmRtdImagePick .skycat1.image.pick \
		-target_image ::.skycat1.image \
		-command "" \
		-verbose 0 \
		-orient Vertical \
		-debug 0 \
		-shorthelpwin ::.skycat1
	    .skycat1.image.pick pick_object
	    update
	    bind $mybind <ButtonRelease-1> {
		exec echo $image2(X) $image2(Y) >> [file dirname $catname]/arcpoint.tmp
		.skycat1.image.pick mark_spot $image2(X) $image2(Y) image7 $mybind 90 1 1
	    }
	}

	#############################################################
	#  Name: tilt
	#
	#  Description:
	#   This function is called when the "Tilted Slits" button
	#   is pressed. It creates the info necessary to a tilted
	#   slit.
	#   Specifically it:
	#   	deletes tiltpoint.tmp file
	#   	calls vmRtdImagePick, which will 
	#		calls update
	#		binds something to someone
	#  		echo's imags2X and image2Y to tiltpoint.tmp file
	#
	#############################################################
	#############################################################

	public method tilt {} {
	    
	    $w_.results tilt_object [code $this search ]
	}

	#############################################################
	#  Name: mps_add_dialog_buttons
	#
	#  Description:
	#   This function is called when init is run for an OT catalog.
	#	It creates the dialog buttons in the object table window.
	#
	#############################################################
	#############################################################

	protected method mps_add_dialog_buttons {instType} {


	    #cba: new edit fields for editing input table rows

	    pack [frame $w_.editFrame-sp -background $ot_bg_] \
		-fill x \
		-expand 1

	    pack [frame $w_.editFrame -borderwidth 2 -relief groove -background $ot_bg_] \
		-side top \
		-fill x \
		-expand 1 \
		-in $w_.editFrame-sp \
		-padx 2 

	    pack \
		[label $w_.editlabel -text "Edit Selected Objects" -bg $ot_bg_] \
		[frame $w_.selentries -background $ot_bg_] \
		-in $w_.editFrame

	    pack [frame $w_.selSizePos -background $ot_bg_] \
		-in $w_.selentries \
		-side top

	    pack \
		[LabelEntry $w_.slitsize_y \
		     -text "Slit Size Y:" \
		     -command [code $this _sel Ssy ] \
		     -valuewidth 3 \
		     -background $ot_bg_ \
		     -anchor $itk_option(-anchor) \
		     -valuefont $itk_option(-valuefont)] \
		[LabelEntry $w_.slitsize_x \
		     -text "Slit Size X:" \
		     -command [code $this _sel Ssx ] \
		     -valuewidth 3 \
		     -background $ot_bg_ \
		     -anchor $itk_option(-anchor) \
		     -valuefont $itk_option(-valuefont)] \
		[LabelEntry $w_.slittilt \
		     -text "Slit Tilt:" \
		     -command [code $this _sel St ] \
		     -background $ot_bg_ \
		     -valuewidth 3 \
		     -anchor $itk_option(-anchor) \
		     -valuefont $itk_option(-valuefont)] \
		[LabelEntry $w_.slitpos_y \
		     -text "Slit Pos Y:" \
		     -command [code $this _sel Spy ] \
		     -valuewidth 3 \
		     -background $ot_bg_ \
		     -anchor $itk_option(-anchor) \
		     -valuefont $itk_option(-valuefont)] \
		[LabelEntry $w_.slitpos_x \
		     -text "Slit Pos X:" \
		     -command [code $this _sel Spx ] \
		     -valuewidth 3 \
		     -background $ot_bg_ \
		     -anchor $itk_option(-anchor) \
		     -valuefont $itk_option(-valuefont)] \
		[button $w_.updateslit \
		     -text "Update Row" \
		     -bg $button_bg_ot_ \
		     -fg $button_text_ot_ \
		     -activebackground $button_bg_active_ot_ \
		     -activeforeground $button_text_active_ot_ \
		     -command [code $this _sel Sall]] \
		-side left \
		-anchor e \
		-in $w_.selSizePos
	    
	    
	    # Apply colors to entry boxes of LabelEntry widgets.
	    $w_.slitsize_y.entry configure -background $ot_bg_
	    $w_.slitsize_x.entry configure -background $ot_bg_
	    $w_.slittilt.entry configure -background $ot_bg_
	    $w_.slitpos_y.entry configure -background $ot_bg_
	    $w_.slitpos_x.entry configure -background $ot_bg_

	    $w_.slitsize_y.entry configure -highlightbackground $ot_bg_
	    $w_.slitsize_x.entry configure -highlightbackground $ot_bg_
	    $w_.slittilt.entry configure -highlightbackground $ot_bg_
	    $w_.slitpos_y.entry configure -highlightbackground $ot_bg_
	    $w_.slitpos_x.entry configure -highlightbackground $ot_bg_

	    # Add help text.
	    add_short_help $w_.slitsize_x {{bitmap b1} = edit slit length of selected item (<cr> updates just this item on selection!)}
	    add_short_help $w_.slitsize_y {{bitmap b1} = edit slit width of selected item (<cr> updates just this item on selection!)}
	    add_short_help $w_.slittilt {{bitmap b1} = edit slit tilt of selected item (<cr> updates just this item on selection!)}
	    add_short_help $w_.slitpos_x {{bitmap b1} = edit slit position in X of selected item (<cr> updates just this item on selection!)}
	    add_short_help $w_.slitpos_y {{bitmap b1} = edit slit position in Y of selected item (<cr> updates just this item on selection!)}
	    add_short_help $w_.updateslit {{bitmap b1} = update selected object with values from the edit fields to the right.}
	    
	    # end of slit entries interface

	    # dialog buttons

	    pack [frame $w_.selbuttons -background $ot_bg_] \
		-side top \
		-in $w_.selentries
	    
	    set selButtonWidth 13
	    
	    #  Fill in the buttons for :
	    #  Acq. Objects, Prioirty 1, 2, and 3 objects, and Ignored objects.
	    #  Types: Rectangular & curved.
	    pack \
		[label $w_.prioLabel -text "Edit Priority:   " -background $ot_bg_ -font $itk_option(-labelfont)] \
		[button $w_.forbidden \
		     -text "Ignore Object" \
		     -width $selButtonWidth \
		     -bg $button_bg_ot_ \
		     -fg $button_text_ot_ \
		     -activebackground $button_bg_active_ot_ \
		     -activeforeground $button_text_active_ot_ \
		     -command [code $this _sel X]] \
		[button $w_.reference \
		     -text "ACQ Object - P0" \
		     -width $selButtonWidth \
		     -bg $button_bg_ot_ \
		     -fg $button_text_ot_ \
		     -activebackground $button_bg_active_ot_ \
		     -activeforeground $button_text_active_ot_ \
		     -command [code $this _sel 0]] \
		[button $w_.compulsory \
		     -text "Priority 1" \
		     -bg $button_bg_ot_ \
		     -fg $button_text_ot_ \
		     -activebackground $button_bg_active_ot_ \
		     -activeforeground $button_text_active_ot_ \
		     -width $selButtonWidth \
		     -command [code $this _sel 1]] \
		[button $w_.p3 \
		     -text "Priority 2" \
		     -bg $button_bg_ot_ \
		     -fg $button_text_ot_ \
		     -activebackground $button_bg_active_ot_ \
		     -activeforeground $button_text_active_ot_ \
		     -width $selButtonWidth \
		     -command [code $this _sel 2]] \
		[button $w_.unselect \
		     -text "Priority 3" \
		     -bg $button_bg_ot_ \
		     -fg $button_text_ot_ \
		     -activebackground $button_bg_active_ot_ \
		     -activeforeground $button_text_active_ot_ \
		     -width $selButtonWidth \
		     -command [code $this _sel 3]] \
		-side left \
		-anchor e \
		-in $w_.selbuttons

	    add_short_help $w_.reference \
		{{bitmap b1} = select an acquisition object, Priority 0}

	    add_short_help $w_.compulsory \
		{{bitmap b1} = select a most important object, Priority 1}

	    add_short_help $w_.p3 \
		{{bitmap b1} = select a second most important object, Priority 2}

	    add_short_help $w_.forbidden \
		{{bitmap b1} = select an object to be ignored}

	    add_short_help $w_.unselect \
		{{bitmap b1} = select the most command object, Priority 3}

	    add_short_help $w_.tilt \
		{{bitmap b1} = open the tilted-slit definition panel}

	    add_short_help $w_.spoc \
		{{bitmap b1} = open the slit positioning and optimization panel}

	    pack [frame $w_.slitbuttons -background $ot_bg_] \
		-side top \
		-fill x \
		-padx 2 \
		-pady 2
	    
	    # Create frames to hold Display and Mask Control buttons
	    pack \
		[frame $w_.displayCmds -borderwidth 2 -relief groove -background $ot_bg_] \
		-side left \
		-padx 20 \
		-fill x \
		-expand 1 \
		-in $w_.slitbuttons

	    pack \
		[frame $w_.controlCmds -borderwidth 2 -relief groove -background $ot_bg_] \
		-side left \
		-padx 2 \
		-ipadx 3 \
		-in $w_.slitbuttons
	    
	    # Add subframes to provide spacing. 
	    pack [frame $w_.displayCmds-sp -background $ot_bg_] \
		-padx 2 \
		-pady 4 \
		-in $w_.displayCmds
	    pack [frame $w_.controlCmds-sp -background $ot_bg_] \
		-padx 50 \
		-pady 4 \
		-ipadx 3 \
		-anchor n \
		-expand 1 \
		-fill x \
		-in $w_.controlCmds

	    # Label Display and control button groupings.
	    pack [label $w_.gmmpsFuncLabel -text "Display Options" -background $ot_bg_ ] \
		-fill x \
		-expand 1 \
		-in $w_.displayCmds-sp
	    pack [label $w_.gmmpsFuncLabel2 -text "Catalog and Mask Control" -background $ot_bg_ ] \
		-fill x \
		-expand 1 \
		-in $w_.controlCmds-sp

	    
	    # More frames, so we can have two rows of buttons each.
	    pack \
		[frame $w_.displayCmds-row1 -background $ot_bg_] \
		[frame $w_.displayCmds-row2 -background $ot_bg_] \
		-fill x  \
		-expand 1 \
		-in $w_.displayCmds-sp
	    pack \
		[frame $w_.controlCmds-col1 -background $ot_bg_] \
		[frame $w_.controlCmds-col2 -background $ot_bg_] \
		-side left \
		-padx 2 \
		-fill x \
		-expand 1 \
		-in $w_.controlCmds-sp
	    
	    # INSTRUMENT SPECIFIC PORTION
	    #
	    #  Add the Clear Slits button.
	    if {$instType == "GMOS-N" || $instType == "GMOS-S"} {
		
		set detName [lindex $detector_ 3 ]        
		
		# Set plot/clear gaps label based on plotting state:
		if {$plotGaps_ == 0} {
		    set gapsText "Plot GMOS $detName Detector Gaps "
		} else {
		    set gapsText "Clear GMOS $detName Detector Gaps"
		}
		
		pack \
		    [button $w_.plotGmosGaps \
			 -text $gapsText \
			 -width 31 \
			 -bg $button_bg_ot_ \
			 -fg $button_text_ot_ \
			 -activebackground $button_bg_active_ot_ \
			 -activeforeground $button_text_active_ot_ \
			 -command [code $this plotGmosGaps]] \
		    -side left \
		    -in $w_.displayCmds-row2
		
		pack \
		    [button $w_.defineBands \
			 -text "Design $instType N&S Mask" \
			 -bg $button_bg_ot_ \
			 -fg $button_text_ot_ \
			 -activebackground $button_bg_active_ot_ \
			 -activeforeground $button_text_active_ot_ \
			 -command [code $this define_bands] \
			 -width 22 ] \
		    -side top \
		    -in $w_.controlCmds-col2
		
		add_short_help $w_.defineBands \
		    {{bitmap b1} = Design a bandshuffle or microshuffle mask.}
		
		add_short_help $w_.plotGmosGaps \
		    {{bitmap b1} = Visualize the gaps between the GMOS detector chips}

	    } elseif {$instType == "F2"} {
		pack \
		    [button $w_.plotGmosGaps \
			 -bg $button_bg_ot_ \
			 -fg $button_text_ot_ \
			 -activebackground $button_bg_active_ot_ \
			 -activeforeground $button_text_active_ot_ \
			 -text "Plot FLAMINGOS-2 Field Layout" \
			 -command [code $this plotF2Detector]] \
		    -side left \
		    -in $w_.displayCmds-row2
		
		add_short_help $w_.plotGmosGaps \
		    {{bitmap b1} = Plot the Flamingos-2 field layout.}
		
	    }
	    
	    pack \
		[button $w_.spoc \
		     -text "Design $instType Mask" \
		     -bg $button_bg_ot_ \
		     -fg $button_text_ot_ \
		     -activebackground $button_bg_active_ot_ \
		     -activeforeground $button_text_active_ot_ \
		     -command [code $this export $instType] \
		     -width 22] \
		-side top \
		-in $w_.controlCmds-col2
	    
	    pack \
		[button $w_.reload \
		     -text "Reload Catalog" \
		     -bg $button_bg_ot_ \
		     -fg $button_text_ot_ \
		     -activebackground $button_bg_active_ot_ \
		     -activeforeground $button_text_active_ot_ \
		     -command [code $this reload_catalog ] \
		     -width 12] \
		[button $w_.close2 \
		     -text "Close Catalog" \
		     -bg $button_bg_ot_ \
		     -fg $button_text_ot_ \
		     -activebackground $button_bg_active_ot_ \
		     -activeforeground $button_text_active_ot_ \
		     -command [code $this close ] \
		     -width 12] \
		-side top \
		-in $w_.controlCmds-col1
	    
	    add_short_help $w_.spoc {{bitmap b1} = Opens the mask design window.}
	    add_short_help $w_.reload {{bitmap b1} = Reload catalog from file.}
	    add_short_help $w_.close2 {{bitmap b1} = Close this catalog.}
	    
	}


	#############################################################
	#  Name: slit_add_dialog_buttons
	#
	#  Description:
	#   This function is called when an ODF is loaded.
	#	It creates the dialog buttons.
	#
	#############################################################
	#############################################################

	protected method slit_add_dialog_buttons {instType} {
	    # dialog buttons
	    
	    pack [frame $w_.slitbuttons-sp -relief groove -background $odf_bg_ ] \
		-fill x \
		-padx 15
	    
	    pack [frame $w_.slitbuttons -borderwidth 2 -background $odf_bg_] \
		-side top \
		-fill x \
		-in $w_.slitbuttons-sp

	    #debugStatement "vmAC::sadb: intrument is $instType"
	    
	    if {$instType == "GMOS-N" || $instType == "GMOS-S"} {
		
		set detName [lindex $detector_ 3 ]
		
		# Set plot/clear bands label based on state of nod and shuffle 
		# window, if it exists. 
		set bandsOn ""
		catch {set skyct [get_skycat ]}
		catch {set bandsOn [$skyct.bandDefUI getShuffleMode ]}
		#debugStatement "bandsOn: $bandsOn"
		if {$bandsOn == "bandShuffle"} {
		    set bandsText "Clear Bands"
		    set plotBands_ 1
		} else {
		    set bandsText "Plot Bands "
		    set plotBands_ 0        		
		}
		
		# Set plot/clear gaps label based on plotting state:
		if {$plotGaps_ == 0} {
		    set gapsText "Plot GMOS $detName Detector Gaps "
		    
		} else {
		    set gapsText "Clear GMOS $detName Detector Gaps"
		    
		}
		
		pack \
		    [button $w_.slits \
			 -text "Plot Slits " \
			 -bg $button_bg_odf_ \
			 -fg $button_text_odf_ \
			 -activebackground $button_bg_active_odf_ \
			 -activeforeground $button_text_active_odf_ \
			 -command [code $this slits]] \
		    [button $w_.plotGmosGaps \
			 -text $gapsText \
			 -bg $button_bg_odf_ \
			 -fg $button_text_odf_ \
			 -activebackground $button_bg_active_odf_ \
			 -activeforeground $button_text_active_odf_ \
			 -command [code $this plotGmosGaps]] \
		    [button $w_.plotBands \
			 -text $bandsText \
			 -bg $button_bg_odf_ \
			 -fg $button_text_odf_ \
			 -activebackground $button_bg_active_odf_ \
			 -activeforeground $button_text_active_odf_ \
			 -command [code $this define_bands drawBandsOnly]] \
		    -side left -pady 2m -in $w_.slitbuttons
		
		add_short_help $w_.slits \
		    {{bitmap b1} = plot slits and spectrum overlay}
		add_short_help $w_.plotGmosGaps \
		    {{bitmap b1} = Visualize the gaps between GMOS detector chips.}
		add_short_help $w_.plotBands \
		    {{bitmap b1} = plot nod & shuffle bands, if available}
		
	    } elseif {$instType == "F2"} {
		pack \
		    [button $w_.slits \
			 -text "Plot Slits " \
			 -bg $button_bg_odf_ \
			 -fg $button_text_odf_ \
			 -activebackground $button_bg_active_odf_ \
			 -activeforeground $button_text_active_odf_ \
			 -command [code $this slits]] \
		    [button $w_.plotGmosGaps \
			 -text "Plot FLAMINGOS-2 Field Layout" \
			 -bg $button_bg_odf_ \
			 -fg $button_text_odf_ \
			 -activebackground $button_bg_active_odf_ \
			 -activeforeground $button_text_active_odf_ \
			 -command [code $this plotF2Detector]] \
		    -side left -pady 2m -in $w_.slitbuttons
		
		add_short_help $w_.slits \
		    {{bitmap b1} = plot slits and spectrum overlay}     
		add_short_help $w_.plotGmosGaps \
		    {{bitmap b1} = plot F2 field layout}
	    }
	    
	    pack \
		[button $w_.reload \
		     -text "Reload Catalog" \
		     -bg $button_bg_odf_ \
		     -fg $button_text_odf_ \
		     -activebackground $button_bg_active_odf_ \
		     -activeforeground $button_text_active_odf_ \
		     -command [code $this reload_catalog ]] \
		[button $w_.close2 \
		     -text "Close Catalog" \
		     -bg $button_bg_odf_ \
		     -fg $button_text_odf_ \
		     -activebackground $button_bg_active_odf_ \
		     -activeforeground $button_text_active_odf_ \
		     -command [code $this close]] \
		-side left -in $w_.slitbuttons 

	    add_short_help $w_.reload \
		{{bitmap b1} = Reload catalog from file.}
	    add_short_help $w_.close2 \
		{{bitmap b2} = Close catalog window.}
	    
	}


	#############################################################
	#  Name: add_dialog_buttons
	#
	#  Description:
	#   This function is called to add buttons to? 
	#   Specifically: Search, Get image, 
	#	It creates the dialog buttons.
	#
	#############################################################
	#############################################################

	protected method add_dialog_buttons {instType} {
	    # dialog buttons
	    #@@cba there is debug puts on next line
	    #debugStatement "vmAC::add_dialog_buttons: Instrument Type is $instType"
	    # should I add a button to explain the instrument or anything?  @@F2
	    #	pack [frame $w_.buttons -borderwidth 2 -relief raised] \
							    #	    -side top -fill x
	    
	    if {$iscat_} {
		set search_label "Search"
	    } else {
		set search_label "Get Image"
	    }
	    
	    button $w_.search \
		-text $search_label \
		-command [code $this search] 
	    #  -side left -expand 1 -pady 2m -in $w_.buttons
	    
	    if {$iscat_} {
		# pack
		button $w_.more \
		    -text "More Info" \
		    -state disabled \
		    -command [code $this more]
		button $w_.preview \
		    -text "Preview" \
		    -state disabled \
		    -command [code $this preview]
		#		-side left -expand 1 -pady 2m -in $w_.buttons
	    }
	    
	    #	pack 
	    button $w_.stop \
		-text "Stop" \
		-state disabled \
		-command [code $this interrupt]
	    button $w_.close \
		-text "Close" \
		-command [code $this close]
	    #-side left -expand 1 -pady 2m -in $w_.buttons
	    
	    # add a help button, if there is a help URL
	    if {"[$w_.cat help]" != ""} {
		pack \
		    [button $w_.help \
			 -text "Help" \
			 -command [code $this help]] \
		    -side left -expand 1 -pady 2m -in $w_.buttons -before $w_.close 
	    }
	}


	#############################################################
	#  Name: conv
	#
	#  Description:
	#   This function adds the x_ccd and y_ccd columns to the
	#   catalog, if they don't exist.  It converts using RA,
	#   DEC, A_WORLD, B_WORLD, THETHA_WORLD.
	#   It is called in search when x_ccd and y_ccd were not found in
	#   the original catalog file.
	#   It will essentially convert a standard catalog that contains the
	#   columns:
	#      ALPHA_SKY, DEC, A_WORLD, B_WORLD, THETA_WORLD
	#   To have the following additional columns.  (These columns are 
	#   calculated):
	#     x_ccd y_ccd A_IMAGE B_IMAGE THETA_IMAGE priority slittype
	#
	#############################################################
	#############################################################
	#*********************8 WARNING, this will never work because
	#*********************8 we do not guarentte this columns being there.
	#******************************
	#************* NEVER CALLED
	public method conv {name} {

	    #
	    #  Set initial values of variables.
	    #
	    #  First get the position of the columns of the mandatory
	    #  columns out of the headings.  Specifically:
	    #  RA, DEC, A_WORLD, B_WORLD, THETA_WORLD
	    #
	    #  locations of all the heads
	    #  headings into $head

	    set head [::cat::.vmastrocat headings]
	    #debugStatement "conv:  entering conv, getting headings:$head "
	    set RA [lsearch $head RA]
	    set DEC [lsearch $head DEC]
	    set A_WORLD [lsearch $head A_WORLD]
	    set B_WORLD [lsearch $head B_WORLD]
	    set THETA_WORLD [lsearch $head THETA_WORLD]

	    set pos {}
	    set radius {0.0 10.0}
	    set nrows {10000}
	    set searchcols 0
	    set minvalues 0
	    set maxvalues 0
	    set nsearchcols 0

	    #
	    # Set the buffer equal to output from "gmmps_sel" call
	    # Pass in filename, and everything else is left open so that
	    # all objects are retrieved - and store this in the buffer array.
	    # This array is essentially all the items in the file that
	    # are selected items.  All columns are in the same order.
	    # 

	    set pi 3.141592654

	    #	set buffer [eval exec gmmps_sel $name $pos $radius $nrows $nsearchcols $searchcols $minvalues $maxvalues]
	    if {[catch {set buffer [eval exec gmmps_sel $name $pos $radius $nrows $nsearchcols $searchcols $minvalues $maxvalues]} ] result } {
		error_dialog "Can't execut gmmps_sel, $result"
	    }
	    set nobj [llength $buffer]

	    #
	    #  Open a file called name.coord for writing.
	    set tmp [open $name.coord w]

	    #	For each object in the catalogue
	    #	
	    for {set n 0} {$n < $nobj} {incr n 1} {
		#
		#  Save the row into variable row.  And get out the
		#  ra and dec (they are in the RA and DEC
		#  positions.
		set row [lindex $buffer $n]
		set ra [lindex $row $RA]
		set dec [lindex $row $DEC]

		#
		#  Get the ra, dec coverted to wcs and degress's and save
		#  in variables coord.
		set coord [$this convert_coords $ra $dec wcs deg]
		set ra [lindex $coord 0]
		set dec [lindex $coord 1]

		#
		#  Get the ra, dec coverted to degree's on the image and
		#  in variables coord.
		set coord [$this convert_coords $ra $dec deg image]
		set x [lindex $coord 0]
		set y [lindex $coord 1]

		#
		#  Save the Theta world, and A_WORLD
		set teta [lindex $row $THETA_WORLD]
		set A [lindex $row $A_WORLD]

		#
		#  Determine what ra1 and dec1 is based on A_WORLD and THETA_WORLD
		set ra1 [expr $A*cos($teta*$pi/180.0) + $ra]
		set dec1 [expr $A*sin($teta*$pi/180.0) + $dec]

		#
		#  Now try to convert this ra1 and dec1 to degree's on the image.
		#  If it fails catch it and set coord to zero's.
		if {[catch {set coord [$this convert_coords $ra1 $dec1 deg image]}]} {
		    set coord {0 0}
		}

		#
		# Set X1 and Y1 to these converted co-ordinates.
		set x1 [lindex $coord 0]
		set y1 [lindex $coord 1]

		#
		#  Save B_WORLD, calculate ra2 and dec2 based on it.
		set B [lindex $row $B_WORLD]
		set ra2 [expr $B*cos((90+$teta)*$pi/180.0) + $ra]
		set dec2 [expr $B*sin((90+$teta)*$pi/180.0) + $dec]

		#
		#  Now try to convert this ra1 and dec1 to degree's on the image.
		#  If it fails catch it and set coord to zero's.
		if {[catch {set coord [$this convert_coords $ra2 $dec2 deg image]}]} {
		    set coord {0 0}
		}

		#
		# Set X2 and Y2 to these converted co-ordinates.
		set x2 [lindex $coord 0]
		set y2 [lindex $coord 1]

		#
		#  Calculate Aimage, Bimage, and Theta Image to be 
		#  Aim=  sqrt of ( (x1-x)^2 + (y1-y)^2 )
		#  Bim=  sqrt of ( (x2-x)^2 + (y2-y)^2 )
		#  tetaim = atan( y1-y/x1-x * 180/pi
		set Aim [expr sqrt(pow(($x1-$x),2)+pow(($y1-$y),2))]
		set Bim [expr sqrt(pow(($x2-$x),2)+pow(($y2-$y),2))]
		if {[catch {set tetaim [expr atan(($y1-$y)/($x1-$x))*180/$pi]}]} {
		    set tetaim 0
		}

		#
		#  Format x, y, Aim, Bim, tetaim and write to file $tmp
		#  including "S"elected and "R"ectangular.
		set x [format "%7g" $x]
		set y [format "%7g" $y]
		set Aim [format "%7g" $Aim]
		set Bim [format "%7g" $Bim]
		set tetaim [format "%7g" $tetaim]

		puts $tmp "$x	$y	$Aim	$Bim	$tetaim	S   R"
	    }
	    ::close $tmp

	    #
	    #  Calculate the number of lines of objects in the catalog by
	    #  finding the underline line, cut of the line number, and 
	    #  subtract that line number from the total number of lines in the
	    #  file.
	    set buffer [exec grep -n "\\--" $name]
	    set headlen [exec echo $buffer | cut -d: -f1]
	    set ncol [exec echo $buffer | wc -w]

	    #
	    #  Open a file for writing called name.col
	    #  Put empty strings into it until you get to the
	    #  header line.
	    set col [open $name.col w]
	    for {set n 3} {$n <= $headlen} {incr n 1} {
		puts $col ""
	    }


	    #
	    #  Now add some extra column headings to it.  See below.
	    puts $col "x_ccd	y_ccd	A_IMAGE	B_IMAGE	THETA_IMAGE	PRIOITY slittype"
	    puts $col "-----	-------	-------	-------	-----------	------- --------"
	    ::close $col

	    #
	    #  Now add to that file  all the extra columns that we wrote to
	    #  the file name.coord in the if statement above.
	    #  Then paste together the name file and the name.col file into 
	    #  a name.tmp file.  Then re-name the name.tmp file to name.
	    exec cat $name.coord >> $name.col
	    exec paste $name $name.col > $name.tmp
	    catch {file delete $name.col}
	    catch {file delete $name.coord}
	    file rename -force $name.tmp $name
	}


	#############################################################
	#  Name: updateCat
	#
	#  Description:
	#   This function is used to write the catalog file, due to updates.
	#   FINISH WRITING THIS LATER.
	#
	#############################################################
	#############################################################
	public proc updateCat { name } {

	    #
	    #  Calculate the number of lines of objects in the catalog by
	    #  finding the underline line, cut of the line number, and 
	    #  subtract that line number from the total number of lines in the
	    #  file.
	    set buffer [exec grep -n "\\--" $name]
	    set headlen [exec echo $buffer | cut -d: -f1]
	    set ncol [exec echo $buffer | wc -w]

	    #
	    #  Open a file for writing called name.col
	    #  Put empty strings into it until you get past the
	    #  header line.
	    set col [open $name.tmps w]
	    for {set n 1} {$n <= $headlen} {incr n 1} {
		puts $col ""
	    }

	    #
	    #  Now cycle thru the and write the file.

	}


	#############################################################
	#  Name: symbol
	#
	#  Description:
	#   This function is used to retrieve from the image information
	#   and to write that information, including the symbol information
	#   to the catalog file.
	#
	#############################################################
	#############################################################
	public proc symbol {name} {

	    # debugStatement "\nproc symbol: Entry. catname = $name"
	    

	    #
	    #  Open a file name.tmp for appending.
	    #  Put in:
	    #    the query results, 
	    #    how to draw the symbols
	    #    also fits keyword information that we extracted above.
	    #    Finish with an "End config entry" line.
	    #
	    set tmp [open $name.tmp a]
	    puts $tmp "QueryResult"
	    puts $tmp ""
	    puts $tmp "# Config entry for original catalog server:"
	    puts $tmp "serv_type: local"
	    puts $tmp "long_name: $name"
	    puts $tmp "short_name: [file tail $name]"
	    puts $tmp "url: $name"
	    puts $tmp "symbol: \{x_ccd y_ccd priority\} \{circle white \{\} \{\} \{\} \{\$priority == \"3\"\}\} \{15 \{\}\} : \{x_ccd y_ccd priority\} \{triangle blue \{\} \{\} \{\} \{\$priority == \"1\"\}\} \{15 \{\}\} : \{x_ccd y_ccd priority\} \{diamond cyan \{\} \{\} \{\} \{\$priority == \"0\"\}\} \{20 \{\}\} : \{x_ccd y_ccd priority\} \{cross yellow \{\} \{\} \{\} \{\$priority == \"X\"\}\} \{15 \{\}\} : \{x_ccd y_ccd priority\} \{square green \{\} \{\} \{\} \{\$priority == \"2\"\}\} \{15 \{\}\}"
	    puts $tmp "# Fits keywords"
	    set buffer ""
	    catch { set buffer [exec grep #fits $name ] }
				puts $tmp "$buffer"
				puts $tmp "# End fits keywords"
				puts $tmp "# Curved slits"
				puts $tmp "# End curved slits"
				puts $tmp "# End config entry"
				::close $tmp

				#
				#  Grep in the catalog file the line "End config entry" and get
				#  the line number.  Which will give you the line where the
				#  headers and data actually starts.
				#  Then append to the above written name.tmp file all of the header
				#  and data out of the original configuration file.  
				#  Then move name.tmp to name.
				set n [exec grep -n "End config entry" $name | cut -d: -f1]
				#puts $n,"[]",$name
				set n [expr $n + 1]

				exec tail -n +$n $name >> $name.tmp
				catch {file delete $name}
				file rename -force $name.tmp $name

				#debugStatement "End proc symbol"
			    }


		#############################################################
		#  Name: col_PRIO
		#
		#  Description:
		#   Add the priority column to the table
		#
		#  Args:
		#    name of the catalog file.
		#############################################################
		#############################################################

		public proc col_PRIO {name} {

		    #
		    # buffer:  Grep for underlines in the catalog.  
		    # headlen: Get that line number of where the headers end.
		    # ncol:    Get the number of sets of unlines in the catalog, or number
		    #            of column's in the catalog.
		    # tmp:	   Get the number of lines in the catalog
		    # datalen: Get number of data lines (tmp-headlen)
		    set buffer [exec grep -n "\\--" $name]
		    set headlen [exec echo $buffer | cut -d: -f1]
		    set ncol [exec echo $buffer | wc -w]
		    set tmp [exec wc -l $name]
		    set datalen [expr [lindex $tmp 0] - $headlen]

		    #
		    #  Open a file for writing called name+.col
		    #  Write in there empty lines, up to where the data starts.
		    set col [open $name.col1 w]
		    for {set n 3} {$n <= $headlen} {incr n 1} {
			puts $col ""
		    }
		    #
		    # Put in name+.col the header for priority
		    puts $col "priority"
		    puts $col  "--------"

		    #
		    # Put in name+.col a column of "3" 's, for selected items.
		    for {set n 1} {$n <= $datalen} {incr n 1} {
			puts $col "3"
		    }
		    ::close $col

		    #
		    #  Mush the 2 files together to get the new column in there.
		    #  Then rename the name.tmp to name
		    exec paste $name $name.col1 > $name.tmp
		    catch {file delete $name.col1}
		    file rename -force $name.tmp $name
		}


		#############################################################
		#  Name: col_ST
		#
		#  Description:
		#   Add the slittype column to the table
		#
		#  Args:
		#    name of the catalog file.
		#############################################################
		#############################################################


		public proc col_ST {name} {
		    #debugStatement "col_ST"
		    #
		    # buffer:  Grep for underlines in the catalog.  
		    # headlen: Get that line number of where the headers end.
		    # ncol:    Get the number of sets of unlines in the catalog, or number
		    #            of column's in the catalog.
		    # tmp:	   Get the number of lines in the catalog
		    # datalen: Get number of data lines (tmp-headlen)
		    set buffer [exec grep -n "\\--" $name]
		    set headlen [exec echo $buffer | cut -d: -f1]
		    set ncol [exec echo $buffer | wc -w]
		    set tmp [exec wc -l $name]
		    set datalen [expr [lindex $tmp 0] - $headlen]

		    #
		    #  Open a file for writing called name+.col
		    #  Write in there empty lines, up to where the data starts.
		    set col [open $name.col2 w]
		    for {set n 3} {$n <= $headlen} {incr n 1} {
			puts $col ""
		    }
		    #
		    # Put in name+.col the header for slittype
		    puts $col "slittype"
		    puts $col  "--------"

		    #
		    # Put in name+.col a column of "R"'s for rectangular.
		    for {set n 1} {$n <= $datalen} {incr n 1} {
			puts $col "R"
		    }
		    ::close $col
		    #debugStatement "stop here to inspect *.col2"
		    #
		    #  Mush the 2 files together to get the new column in there.
		    #  Then rename the name.tmp to name
		    set mx [sleep 1]
		    set mx [exec paste $name $name.col2 > $name.tmp]

		    # debugStatement "exec said $mx"
		    catch {file delete $name.col2}
		    file rename -force $name.tmp $name
		}

		#############################################################
		#  Name: col_STSIZE
		#
		#  Description:
		#   Add the slitsize_y column to the table
		#
		#  Args:
		#    name of the catalog file.
		#############################################################
		#############################################################

		public proc col_STSIZE {name} {

		    #
		    # buffer:  Grep for underlines in the catalog.  
		    # headlen: Get the line number of where the headers end.
		    # ncol:    Get the number of sets of unlines in the catalog, or number
		    #            of column's in the catalog.
		    # tmp:	   Get the number of lines in the catalog
		    # datalen: Get number of data lines (tmp-headlen)
		    set buffer [exec grep -n "\\--" $name]
		    set headlen [exec echo $buffer | cut -d: -f1]
		    set ncol [exec echo $buffer | wc -w]
		    set tmp [exec wc -l $name]
		    set datalen [expr [lindex $tmp 0] - $headlen]
		    #debugStatement "col_STSIZE: headlen=$headlen, ncol=$ncol, tmp=$tmp datalen=$datalen.."

		    #
		    #  Open a file for writing called name+.col
		    #  Write in there empty lines, up to where the data starts.
		    set col [open $name.col3 w]
		    for {set n 3} {$n <= $headlen} {incr n 1} {
			puts $col ""
		    }

		    #
		    # Put in name+.col the header for slitsize_y
		    puts $col "slitsize_y"
		    puts $col  "----------"

		    #
		    # Put in name+.col a column of "1.08" arcsec (15pixels)'s for the slit length.
		    for {set n 1} {$n <= $datalen} {incr n 1} {
			puts $col "5.0"
		    }
		    ::close $col

		    #
		    #  Mush the 2 files together to get the new column in there.
		    #  Then rename the name.tmp to name
		    exec paste $name $name.col3 > $name.tmp
		    catch {file delete $name.col3}
		    file rename -force $name.tmp $name
		}


		#############################################################
		#  Name: export
		#
		#  Description:
		#   This function is called when the "Design Mask" button
		#   is pressed. It creates the Master ODF file(s). 
		#   It calls the gmmps_spoc method.  Which will bring up
		#	  the popup to allow you to run the spoc algorithm.
		#
		#############################################################
		#############################################################

		public method export {instType} {
		    #
		    #  Determine if slitpos_x and slitpos_y exist in the
		    #  catalog files.
		    set addPX ""
		    set addPY ""
		    catch {set addPX [exec grep slitpos_x $itk_option(-catalog) ]}
		    catch {set addPY [exec grep slitpos_y $itk_option(-catalog) ]}
		    if { $addPX == "" } {
			set addPX "yes" 
		    } else {
			set addPX "no" 
		    }
		    if { $addPY == "" } {
			set addPY "yes" 
		    } else {
			set addPY "no" 
		    }
		    
		    # Make sure slit tilts are valid.
		    if {[check_tilts -45.0 45.0 ]} {
			return
		    }
		    
		    # Find detector info.
		    setDetInfo $instType
		    
		    # destroy any open spoc window.
		    catch {destroy $spoc_}
		    
		    #
		    #  Call the gmmps_spoc.tcl init function
		    #debugStatement "vmAC::export: opening gmmps_spoc interface for instrument $instType"
		    set spoc_ [utilReUseWidget cat::gmmps_spoc .skycat1.spoc $this $itk_option(-catalog) $itk_option(-xsize) $itk_option(-tilt) $addPX $addPY $itk_option(-pixs) $itk_option(-bw) $this.results $instType $config_file_ $DET_IMG_ $DET_SPEC_ ]
		    
		}


		#############################################################
		#  Name: check_tilts
		#
		#  Description:
		#   Makes sure there are no invalid slittilts in the catalog. 
		#############################################################
		#############################################################

		protected method check_tilts {min max} {
		    #
		    # Make sure there are no invalid slittilts.
		    set rows [$this.results get_contents ]
		    set columns [$this.results get_headings ]
		    set tilt_col [lsearch  -regex $columns {(?i)slittilt} ]
		    set id_col [lsearch -regex $columns {(?i)ID} ]
		    set prio_col [lsearch -regex $columns {(?i)priority} ]
		    set bad_tilt_rows [list ]
		    
		    # No slit tilts is good slit tilts.
		    if {$tilt_col == -1} {
			return 0
		    }
		    
		    set cnt 0
		    set changed 0
		    foreach row $rows {
			set tilt [format %.6f [lindex $row $tilt_col ]]
			
			if {$tilt > $max || $tilt < $min} {
			    set id [lindex $row $id_col ]
			    
			    set choice [choice_dialog "Slittilt $tilt for object $id is outside valid range ($min to $max degrees)." {{Set Priority to Ignore} {Reset Angle to 0.0} Cancel} {Set Priority to Ignore} $w_ ]
			    puts "Slittilt for object $id is invalid ($tilt)."
			    
			    if {$choice == "Cancel"} {
				
				puts "Make Mask cancelled."
				return 1
				
			    } elseif {$choice == "Reset Angle to 0.0" } {
				
				set newrow [lreplace $row $tilt_col $tilt_col 0.000000 ]
				$this.results set_row $row $newrow
				set changed 1
				
			    } elseif {$choice == "Set Priority to Ignore" } {
				
				set newrow [lreplace $row $prio_col $prio_col "X" ]
				$this.results set_row $row $newrow
				set changed 1
			    }
			    
			    incr cnt
			}
		    }
		    
		    # Save changes to file.
		    if {$changed} {
			$this.results save_to_file $itk_option(-catalog) [$this.results get_contents ] "" 1
		    }
		    
		    return 0
		}


		#############################################################
		#  Name: disable_search
		#
		#  Description:
		#   Disable the search button, if the arg is 1, otherwise enable it
		#############################################################
		#############################################################

		public method disable_search {{arg 1}} {
		    if {$arg} {
			set search_state_ disabled
		    } else {
			set search_state_ normal
		    }
		}


		#############################################################
		#  Name: help
		#
		#  Description:
		#   If the catalog has a "help" URL, try to display it in 
		#   netscape using the netscape remote interface.
		#############################################################
		#############################################################

		public method help {} {
		    set url [$w_.cat help]
		    if {"$url" != ""} {
			send_to_netscape $url
		    } else {
			info_dialog "Sorry, no help is available for this catalog"
		    }
		}


		#############################################################
		#  Name: add_progress_bar
		#
		#  Description:
		#   Add a progress bar to display the progress of data transfers
		#############################################################
		#############################################################

		protected method add_progress_bar {} {
		    # add a progress bar at the botton
		    #	pack [ProgressBar $w_.progress] \
							  #	    -side top -fill x
		}


		#############################################################
		#  Name: layout_dialog
		#
		#  Description:
		#   Do the dialog window layout.  Specifically:
		#     Open the catalog
		#     Add the menu bar and search options,
		#     Get the named servers
		#     create a results_ table 
		#     Add the dialog buttons and progress bar
		#############################################################
		#############################################################

		protected method layout_dialog {} {
		    set instType $itk_option(-instType)
		    set catType  $itk_option(-catType)

		    # debugStatement "vmAC:layout_dialog instType = $instType"

		    open_catalog
		    add_menubar
		    add_search_options
		    get_name_servers
		    
		    # If not an image server, then assume its our type.
		    if {$iscat_} {
			add_result_table
		    }
		    
		    add_dialog_buttons $instType
		    add_progress_bar
		    
		    
		    # Add bindings for GUI editing.
		    # Load colorscheme.
		    if {$catType == 1 || $catType == 3} {
			# load OT colorscheme.
			set bg $ot_bg_
			set bg2 $ot_bg2_
			set fg_text $ot_text_
			set catalog_bg_active $catalog_bg_active_ot_
			set catalog_text_active $catalog_text_active_ot_
			set catalog_bg $catalog_bg_ot_
			set catalog_text $catalog_text_ot_
			set button_bg $button_bg_ot_
			set button_text $button_text_ot_
			set button_bg_active $button_bg_active_ot_
			set button_text_active $button_text_active_ot_
			set menu_bg $menu_bg_ot_
			set menu_text $menu_text_ot_
			set menu_bg_active $menu_bg_active_ot_
			set menu_text_active $menu_text_active_ot_
		    } else {
			# load ODF colorscheme.
			set bg $odf_bg_
			set bg2 $odf_bg2_
			set fg_text $odf_text_
			set catalog_bg_active $catalog_bg_active_odf_
			set catalog_text_active $catalog_text_active_odf_
			set catalog_bg $catalog_bg_odf_
			set catalog_text $catalog_text_odf_
			set button_bg $button_bg_odf_
			set button_text $button_text_odf_
			set button_bg_active $button_bg_active_odf_
			set button_text_active $button_text_active_odf_	
			set menu_bg $menu_bg_odf_
			set menu_text $menu_text_odf_
			set menu_bg_active $menu_bg_active_odf_
			set menu_text_active $menu_text_active_odf_
		    }

		    # Change tables settings here because the widget doesn't seem to want to 
		    # listen to args we pass in.  
		    $results_ configure \
			-height 18 \
			-width 80 \
			-background $bg \
			-foreground $fg_text \
			-selectbackground $catalog_bg_active \
			-selectforeground $catalog_text_active \
			-activebackground $button_bg_active

		    $results_.listbox configure \
			-background $catalog_bg \
			-foreground $catalog_text
		    $results_.headbox configure \
			-background $catalog_bg \
			-foreground $catalog_text \
			-selectbackground $catalog_bg \
			-selectforeground $catalog_text
		    $results_.tf configure \
			-background $bg
		    $results_.vf.vscroll configure  \
			-troughcolor $bg2 \
			-background $button_bg
		    $results_.hf.hscroll configure  \
			-troughcolor $bg2 \
			-background $button_bg
		    
		    #  Change menu colors here. 
		    $w_.menubar configure \
			-background $menu_bg \
			-highlightbackground $menu_bg_active\
			-highlightcolor $menu_text_active
		    
		    $w_.menubar.file configure \
			-background $menu_bg \
			-foreground $menu_text \
			-activebackground $menu_bg_active\
			-activeforeground $menu_text_active 
		    
		    $w_.menubar.edit configure \
			-background $menu_bg \
			-foreground $menu_text \
			-activebackground $menu_bg_active\
			-activeforeground $menu_text_active 


		    $w_.menubar.options configure \
			-background $menu_bg \
			-foreground $menu_text \
			-activebackground $menu_bg_active\
			-activeforeground $menu_text_active
		    
		    
		    # Apply custom color scheme.
		    #apply_color_scheme catType
		}

		#############################################################
		#  Name: add_canvas_bindings
		#
		#  Description:
		#   GUI canvas movement tags.
		#############################################################
		#############################################################

		protected method add_canvas_bindings {} {

		    # Move slit bindings for Y direction. 
		    $target_canvas_ bind "$idTag-Move-Y" <1> [code $this moveSlit $idTag "Y" %x %y "in" $objectData]
		    $target_canvas_ bind "$idTag-Move-Y" <ButtonRelease-1> [code $this moveSlit $idTag "Y" %x %y "out" $objectData]
		    $target_canvas_ bind "$idTag-Move-Y" <Motion> [code $this moveSlitAnimate "Y" %x %y $idTag]
		    
		    # Move slit bindings for X direction. 
		    $target_canvas_ bind "$idTag-Move-X" <1> [code $this moveSlit $idTag "X" %x %y "in" $objectData]
		    $target_canvas_ bind "$idTag-Move-X" <ButtonRelease-1> [code $this moveSlit $idTag "X" %x %y "out" $objectData]
		    $target_canvas_ bind "$idTag-Move-X" <Motion> [code $this moveSlitAnimate "X" %x %y $idTag]



		}




		#############################################################
		#  Name: apply_color_scheme
		#
		#  Description:
		#   Applies the custom color scheme defined in protected variables.
		#   (note there is also some colorscheme code scattered about)
		#############################################################
		#############################################################

		protected method apply_color_scheme {catType} {



		}





		#############################################################
		#  Name: make_short_help
		#
		#  Description:
		#   Add a short help window and set the help texts
		#############################################################
		#############################################################

		protected method make_short_help {} {
		    TopLevelWidget::make_short_help
		    
		    add_short_help $results_ \
			{Query results: {bitmap b1} = select object, \
			     double click {bitmap b1} = label object, \
			     {bitmap dragb2} = scroll list}
		    
		    if {$iscat_} {
			add_short_help $w_.search \
			    {{bitmap b1} = start catalog search}
			
			add_short_help $w_.preview \
			    {{bitmap b1} = view the preview image or plot data for the selected object}
			
			add_short_help $w_.more \
			    {{bitmap b1} = display more information for the selected object}
			
		    } else {
			add_short_help $w_.search \
			    {{bitmap b1} = fetch the image from the image server}
		    }
		    
		    add_short_help $w_.stop \
			{{bitmap b1} = interrupt the current catalog operation}
		    
		    if {[winfo exists $w_.help]} {
			add_short_help $w_.help \
			    {{bitmap b1} = Display help information for catalog in netscape}
		    }
		    
		    add_short_help $w_.close {{bitmap b1} = close this window}
		    
		    #	add_short_help $w_.progress {Progress bar: displays status of work in progress}

		}


		#############################################################
		#  Name: select_catalog
		#
		#  Description:
		#   This proc is called when the named catalog selected from the menu.
		#   It will check for specific column headings, set flags if the PRIORITY
		#   slittype, and X/Y column headings are missing. It will then open
		#   the catalog and save an instance of it.
		#
		#  Args:
		#   name is the name of the catalog file.
		#
		#   serv_type is the type of the catalog: one of: catalog, archive, 
		#   local, imagesvr, etc...
		#
		#   id is an optional unique id to be associated with a new catalog widget.
		#	Not used at all?
		#
		#   classname is the name of the vmAstroCat subclass to use to create new
		#   catalog widgets (defaults to "vmAstroCat").
		#
		#   $debug is a flag set from the command line arg.
		#
		#   $w should be the top level window of the caller (optional).
		#############################################################
		#############################################################

		public proc select_catalog {name serv_type id classname {debug 0} tType {w ""} instType} {
		    #
		    #  Declare global var.s
		    ############ WARNING, have to attached these global variables to a specific
		    ############ window.  Otherwise, when I open another window where slitsize_x
		    ############ is there, and then go back to the original window, where it isn't
		    ############ addXSIZE will be invalid.
		    #
		    global acDebugLevel
		    set acDebugLevel 0
		    
		    #debugStatement "vmAC: select_catalog"
		    #debugStatement "select_catalog: entry..... tType=$tType"

		    set addSIZE "no"
		    set addXSIZE "no"
		    set addTILT "no"
		    set addPRIORITY "no"
		    set addSLITTYPE "no"
		    set itk_option(-prior) "no"
		    set itk_option(-xsize) "no"
		    set itk_option(-ysize) "no"
		    set itk_option(-tilt) "no"
		    set itk_option(-stype) "no"
		    set itk_option(-catType) $tType
		    set gmos_cat ""
		    set col_SIZE ""
		    set col_XSIZE ""
		    set col_PRIORITY ""
		    set col_SLITTYPE ""
		    set col_XSLIT ""
		    set col_TILT ""
		    set raDecCol_s ""
		    set xyCol_s ""
		    set tmpHeadings ""
		    
		    #  This type of file should be a FITS file, read it
		    #  in and convert to a catalog file.
		    #debugStatement "vmAC:select_catalog tType=$tType"
		    if { $tType == 3 || $tType == 4 } {
			#debugStatement "select_catalog tType=$tType"
			set index1 [string last . $name ]
			if { $index1 != -1 } {
			    set index1 [expr $index1 - 1 ]
			    set inName [string range $name 0 $index1 ]
			} else {
			    set inName $name
			}
			#debugStatement "name=$name index1=$index1 inName = $inName"
			#debugStatement "#About to call readInFitsFile"
			set x -1
			if {[catch {set x [ readInFitsFile $inName $tType]} msg]} {
			    #debugStatement "After exec3."
			    error_dialog $msg 
			    return 
			} else {
			    if { $x == -1 } {
				#debugStatement "select_catalog: fits file exit status: $x. Failed."
				return
			    } else {
				#debugStatement "select_catalog: fits conversion exit status: $x. Passed."
				set name [append inName ".cat"]
				#debugStatement "#new name=$name"
				# @@cba: get the image info in the cat file!
				symbol $name
			    }
			}
		    } else {
			#
			#  ELSE this should be a catalog file, grep for "QueryResult"
			#  should be there for a catalog file.
			set catCheck ""
			catch {set catCheck [ exec grep QueryResult $name ] }
			if { $catCheck == "" } {
			    error_dialog "Error, not an ascii catalog file."
			    return 
			}
		    }
		    
		    # Set the pixel scale.
		    set pixret [getPixelScale ]
		    set pixScale [lindex $pixret 0 ]
		    set detInst  [lindex $pixret 1 ]
		    #debugStatement "Pixelscale found in select_catalog: $pixScale"
		    
		    # Set instrument type if it is not already set.
		    if {$instType == "NONE"} {
			set instType [getInstType $name $pixScale $detInst ]
			
			if {$instType == 1} {
			    error_dialog "Invalid preimage file. Cannot load catalog."
			    return
			}
			
		    }
		    
		    #debugStatement "vmAC:select_catalog after file open"

		    # Also, rewrite file making sure the catalog data is tab-separated. 
		    #  Get the column headings and save them.
		    set buffer [exec  grep -n "\\--" $name]
		    set headlen [exec echo $buffer | cut -d: -f1]
		    set ncol [exec echo $buffer | wc -w]

		    set cat [open $name r ]
		    set cattmp [open $name.tmp w ]

		    for {set n 1} {$n <= $headlen} {incr n 1} {
			gets $cat line
			puts $cattmp $line
		    }
		    
		    # We do no want space separated lists, we want tab separated lists.
		    while {[gets $cat line ] >= 0} {
			set line [concat $line ]
			set line [join $line "\t" ]
			puts $cattmp $line
		    }
		    
		    ::close $cat
		    ::close $cattmp

		    file rename -force $name.tmp $name
		    catch {file delete $catname.tmps}
		    
		    set itk_option(-headingx) $line
		    set headings_ $line
		    #debugStatement "**** header saved: $itk_option(-headingx) "
		    
		    #
		    #  Grep for x_ccd, y_ccd, ID, slitsize_y, priority and 
		    #  slittype in the catalog.  First 3 are mandatory.  
		    #
		    catch {set raDecCol_s [exec grep -i RA $name | grep -i DEC | grep -i MAG ]}
		    catch {set xyCol_s [exec grep -i  x_ccd $name | grep  -i y_ccd | grep -i  ID ]}
		    catch {set col_SIZE [exec grep  -i slitsize_y $name ]}
		    catch {set col_XSIZE [exec grep  -i slitsize_x $name ]}
		    catch {set col_PRIORITY [exec grep  -i priority $name ]}
		    catch {set col_SLITTYPE [exec grep  -i slittype $name ]}
		    catch {set col_XSLIT [exec grep  -i slitpos_x $name ]}
		    catch {set col_TILT [exec grep  -i slittilt $name ]}
		    catch {set tmpHeadings [exec grep ID  -i $name | grep  -i RA | grep  -i DEC | grep  -i x_ccd | grep  -i y_ccd ]}
		    #puts "col_SIZE = $col_SIZE"
		    #debugStatement "select_catalog: About to deter col's that exist, tmpHeadings=$tmpHeadings"

		    #
		    #  If we found the minimum table headings that we need then cont.
		    #

		    if {$xyCol_s == ""} {
			error_dialog "AHHHH, can't find ID, x_ccd or y_ccd, stop!!"
			return 0
			
		    } elseif {$raDecCol_s == ""} {
			error_dialog "ERROR, For now, must have RA/DEC/MAG..."
			return 0
		    } else {
			if { $tType != 2 || $tType != 4} {

			    #
			    #  If we didn't find the extra columns we need then
			    #  set the flag to indicate we need to add them.
			    #
			    if {$col_SIZE == ""} {
				#
				#  No column with name slitsize_y, so set flag to add it.
				symbol $name
				$astrocat_ entry remove $name
				cat::CatalogInfo::save {} $w 0
				$astrocat_ reload
				set addSIZE "yes"
				set itk_option(-ysize) "yes"
				# debugStatement "select_catalog: addSIZE= $addSIZE"
			    }
			    if {$col_XSIZE == ""} {
				#
				#  No column with name slitsize_x, so set flag.
				set addXSIZE "yes"
				set itk_option(-xsize) "yes"
				#  debugStatement "select_catalog: addXSIZE= $addXSIZE"
			    }
			    if {$col_TILT == ""} {
				#
				#  No column with name slittilt, so set flag.
				set addTILT "yes"
				set itk_option(-tilt) "yes"
				#debugStatement "select_catalog: addTILT= $addTILT"
			    }
			    if {$col_PRIORITY == ""} {
				#
				#  No column with name priority, so add them.
				symbol $name
				$astrocat_ entry remove $name
				cat::CatalogInfo::save {} $w 0
				$astrocat_ reload
				#debugStatement "select_catalog: no priority col"
				set addPRIORITY "yes"
				set itk_option(-prior) "yes"
				#debugStatement "select_catalog: addPRIOIRTY= $addPRIORITY"
			    }
			    if {$col_SLITTYPE == "" } {
				#
				#  No column with name slittype, so add them.
				symbol $name
				$astrocat_ entry remove $name
				cat::CatalogInfo::save {} $w 0
				$astrocat_ reload
				set addSLITTYPE "yes"
				set itk_option(-stype) "yes"
				#debugStatement "select_catalog: no slittype col"
			    }
			}
		    }

		    #
		    #  Complicated way to set the column positions of id, ra, dec.
		    #  This gets used later to set the headings for the gmmps_sel call.
		    set col_numb 0
		    set ids ""
		    set ras ""
		    set decs ""
		    
		    foreach i $tmpHeadings {
			if { $i == "ID" } { 
			    set ids "id_col $col_numb"
			} elseif {$i == "RA" } { 
			    set ras "ra_col $col_numb"
			} elseif { $i == "DEC" } { 
			    set decs "dec_col $col_numb"
			}
			incr col_numb
		    }
		    set itk_option(-uphead) [list $ids $ras $decs]
		    #debugStatement "Setting updateHeadings: $itk_option(-uphead)"
		    
		    #  Open the catalog, and create an instance of it.  
		    #debugStatement "select_catalog: about to call new_catalog"

		    # @@cba:@@test
		    #info_dialog "Loading Catalog $name"
		    #debugStatement "name = $name $classname"

		    new_catalog $name $id $classname $debug "catalog" $tType $addPRIORITY $addSLITTYPE $addTILT $addSIZE $addXSIZE $w $instType $pixScale

		    #debugStatement "select_catalog: Exit. \n"
		}


		#############################################################
		#  Name: set_namesvr
		#
		#  Description:
		#   Set the name server to use to resolve object names
		#############################################################
		#############################################################

		public method set_namesvr {name} {
		    $searchopts_ config -namesvr $name

		}


		#############################################################
		#  Name: interrupt
		#
		#  Description:
		#   Interrupt the current search 
		#############################################################
		#############################################################

		public method interrupt {} {
		    $w_.batch interrupt
		    $searchopts_ interrupt
		    set_feedback off
		    catch {$results_ config -title "Search Results"}
		    set_state normal
		}


		#############################################################
		#  Name: set_state
		#
		#  Description:
		#   Set/reset widget states while busy 
		#############################################################
		#############################################################

		public method set_state {state} {
		    set state_ $state
		    if {"$state" == "normal"} {
			catch {blt::busy release $w_.options}
			catch {focus -lastfor $w_.options}
			$w_.search config -state $search_state_
			$w_.stop config -state disabled
		    } else {
			catch {focus .}
			catch {blt::busy hold $w_.options}
			$w_.search config -state disabled
			catch {
			    $w_.more config -state disabled
			    $w_.preview config -state disabled
			}
			$w_.stop config -state normal
		    }
		    update idletasks
		    #	$w_.progress reset
		}


		#############################################################
		#  Name: feedback
		#
		#  Description:
		#   This method is called by the fileevent handler during the data transfer 
		#   from the HTTP C++ class to give the user feedback about how much 
		#   there is left to copy, etc...
		#   Read a line from the feedback pipe, which should contain text
		#   to be displayed in the progress widget.
		#
		#   When we are in busy (disabled, waiting) mode, we also 
		#   interpret some special messages, such as: "total length: n bytes" 
		#   to update the progress bar.
		#
		#############################################################
		#############################################################

		public method feedback {} {
		    set text [gets $rfd_]
		    if {"$state_" != "normal"} {
			if {[scan $text {total length: %d bytes} n] == 1} {
			    #		$w_.progress config -to $n
			} elseif {[scan $text {read %d bytes} n] == 1} {
			    #		$w_.progress config -value $n
			} elseif {[scan $text {url: %s} url] == 1} {
			    catch {
				set fd [open $logfile_name_ a+]
				puts $fd "[clock format [clock seconds]]\n$itk_option(-catalog)\n$url\n\n"
				::close $fd
			    } 
			}
		    }
		    #	$w_.progress config -text $text
		    update idletasks
		}


		#############################################################
		#  Name: set_info
		#
		#  Description:
		#   This method could be used by a plugin to set the contents of 
		#   the listing without doing a query.
		#   The arguments are the table headings (list of strings) and 
		#   the table data (list of rows, where each row is a list of values).
		#   The "more" flag is optional and indicates that there would 
		#   be more data.
		#############################################################
		#############################################################

		public method set_info {headings info {more 0}} {
		    query_done "" $headings $info $more
		}



		#############################################################
		#  Name: query_done
		#
		#  Description:
		#   This method is called when the background query is done.
		#
		#   errmsg - If this is not empty, it contains an error message and
		#   the following args should be ignored. If there were no
		#   errors, this arg is an empty string.
		#
		#   headings  - are the column headings 
		#
		#   info - is a list of rows (result of query)
		#
		#   more - is a flag set to 1 if there were more rows available that were
		#   not returned due to the maxrows limit set.
		#
		#############################################################
		#############################################################

		public method query_done {errmsg headings info more} {
		    #debugStatement "vmAC:query_done $headings"

		    if {"$errmsg" != ""} {
			# check if we need user/passwd info. errmsg should have the format:
			# "Authorization Required for <realm> at <host>"
			if {[regsub {Authorization Required} $errmsg {Enter username} msg]} {
			    lassign [passwd_dialog $msg] username passwd
			    if {"$username" != "" && "$passwd" != ""} {
				lassign $errmsg {} {} {} realm {} host
				$w_.cat authorize $username $passwd $realm $host
				$searchopts_ search
			    }
			} else {
			    set_state normal
			    catch {$results_ config -title "Search Results"}
			    # error messages starting with "***" are only 
			    # displayed in progress win
			    if {"[string range $errmsg 0 2]" != "***"} {
				error_dialog $errmsg $w_
			    }
			    #				after 0 [list $w_.progress config -text $errmsg]
			}
		    } else {
			if {! $iscat_} {
			    # for image servers, the info is the image file name
			    set filename $info
			    # load the image and remove the temp file
			    display_image_file $filename
			    catch {file delete $filename}
			} else {
			    busy {
				#	$results_ save_yview
				#	$results_ save_selection
				set prev_headings $itk_option(-headingx)
				set itk_option(-headingx) $headings
				set headings_ $headings
				set info_ $info
				# debugStatement "query_done: headings=$headings..."


				# update table
				$results_ config -headings $itk_option(-headingx)
				if {$reset_columns_} {
				    set reset_columns_ 0
				    reset_table
				}

				if {"$prev_headings" != "$itk_option(-headingx)"} {
				    $results_ update_options
				    set_menu_states
				}

				$results_ config -info $info
				
				# need to know the equinox of the results, if using world coords
				if {[$w_.cat iswcs]} {
				    lassign [$w_.searchopts get_pos_radius] {} {} equinox
				    $results_ config -equinox $equinox
				} 

				if {$more} {
				    set more "+"
				} else {
				    set more ""
				}
				$results_ config -title "Search Results ([$results_ total_rows]$more)"

				# note column indexes in array (use upper case to simplify search)
				catch {unset col_}
				set n -1
				foreach i $itk_option(-headingx) {
				    set col_($i) [incr n]
				}
				
				# plot stars
				#puts "plot time: [expr [lindex [time plot] 0]/1000000.]"
				#debugStatement "Beginning symbol plots?"				
				
				plot

				#debugStatement "Done plotting?"

				$results_ restore_yview
				$results_ restore_selection
			    }
			}
			set_state normal
		    }
		}


		#############################################################
		#  Name: search
		#
		#  Description:
		#   Start the catalog search based on the current search options
		#   and display the results in the table.
		#############################################################
		#############################################################

		public method search {args} {

		    #debugStatement "method search entry"
		    
		    
		    set addPRIORITY $itk_option(-prior)
		    set addSIZE $itk_option(-ysize)
		    set addSLITTYPE $itk_option(-stype)
		    #debugStatement "search: start addSIZE = $addSIZE, addPRIORITY=$addPRIORITY,  "
		    # start the query in the background
		    catch {
			$results_ config -title "Searching......"
			$results_ clear
		    }
		    set_state disabled
		    set servtype [$w_.cat servtype]

		    #
		    #  Depending on the server type, report different progress.
		    if {"$servtype" == "local"} {
			#	    $w_.progress config -text "Searching catalog..."
		    } elseif {"$servtype" == "imagesvr"} {
			#	    $w_.progress config -text "Attempting to contact image server..."
		    } else {
			#	    $w_.progress config -text "Attempting to contact catalog server..."
		    }

		    #
		    #  Set the catalog name to the -catalog option.
		    set catname $itk_option(-catalog)
		    #debugStatement "search: catname = $catname "


		    #
		    #  If the command line option said to convert then call the conv 
		    #  method which will calculate and add additional columns required
		    #  from the gmms software.
		    if {$addSIZE == "yes" } {
			if {$addSLITTYPE == "no" && $addPRIORITY == "no"} {
			    error_dialog "error can't add slitsize_y col, BIG error"
			    return 0
			} else {
			    #
			    #  This adds slitsize_y column.
			    #debugStatement "search: About to call col_STSIZE"
			    col_STSIZE $catname
			    set addSIZE "no"
			    set itk_option(-ysize) "no"
			}
		    } 
		    if {$addPRIORITY == "yes"} {
			#
			#  Else cmd line option said to add the priority column.
			#  Make those columns be "S"elected and "R"ectangular
			#debugStatement "search: About to add priority col"
			col_PRIO $catname
			set addPRIORITY "no"
			set itk_option(-prior) "no"
		    }
		    #debugStatement "vmAC: search addSLITTYPE=$addSLITTYPE"
		    if {$addSLITTYPE == "yes"} {
			# debugStatement "adding slittype column"
			#
			#  Else cmd line option said to add the slittype column.
			#  Make those columns be "S"elected and "R"ectangular
			# debugStatement "search: About to add slittype col"
			col_ST $catname
			set addSLITTYPE "no"
			set itk_option(-stype) "no"
		    }

		    #
		    #  Look busy and call vmAstroQuery:search, which will call do_query (bg),
		    #  and ultimately gmmps_sel, which determines what to get out of the
		    #  file and then query_done which updates the  display.
		    #
		    #    $w_.progress look_busy
		    #debugStatement "about to searchopts_ search $itk_option(-catalog) $itk_option(-uphead)"

		    $searchopts_ search $itk_option(-catalog) $itk_option(-uphead)
		    #debugStatement "search: exit... "
		}

		#############################################################
		#  Name: getimage_from_args
		#
		#  Description:
		#   Get the requested image from the image server based on 
		#   the given arguments for the world coord position or 
		#   name and the given width and height
		#############################################################
		#############################################################

		public method getimage_from_args {ra dec name equinox width height} {
		    $searchopts_ set_pos_width_height [list $ra $dec $equinox $width $height $name]
		    search
		}


		#############################################################
		#  Name: check_local_catalog
		#
		#  Description:
		#   If the given name is the name of a local catalog, check 
		#   that the file exists, and if not, ask the user for a new 
		#   path name or remove it from the menu.
		#############################################################
		#############################################################

		public proc check_local_catalog {name {id ""} {classname vmAstroCat} 
						 {debug 0} {type "catalog"} tType addPRIORITY addSLITTYPE addTILT addSIZE addXSIZE {w ""}} {
		    if {"[$astrocat_ servtype $name]" != "local"} {
			return 0
		    }
		    set file [$astrocat_ url $name]
		    if {[file isfile $file]} {
			return 0
		    }

		    set msg "The catalog file [file tail $file] does not exists. \
                     Do you want to specify a new name?"
		    switch [choice_dialog $msg {{Specify a new name} Cancel {Remove from menu}} {Specify a new name} $w] {
			{Specify a new name} {
			    set file [filename_dialog]
			    if {[file isfile $file]} {
				$astrocat_ entry remove $name
				new_catalog $file $id $classname $debug $type $tType  $addPRIORITY $addSLITTYPE $addTILT $addSIZE  $addXSIZE $w
				if {[winfo exists .catinf]} {
				    .catinf reinit_tree
				}
			    }
			}
			Cancel {
			    return 1
			}
			{Remove from menu} {
			    $astrocat_ entry remove $name
			    if {[winfo exists .catinf]} {
				.catinf reinit_tree
			    }
			}
		    }

		    # update config file and menus
		    cat::CatalogInfo::save "" $w 0
		    update_catalog_menus
		    return 1
		}


		#############################################################
		#  Name: set_pos_radius
		#
		#  Description:
		# set the values for the position and radius entries from the given
		# list, which should be in the format {ra dec equinox radius} if
		# we are using wcs, otherwise {x y radius}.
		#############################################################
		#############################################################

		public method set_pos_radius {list} {
		    $searchopts_ set_pos_radius $list
		}


		#############################################################
		#  Name: set_pos_width_height
		#
		#  Description:
		# set the values for the position, width and height entries from 
		# the given list, which should be in the format 
		# {ra dec equinox width height (in arcmin)} for wcs, or {x y width height},
		# for pixel coordinates. 
		#############################################################
		#############################################################

		public method set_pos_width_height {list} {
		    $searchopts_ set_pos_width_height $list
		}

		#############################################################
		#  Name: select_result_row
		#
		#  Description:
		#   This method is called whenever a result row is selected.
		#   Note the selected values and enable/disable some buttons
		#############################################################
		#############################################################

		public method select_result_row {} {
		    $w_.results select_result_row

		    if {! [info exists col_(PREVIEW)] && ! [info exists col_(MORE)]} {
			return
		    }
		    set row [lindex [$results_ get_selected] 0]
		    if {[llength $row] == 0} {
			return
		    }
		    set preview [get_col PREVIEW $row]
		    set more [get_col MORE $row]
		    set object_name_ [lindex $row 0]

		    # note: HST returns P=http:....
		    if {! [regsub {P=(.*)} $preview {\1} preview_url_]} {
			set preview_url_ $preview
		    }
		    if {"$preview_url_" == ""} {
			$w_.preview config -state disabled
		    } else {
			$w_.preview config -state normal
		    }

		    # note: HST returns M=http:....
		    if {! [regsub {M=(.*)} $more {\1} more_url_]} {
			set more_url_ $more
		    }
		    if {"$more_url_" == ""} {
			$w_.more config -state disabled
		    } else {
			$w_.more config -state normal
		    }
		}


		#############################################################
		#  Name: preview
		#
		#  Description:
		#    Assuming there is a field called "preview" that is the URL of an
		#    image, get the image or table data for the selected line over HTTP 
		#    and display it
		# 
		#    The preview data may be an image or other data, such as a tab table to
		#    display as a graph (like the "cuts" (RtdImageSpectrum) graph).
		#############################################################
		#############################################################

		public method preview {} {
		    if {[$results_ num_selected] == 0} {
			warning_dialog "Please select an object from the list first" $w_
			return
		    }
		    if {! [info exists col_(PREVIEW)]} {
			warning_dialog "Previewing is not supported for this catalog" $w_
			return
		    }
		    if {"$preview_url_" == ""} {
			warning_dialog "No preview for this object" $w_
			return
		    }
		    
		    # if it is a file URL, we can just get the file
		    if {[string first file: $preview_url_] == 0} {
			set filename [string range $preview_url_ 5 end]
			preview_done 0 [list $filename image/x-fits]
		    } else {
			# HTTP: do it in the background, since it may take a while
			set_state disabled
			#	    $w_.progress config -text "Attempting to contact server..."
			#	    $w_.progress look_busy

			# set up the feedback pipe
			set_feedback on

			$w_.batch bg_eval [code $this get_preview]
		    }
		}


		#############################################################
		#  Name: get_preview
		#
		#  Description:
		#   Get the preview image given by $preview_url_ into a temp file
		# 
		#############################################################
		#############################################################

		protected method get_preview {} {
		    # get the preview and call preview_done
		    $w_.cat getpreview -url $preview_url_
		}


		#############################################################
		#  Name: preview_done
		#
		#  Description:
		#    This method is called when we have received the preview data
		#    The "status" argument is the status of the background http 
		#    get operation (0 if ok).
		#    The result is a list of {filename Content-type}, where 
		#    filename contains the data and Content-type indicates the 
		#    type of the preview data. Decompression is already
		#    taken care of, so we only expect to get a FITS image here
		#    or a tab table to plot.
		#
		#    If we get an authorization error, we ask the user to 
		#    type in a username and password and retry the operation
		#    with the new info. This is needed for some
		#    HTTP server sites that have special restrictions on 
		#    data access (especially image access).
		#############################################################
		#############################################################

		protected method preview_done {status result} {
		    set_feedback off

		    if {$status} {
			# check if we need user/passwd info. errmsg should have the format:
			# "Authorization Required for <realm> at <host>"
			if {[regsub {Authorization Required} $result {Enter username} msg]} {
			    lassign [passwd_dialog $msg] username passwd
			    if {"$username" != "" && "$passwd" != ""} {
				lassign $result {} {} {} realm {} host
				$w_.cat authorize $username $passwd $realm $host
				preview
			    }
			} else {
			    error_dialog $result $w_
			}
		    } else {
			if {[llength $result] != 2} {
			    error_dialog "error getting preview data (result = $result)"
			} else {
			    lassign $result filename type
			    if {"$type" == "image/x-fits"} {
				# load the image and remove the temp file
				display_image_file $filename
			    } elseif {"$type" == "text/tab-separated-values" \
					  || "$type" == "text/x-starbase" \
					  || "$type" == "text/plain" \
					  || "$type" == ""} {
				PreviewPlot $w_.pplot[incr count_] \
				    -file $filename \
				    -name $object_name_ \
				    -shorthelpwin $this \
				    -transient 1
			    } else {
				error_dialog "unrecognized Content-type for preview data: $type"
			    }
			    if {[string first file: $preview_url_] != 0} {
				# if the image was retrieved via http, remove it
				catch {file delete $filename}
			    }
			}
		    }
		    set_state normal
		    
		    # restore preview button state...
		    select_result_row
		}

		#############################################################
		#  Name: send_to_netscape
		#
		#  Description:
		#   Send a URL to be displayed by netscape
		#############################################################
		#############################################################

		public proc send_to_netscape {url} {
		    if {[catch {exec netscape -remote "openURL($url)"} msg1]} {
			if {! [string match $msg1 "netscape: warning"]} {
			    if {[catch {exec netscape $url &} msg2]} {
				warning_dialog "couldn't start netscape: $msg1"
			    } else {
				info_dialog "Starting netscape..."
			    }
			}
		    } else {
			info_dialog "Sending help URL to netscape..."
		    }
		}


		#############################################################
		#  Name: more
		#
		#  Description:
		#   Use the "more" URL to display more info about an object 
		#   using netscape or mosaic
		#############################################################
		#############################################################

		public method more {} {
		    if {[$results_ num_selected] == 0} {
			warning_dialog "Please select an object from the list first" $w_
			return
		    }
		    if {"$more_url_" == ""} {
			warning_dialog "No more information is available for this object" $w_
			return
		    }
		    # start netscape
		    #	$w_.progress config -text "sending URL to netscape..."
		    update idletasks
		    send_to_netscape $more_url_
		}

		#############################################################
		#  Name: get_col
		#
		#  Description:
		#   Return the value for the given column name in the given row
		#############################################################
		#############################################################

		public method get_col {name row} {
		    if {[info exists col_($name)]} {
			return [lindex $row $col_($name)]
		    } 
		}


		#############################################################
		#  Name: new_catalog
		#
		#  Description:
		#   This proc is used to open a window for the named catalog, 
		#   or reuse the existing one for the catalog, if it is already open.
		#   Specifically:
		#     check that type = "local"
		#     save this instance
		#     open the catalog
		#     put the window under top level window
		#
		#  Args:
		#   name is the long name of catalog from config file
		#
		#   id is an optional unique id to be associated with a new catalog widget.
		#
		#   classname is the name of the vmAstroCat subclass to use to create new
		#   catalog widgets (defaults to "vmAstroCat").
		#
		#   debug is a flag: if true, run queries in foreground
		#
		#   type is the catalog serv_type (here it is enough to specify "catalog" or
		#   "imagesvr".
		#
		#   w should be the top level window of the caller, if specified.
		#############################################################
		#############################################################

		public proc new_catalog {name {id ""} {classname vmAstroCat} {debug 0} 
					 {type "catalog"} tType addPRIORITY addSLITTYPE addTILT addSIZE addXSIZE {w ""} iType {pixScale ""}} {
		    
		    # if it is a local catalog, 
		    #   make sure it exists still, or get a new name
		    if {[check_local_catalog $name $id $classname $debug $type $tType $addPRIORITY $addSLITTYPE $addTILT $addSIZE $addXSIZE $w] != 0} {
			return
		    }
		    
		    # Send last loaded image and catalog info to config file.
		    set loadedImage [$target_image_ cget -file ]
		    gmmps_config [list "catalog" "image" "instrument" "catalog_type"] [list $name $loadedImage $iType $tType] 
		    
		    #  Save this as an instance name, id
		    set i "$name,$id"
		    if {[info exists instances_($i)] && [winfo exists $instances_($i)]} {
			utilRaiseWindow $instances_($i)

			if {"[$instances_($i).cat servtype]" == "local"} {
			    # for local catalogs, search automatically when opened
			    $instances_($i) search
			}
			return
		    }
		    
		    # Removed next four lines, because they don't do anything useful
		    # -dfennell
		    #set col_RA ""
		    #set col_XSLIT ""
		    #catch {set col_RA [exec grep - i ra $name ]}
		    #catch {set col_XSLIT [exec grep - i slitpos_x $name ]}
		    
		    #  Open the catalog
		    if {[catch {$astrocat_ open $name} msg]} {
			error_dialog $msg
			return
		    }
		    
		    # If $w was specified, put the window under that top level window
		    # so we get the window numbers right (for cloning, see TopLevelWidget).
		    if {[winfo exists $w]} {
			set instname $w.vmac[incr n_instances_]
		    } else {
			set instname .vmac[incr n_instances_]
		    }
		    
		    set instances_($i) \
			[$classname $instname \
			     -id $id \
			     -debug $debug \
			     -catalog $name \
			     -catalogtype $type \
			     -transient 0 \
			     -center 0 \
			     -catType $tType \
			     -prior $addPRIORITY \
			     -stype $addSLITTYPE \
			     -xsize $addXSIZE \
			     -ysize $addSIZE \
			     -tilt $addTILT \
			     -pixs $pixScale \
			     -instType $iType]
		    
		}


		public method gmmps_config_ {keys {data ""}} {
		    #debugStatement "gmmps_config_ $keys $data"
		    
		    return [gmmps_config $keys $data ]
		}


		#############################################################
		#  Name: gmmps_config
		#
		#  Description:
		#	Read or write configuration data to $config_file_
		#	keys is a list of keys
		#	data is a lits of associated data
		#
		#	If no data argument is provided we are assumed to be in 
		#		read mode and the val corresponding to the keys are 
		#		returned in a list, in order they were given. 
		#	If a data argument is provided then that data is written
		#		to the config file, based on its corresponding keys.
		#		NOTE: all keys must have valid data for the method to complete.
		#############################################################
		#############################################################
		public proc gmmps_config {keys {data ""}} {
		    
		    #debugStatement "gmmps_config entry"
		    
		    # Make sure keys and data are the same length if we are in write mode.
		    if {$data != "" && ([llength $keys ] != [llength $data ])} {
			return ""
		    }
		    
		    # If config file does not exist, create a new one. 
		    if {![file exists $config_file_ ]} {
			set newConfig [open $config_file_ w ]
			::close $newConfig
		    }

		    # Open config file for reading. 
		    set gmmpsConfig [open $config_file_ r ]
		    
		    set fileKeys [list ]
		    set fileData [list ]
		    set blanksAndComments [list ]
		    set cnt 0
		    
		    # Populate fileKeys and fileData lists from config file.
		    while {[gets $gmmpsConfig pair] >= 0} {	
			# Ignore comments and empty lines.
			if {$pair == "" || [string index [string trim $pair ] 0 ] == "#"} { 
			    # Store positions and contents of any blank lines and comments,
			    # so they can be restored later.
			    set blanksAndComments [lappend blanksAndComments [list $cnt $pair ]]
			    incr cnt
			    continue
			}
			
			set pair [split [string trim $pair ] "=" ]
			set fileKeys [lappend fileKeys [string trim [lindex $pair 0 ]]]
			set fileData [lappend fileData [string trim [lindex $pair 1 ]]]
			incr cnt
		    }
		    
		    # Done reading, close file. Will reopen if a write is needed. 
		    ::close $gmmpsConfig
		    
		    if {$data == ""} {
			set targetData [list ]
		    } else {
			set targetKeys $fileKeys
			set targetData $fileData
		    }
		    
		    # Gather relevant data in targetKeys and targetData. 
		    set cnt 0
		    foreach key $keys {
			# Search for key in keys loaded form file.
			set idx [lsearch -exact $fileKeys [string trim $key ]]
			
			set newDat [lindex $data $cnt ]
			set oldDat [lindex $fileData $idx ]
			
			if {$idx != -1} {
			    # Key present in config file.
			    
			    if {$data == ""} {
				# In read mode append the data to output list.	
				set targetData [lappend targetData $oldDat ]
			    } else {
				# In write mode edit data in targetData.
				set targetData [lreplace $targetData $idx $idx $newDat ]
			    }
			} else {
			    # Key not present in config file.
			    
			    if {$data == ""} {
				# In read mode all keys must be present.
				return ""
			    } else {
				# In write mode we want to add new data.
				set targetKeys [lappend targetKeys $key ]
				set targetData [lappend targetData $newDat ]
			    }
			}
			
			incr cnt
		    }
		    
		    if {$data == ""} {
			# Read mode return data values.
			#debugStatement "gmmps_config: returning $targetData"
			return $targetData
		    } else {
			# Write mode edit file.
			
			# Open config file for writing.
			set gmmpsConfig [open $config_file_ w ]
			
			set keyCnt 0
			set lineCnt 0
			
			# Write all keys back into config file.
			foreach key $targetKeys {
			    
			    # Write blank lines and comments back into place. 
			    if {[llength $blanksAndComments ] != 0 } {
				set tempBAC $blanksAndComments
				foreach blankLine $tempBAC {
				    if {[lindex $blankLine 0 ] == $lineCnt} {
					puts $gmmpsConfig [format "%s" [lindex $blankLine 1 ]]

					# Remove first element from blank and comment list. 
					if {[llength $blanksAndComments ] > 1} {
					    set blanksAndComments [lrange $blanksAndComments 1 end ] 
					} else {
					    set blanksAndComments ""
					}
					
					incr lineCnt
				    } else {
					break
				    }
				}
			    }
			    
			    set curData [lindex $targetData $keyCnt ]
			    set line "$key=$curData"
			    
			    puts $gmmpsConfig $line
			    
			    incr lineCnt
			    incr keyCnt
			}

			# Write blank lines and comments back into place. 
			# (for after config values are done.)
			if {[llength $blanksAndComments ] != 0 } {
			    set tempBAC $blanksAndComments
			    foreach blankLine $tempBAC {
				if {[lindex $blankLine 0 ] == $lineCnt} {
				    puts $gmmpsConfig [format "%s" [lindex $blankLine 1 ]]

				    # Remove first element from blank and comment list. 
				    if {[llength $blanksAndComments ] > 1} {
					set blanksAndComments [lrange $blanksAndComments 1 end ] 
				    } else {
					set blanksAndComments ""
				    }
				    incr lineCnt
				} else {
				    break
				}
			    }
			}
			
			::close $gmmpsConfig
			
			# Return success.
			return 0
		    }
		}

		#############################################################
		#  Name: instances
		#
		#  Description:
		#   Return a Tcl list of instances of this class. By default
		#   (and for backward compatibility) only catalog windows are
		#   included - not image servers. If "choice" is "all", then
		#   all instances are returned. If choice is imagesvr, only 
		#   image server windows are returned.
		#############################################################
		#############################################################

		public proc instances {{choice catalogs}} {
		    set list {}
		    if {[info exists instances_]} {
			foreach i [array names instances_] {
			    if {[winfo exists $instances_($i)]} {
				if {[$instances_($i) iscat]} {
				    # instance of a catalog window
				    if {"$choice" == "catalogs" || "$choice" == "all"} {
					lappend list $instances_($i)
				    }
				} else {
				    # instance of an image server window
				    if {"$choice" == "imagesvr"} {
					lappend list $instances_($i) 
				    }
				}
			    }
			}
		    }
		    return $list
		}


		#############################################################
		#  Name: get_instance
		#
		#  Description:
		#   Return the widget instance for the given catalog, if found,
		#   otherwise an empty string
		#############################################################
		#############################################################

		public proc get_instance {catalog} {
		    foreach w [cat::vmAstroCat::instances all] {
			if {"[$w cget -catalog]" == "$catalog"} {
			    return $w
			}
		    }
		}


		#############################################################
		#  Name: iscat
		#
		#  Description:
		#   Return 1 if this window is for a searchable catalog, 
		#   and 0 otherwise (in which case it must be an image server window).
		#############################################################
		#############################################################

		public method iscat {} {
		    return $iscat_
		}


		# ---------------------------------------------------------------
		# The following methods deal with images and plotting of objects
		# in an image and are meant to be defined in a derived class.
		# See the skycat/skycat/interp/library/SkySearch.tcl for an example.
		# ---------------------------------------------------------------


		# see plot method in vmSkySearch.tcl -df 

		public method plot {} {
		}


		# display the given image file (To be defined in a subclass)

		public method display_image_file {filename} {
		}


		#  return the equinox

		public method get_equinox {} {
		    return [component searchopts get_equinox]
		}


		#  return the name of the TableList
		public method get_table {} {
		    if {[info exists itk_component(results)]} {
			return $itk_component(results)
		    }
		    error "This catalog is an image server"
		}


		#############################################################
		#  Name: getPixelScale
		#
		#  Description:
		#   Finds the pixel scale based on instrument type and detector ID.
		#  
		#############################################################
		#############################################################

		public proc getPixelScale {} {
		    set home $::env(GMMPS)
		    global image2

		    set targetDetId [string trim [get_detid ]]
		    
		    # Find binning from fits image.
		    set ccdsum [string trim [$target_image_ fits get "CCDSUM" ]]

		    # Try again with different method if nothing returned, this can happen
		    # if INHERIT = T. This reads CCDSUM from the extension, just as the 
		    # above method does. Bryan Miller, 2013-03-18
		    if {$ccdsum == ""} {
			set ccdsum [get_ccdsum_image ]
			#puts "get_ccdsum_image: $ccdsum"
		    }

		    set binning [lindex $ccdsum 0]
		    #puts "binning: $binning"

		    # Open detector config file for reading. 
		    set detFile [open "$home/config/Detectors.dat" r ]
		    
		    set pixScale 0.0
		    set inst ""

		    while {[gets $detFile line] >= 0} {
			
			if {$targetDetId == "NOT FOUND"} {
			    break
			} elseif {$line == ""} {
			    continue
			} elseif {[string index $line 0 ] == "#" } {
			    continue
			}
			
			set line [split $line "\t" ]	
			
			# Extract instrument type and detector id from detector property list. 
			# List order: (more info in config/Detectors.dat)
			# Currently-Installed Instrument-Name Detector-Id Detector-Name Gap1 Gap2 GapL Pixel-Scale
			set inst  [string trim [lindex $line 1 ]]
			
			# What is the deal with flam? 
			if {$inst == "flam" || $inst == "Flam"} {
			    set inst "F2"
			}
			
			set detId [string trim [lindex $line 2 ]]
			
			#debugStatement "setPixelScale: .$detId. == .$targetDetId."
			
			if {$detId == $targetDetId} {
			    # If this id matches the id we are searching for, then find the pixelscale.
			    
			    if {$inst == "F2"} {
				set aofold [$target_image_ fits get AOFOLD]

				# Set F2 pixelscale based on aofold value. 
				# more explanation in config/Detectors.
				if {$aofold == "in"} {
				    set pixScale [string trim [lindex $line 5 ]]
				} else {
				    set pixScale [string trim [lindex $line 4 ]]
				}
				
				break
			    } else {
				# GMOS.
				
				set pixScale [string trim [lindex $line 7 ]]
				break
			    }
			}
		    }
		    
		    ::close $detFile
		    
		    # Make sure a pixelScale was found and apply binning.
		    if {$pixScale != 0.0} {
			set pixScale [expr $binning * $pixScale ]
		    } else {
			# If pixelScale was not found we assume this is a fake preimage. 
			# So, search for pixelscale in image .fits.
			
			#puts "No match for $inst, $targetDetId... must be fake preimage?"
			
			set pixScale [get_pixelscale_image ] 
			if {$pixScale == "NOT FOUND"} {
			    set pixScale [string trim [$target_image_ fits get "PIXSCALE" ]] 
			}
			
			#set pixScale [expr $binning * [get_pixelscale_image ]]
			
			if {$pixScale == ""} {
			    set pixScale [expr $binning * 0.073 ]
			    error_dialog "No pixel scale found. Setting to default: $pixScale."
			} else {
			    puts "Warning: Detector ID not found or did not match. Pixel scale taken from the fits image: $pixScale"
			}
		    }
		    #puts "found pixelscale $pixScale for inst $inst"
		    
		    return [list $pixScale $inst ]
		    
		    # read from the fits file the instType:
		    #catch  set itk_option(-pixs) [exec grep "#fits PIXSCALE =" $itk_option(-catalog) | cut -f2 -d= ] 
		}


		#############################################################
		#  Name: getInstType
		#
		#  Description:
		#   Finds the instrument type from the catalog file. If that is not possible
		#	then set it based on pixelScale.
		#  
		#############################################################
		#############################################################

		protected proc getInstType {name pixScale detInst} {

		    set instType ""

		    # Grep for instrument type.
		    if {[catch {set instType [exec grep "#fits INSTRUME=" $name | cut -f2 -d= ]} msg]} {
			catch {set instType [exec grep "#fits INSTRUME =" $name | cut -f2 -d= ]}
		    }
		    # Make sure that the instrument name is capitalized, BM 2013-08-14
		    set instType [string toupper $instType]
		    
		    if {$instType != ""} {
			# Grepping worked, do a bit of formatting.
			
			# Clean up instType a bit. 
			set instType [string trim $instType ]
			if {[string index $instType 0 ] == "'" } {
			    set instType [string range $instType 1 end ]
			}
			
			if {[string index $instType end ] == "'"} {
			    set instType [string range $instType 0 end-1 ]
			}
			
			set instType [string trim $instType ]
		    } else {
			# Grepping DID NOT work, so we need to check our pixelscale
			# against the one in the catalog file. 
			
			if {[catch {set pixScaleCat [exec grep "#fits PIXSCALE =" $name | cut -f2 -d= ]} msg]} {
			    # even grepping the catalog file didn't work! Last resort...
			    
			    set instType [get_insttype_image ]
			    #debugStatement "instType set from image: $instType"
			    
			    if {$instType != "NOT FOUND"} {
				catch {destroy .badInst}	
				tk_dialog .badInst "Instrument type not found" "Warning: Instrument was not found in the catalog file. $instType loaded from the preimage.\nPlease be sure you are using the correct preimage for this catalog." "" 0 "OK"
				catch {destroy .badInst}
			    } else {
				# No mas! 
				
				return 1
			    }
			    
			} else {
			    # Grep for pixelscale successful. 
			    
			    set otPixScale [format %f [string trim $pixScaleCat ]]
			    
			    if {$pixScale < [expr $otPixScale + 0.00001 ] && $pixScale > [expr $otPixScale - 0.00001]} {
				set instType $detInst
			    } else {
				set instType [get_insttype_image ]
				#debugStatement "instType set from image: $instType"
				
				if {$instType != "NOT FOUND"} {
				    catch {destroy .badInst}	
				    tk_dialog .badInst "Instrument type not found" "Warning: Instrument type was not found in catalog file. $instType loaded from preimage.\nPlease be sure you are using the correct preimage for this catalog." "" 0 "OK"
				    catch {destroy .badInst}
				} else {
				    # No mas! 
				    
				    return 1
				}
			    }
			}
		    }
		    
		    # REALLY old Flamingos-2 images might have this name.
		    if {$instType == "Flam" || $instType == "flam"} {
			set $instType "F2"
		    }
		    
		    return $instType
		}


		#############################################################
		#  Name: setDetInfo
		#
		#  Description:
		#   Search for detector information in the preimage file and detector
		#	configuration file.
		#############################################################
		#############################################################	

		protected method setDetInfo {instType} {
		    
		    # Find DET_SPEC_ based on instType and info in detector config file.
		    set detFile [open "$home_/config/Detectors.dat" r ]
		    
		    while {[gets $detFile line ] >= 0} {	
			#Parse detector config data(see config/Detectors.dat for more info)
			set line [split $line "\t" ]
			
			set lineInst [string trim [lindex $line 1 ]]
			set installed [string trim [lindex $line 0 ]]
			
			if {$lineInst == $instType && $installed == 1} {
			    # Found the installed detector for our instrument! 
			    
			    set DET_SPEC_ [string trim [lindex $line 2 ]]
			    break
			}
		    }
		    ::close $detFile
		    
		    # Find DET_IMG from preimage det_id.
		    set DET_IMG_ [get_detid ]
		    if {$DET_IMG_ == "NOT FOUND" || $DET_IMG_ == ""} {
			set DET_IMG_ "UNKNOWN"
		    }
		    
		    return
		}



		#############################################################
		#  Name: is_sorted
		#
		#  Description: Inputs a list of column rows and outputs 1 or 0, depending
		#	if the list is in sorted ascending order in the ID column, or not. 
		#   
		#############################################################

		protected method is_sorted {rows} {
		    set id_col 0
		    set last -1
		    
		    foreach row $rows {
			set cur [lindex $row $id_col ]
			if {$cur > $last} {
			    set last $cur
			} else {
			    return 0
			}
		    }
		    return 1
		}


		#############################################################
		#  Name: sort_rowlist
		#
		#  Description: Sorts a given rowlist by id, and returns the sorted list 
		#	(increasing order).
		#############################################################

		protected method sort_rowlist {rows} {
		    set id_col 0
		    # Sort this data. 
		    set sorted_rows [lsort -integer -index $id_col $rows ]
		    return $sorted_rows
		}



		#############################################################
		# Fetch detector id from the fits image. 
		#############################################################

		protected proc get_detid {} {
		    # Find target ID from fits image.		

		    set targetDetId "NOT FOUND"
		    set fits [$target_image_ hdu fits 1 ]
		    foreach header [split $fits "\n" ] {
			if {[string range $header 0 4 ] == "DETID"} {
			    set targetDetId [lindex [split $header "'" ] 1 ]
			    break
			}
		    }
		    
		    return [string trim $targetDetId ]
		}

		#############################################################
		#############################################################

		protected proc get_insttype_image {} {
		    set targetInst "NOT FOUND"
		    set fits [$target_image_ hdu fits 1 ]
		    
		    foreach header [split $fits "\n" ] {
			#puts $header
			if {[string range $header 0 7 ] == "INSTRUME"} {
			    # Make sure targetInst is upper case, BM 2014-08-14
			    set targetInst [string toupper [lindex [split $header "'" ] 1 ]]
			    break
			}
		    }
		    
		    return [string trim $targetInst ]
		}


		#############################################################
		# Fetch pixelscale from fits image. 
		#############################################################

		protected proc get_pixelscale_image {} {
		    
		    set pixelScale "NOT FOUND"
		    set fits [$target_image_ hdu fits 1 ]
		    foreach header [split $fits "\n" ] {
			if {[string range $header 0 7 ] == "PIXSCALE"} {
			    set pixelScale [lindex [split [lindex [split $header "=" ] 1 ] "/" ] 0 ]
			    break
			}
		    }
		    
		    return [string trim $pixelScale ]
		}

		#############################################################
		# Fetch ccdsum from fits image extension, not the PHU (Bryan Miller, 2013/03/18)
		#############################################################

		protected proc get_ccdsum_image {} {
		    
		    set ccdsum "NOT FOUND"
		    set fits [$target_image_ hdu fits 1 ]
		    foreach header [split $fits "\n" ] {
			if {[string range $header 0 5 ] == "CCDSUM"} {
			    set ccdsum [lindex [split [lindex [split $header "=" ] 1 ] "/" ] 0 ]
			    break
			}
		    }
		    
		    return [string trim $ccdsum "' "]
		}

		#############################################################
		# Return the name of this skycat instance. 
		#############################################################

		public method get_skycat {} {
		    
		    #debugStatement "getSKYCATTTTTTTT: $this"
		    #debugStatement [string range $this 2 8 ]
		    
		    if {[string range $this 2 8 ] == ".skycat"} {
			return [string range $this 2 9 ]
		    } else { 
			return 1
		    }
		}



		#############################################################
		# -- options --
		#############################################################

		# name of catalog
		itk_option define -catalog catalog Catalog "" {
		    # make sure we use full path name for local catalogs
		    set f $itk_option(-catalog)
		    if {[file exists $f] && "[string index $f 0]" != "/"} {
			set itk_option(-catalog) [pwd]/$f
		    }
		}

		# type of catalog (catalog or archive)
		itk_option define -catalogtype catalogType CatalogType "catalog"

		# Optional unique id, used in searching for already existing catalog widgets.
		itk_option define -id id Id ""

		# Contains the type of file loaded in, Object Tbl, Master OT , or fitsFile 
		itk_option define -catType catType CatType "1"

		# Options indicating col's that exist
		itk_option define -prior prior Prior "no"
		itk_option define -stype stype SType "R"
		itk_option define -xsize xsize XSize "no"
		itk_option define -ysize ysize YSize "no"
		itk_option define -tilt tilt Tilt "no"
		itk_option define -uphead uphead UpHead {}
		itk_option define -headingx headingx Headingx {}
		itk_option define -pixs pixs Pixs "0.0725" 


		# list of catalog column headings (from results of most recent query)
		protected variable headings_ {}

		# result from most recent query (list of rows)
		protected variable info_ {}

		# font to use for labels
		itk_option define -labelfont labelFont LabelFont -Adobe-helvetica-bold-r-normal--12*

		# font to use for values
		itk_option define -valuefont valueFont ValueFont -Adobe-helvetica-medium-r-normal--12*

		# font to use for ra,dec labels
		itk_option define -wcsfont wcsFont WcsFont -*-symbol-*-*-*-*-14-*-*-*-*-*-*-*

		# set the width for  displaying labels
		itk_option define -labelwidth labelWidth LabelWidth 12

		# set the anchor for labels
		itk_option define -anchor anchor Anchor e

		# flag: if true, run queries in foreground for better debugging
		itk_option define -debug debug Debug 0

		# set central wavelength
		itk_option define -bw bw Bw ""

		# vmAstroQuery widget used to manage search options
		protected variable searchopts_ {}

		# QueryResult widget used to display search results
		protected variable results_ {}

		# array(uppercase col name) of col index from catalog headings
		protected variable col_

		# preview URLs for selected object
		protected variable preview_url_ {}

		# more-info URLs for selected object
		protected variable more_url_ {}

		# pipe to read feedback during HTTP transfer
		protected variable rfd_

		# pipe to write feedback during HTTP transfer
		protected variable wfd_

		# current state: normal, disabled (i.e.: waiting)
		protected variable state_ {normal}

		# log file handle (used to log URLs)
		protected variable logfile_name_ 

		# flag: set at end of constructor
		protected variable initialized_ 0

		# index of this object in the instances_ array
		protected variable instance_idx_ {}

		# name of File menu widget
		protected variable file_menu_

		# name of Edit menu widget
		protected variable edit_menu_

		# name of Options menu widget
		protected variable options_menu_

		# name of Name Server menu widget
		protected variable ns_menu_

		# flag: set when the column headings are changed between TCS and normal
		protected variable reset_columns_ {0}

		# currently selected object (Id field in row)
		protected variable object_name_ {}

		# count used for filename generation
		protected variable count_ 0

		# flag: true if catalog is not an image server
		protected variable iscat_ 1

		# flag: true if searching is allowed
		protected variable search_state_ normal

		# flag: true if slits are being plotted for this catalog
		protected variable plotSlits_ 0

		# flag: true if nod and shuffle bands are being plotted.
		protected variable plotBands_ 0

		# The detector ID of the image loaded into skycat.
		protected variable DET_IMG_ ""

		# The detector ID that the mask is intended for.
		protected variable DET_SPEC_ ""


		# Formatting variables. 

		protected variable ot_text_ "black"
		protected variable ot_bg_ "grey86"
		protected variable ot_bg2_ "grey86"

		protected variable button_text_ot_ "black"
		protected variable button_text_active_ot_ "black"
		protected variable button_bg_ot_ "grey86"
		protected variable button_bg_active_ot_ "white"

		protected variable catalog_text_ot_ "black"
		protected variable catalog_text_active_ot_ "black"
		protected variable catalog_bg_ot_ "white"
		protected variable catalog_bg_active_ot_ "yellow"

		protected variable menu_bg_ot_ "firebrick"
		protected variable menu_text_ot_ "white"
		protected variable menu_bg_active_ot_ "white"
		protected variable menu_text_active_ot_ "black"

		protected variable detRadioColor_ "black"

		protected variable odf_text_ "black"
		protected variable odf_bg_ "grey86"
		protected variable odf_bg2_ "grey86"

		protected variable button_text_odf_ "black"
		protected variable button_text_active_odf_ "black"
		protected variable button_bg_odf_ "grey86"
		protected variable button_bg_active_odf_ "white"

		protected variable catalog_text_odf_ "black"
		protected variable catalog_text_active_odf_ "black"
		protected variable catalog_bg_odf_ "white"
		protected variable catalog_bg_active_odf_ "yellow"

		protected variable menu_bg_odf_ "DodgerBlue3"
		protected variable menu_text_odf_ "white"
		protected variable menu_bg_active_odf_ "white"
		protected variable menu_text_active_odf_ "black"


		#	protected variable ot_text_ "black"
		#	protected variable ot_bg_ "snow"
		#	protected variable ot_bg2_ "snow"

		#	protected variable button_text_ot_ "white"
		#	protected variable button_text_active_ot_ "MidnightBlue"
		#	protected variable button_bg_ot_ "OrangeRed2"
		#	protected variable button_bg_active_ot_ "white"

		#	protected variable catalog_text_ot_ "black"
		#	protected variable catalog_text_active_ot_ "white"
		#	protected variable catalog_bg_ot_ "snow2"
		#	protected variable catalog_bg_active_ot_ "MidnightBlue"

		#	protected variable odf_text_ "black"
		#	protected variable odf_bg_ "snow"
		#	protected variable odf_bg2_ "snow"

		#	protected variable button_text_odf_ "white"
		#	protected variable button_text_active_odf_ "OrangeRed2"
		#	protected variable button_bg_odf_ "MidnightBlue"
		#	protected variable button_bg_active_odf_ "white"

		#	protected variable catalog_text_odf_ "black"
		#	protected variable catalog_text_active_odf_ "white"
		#	protected variable catalog_bg_odf_ "snow2"
		#	protected variable catalog_bg_active_odf_ "OrangeRed2"

		#protected variable ot_menu_bg_ "blue"
		#protected variable ot_menu_text_ "black"


		# -- common variables (common to all instances of this class) --

		# C++ astrocat object used by static member procs
		protected common astrocat_ [astrocat ::cat::.vmastrocat]

		# A reference to the band_def_UI window. Useful if we want to destroy
		# previous windows on new window loads.
		protected common bw_ ""

		# A reference to the mask design window, if open.
		protected common spoc_ ""

		# array mapping catalog name to widget/class name
		protected common instances_

		# instance count
		protected common n_instances_ 0

		# current instance name
		protected common current_instance_ {}

		# array(TopLevelWidget instance) of command used to update the
		# Data-Servers menu. This is used to make it posible to update
		# all instances of this menu in various top level windows.
		protected common catalog_menu_info_

		# flag: set to 1 after we checked for a proxy server
		protected common checked_proxies_ 0

		# Reference to the skycat image widget. 
		protected common target_image_ image2

		protected common home_ $::env(GMMPS)

		# GMMPS configuration filename
		protected common config_file_ [format "%s/.gmmps.cfg" $::env(HOME) ]

		# Information about the detector being used. A list. 
		# Will be set to 0 on construction.
		# GMOS Order: Currently-Installed Instrument-Name Detector-Id Detector-Name Gap1 Gap2 GapL Pixel-Scale
		# For F2:     Curently-Installed Instrument-Name Detector-Id Detector-Name Pixel-Scale Pixel-Scale-AO
		protected common detector_ 0

		# flag used when plotting GMOS detector gaps.
		protected common plotGaps_ 0


	    }
